/***********************************************************************

    This file is part of
    FEINS, Finite Element Incompressible Navier-Stokes solver,
    which is expanding to a more general FEM solver and toolbox,
    Copyright (C) 2003--2008, Rene Schneider 
    <rene.schneider@mathematik.tu-chemnitz.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.

    Minor contributions to this program (for example bug-fixes and
    minor extensions) by third parties automatically transfer the
    copyright to the general author of FEINS, to maintain the
    possibility of commercial re-licensing. If you contribute but wish
    to keep the copyright of your contribution, make that clear in
    your contribution!

    Non-GPL licenses to this program are available upon request from
    the author.

************************************************************************/
/*
FILE mesh.c
HEADER mesh.h

TO_HEADER:


// useful text macros
#include "feins_macros.h"
// data structures 
#include "datastruc.h"
#include "sparse_struct.h"
*/



/* function prototypes */
#include "mesh.h"


/* prototypes of external functions */
//#include <math.h>
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <string.h>

#include "sparse.h"

/*FUNCTION*/
void mesh_init( struct mesh *m
/* initialises the mesh to be empty,

   Output: m        - an empty mesh, all inner pointers set to NULL,
                      array lengths etc. to zero
*/
		){
  (*m).dim     = 0;
  (*m).problem = 0;

  (*m).vx_w   = 0;
  (*m).vx_nr  = 0;
  (*m).vx_max = 0;
  (*m).vertex = NULL;

  (*m).el_w   = 0;
  (*m).el_nr  = 0;
  (*m).el_max = 0;
  (*m).elem   = NULL;

  (*m).eg_w   = 0;
  (*m).eg_nr  = 0;
  (*m).eg_max = 0;
  (*m).edge   = NULL;

  (*m).fc_w   = 0;
  (*m).fc_nr  = 0;
  (*m).fc_max = 0;
  (*m).face   = NULL;

  (*m).vo_w   = 0;
  (*m).vo_nr  = 0;
  (*m).vo_max = 0;
  (*m).vols   = NULL;

  (*m).hi_w   = 0;
  (*m).hi_nr  = 0;
  (*m).hi_max = 0;
  (*m).hier   = NULL;

  (*m).eh_w   = 0;
  (*m).eh_nr  = 0;
  (*m).eh_max = 0;
  (*m).elhier = NULL;

  (*m).bd_w   = 0;
  (*m).bd_nr  = 0;
  (*m).bd_max = 0;
  (*m).bound = NULL;

  (*m).ps_w   = 0;
  (*m).ps_nr  = 0;
  (*m).ps_max = 0;
  (*m).pcsurf = NULL;

  (*m).pv_w   = 0;
  (*m).pv_nr  = 0;
  (*m).pv_max = 0;
  (*m).pcvol = NULL;

  (*m).pc_w   = 0;
  (*m).pc_nr  = 0;
  (*m).pc_max = 0;
  (*m).pccrit = NULL;

  (*m).fu_w   = 0;
  (*m).fu_nr  = 0;
  (*m).fu_max = 0;
  (*m).func   = NULL;

  (*m).pa_w   = 0;
  (*m).pa_nr  = 0;
  (*m).pa_max = 0;
  (*m).para   = NULL;

  (*m).sg_w   = 0;
  (*m).sg_nr  = 0;
  (*m).sg_max = 0;
  (*m).sseg   = NULL;

  (*m).sp_w   = 0;
  (*m).sp_nr  = 0;
  (*m).sp_max = 0;
  (*m).spar   = NULL;

  return;
}




/********************************************************************/
/*                                                                  */
/*         T1 element stuff                                         */
/*                                                                  */
/********************************************************************/


/*FUNCTION*/
int mesh_refine_uniform_t1( struct mesh *m 
/* refines the all elements of the mesh m, all necessarry new parts are
   created thereby

   In/Out: m        - the mesh, is modified accordingly, e.g. new
                      elements, faces, edges, vertices, boundary
                      elements and hierarchy entries are created,
		      the element el and the face el are thereby
		      replaced by one of their childs, the edges
		      remain to allow other elements to stay intact,
		      if necessary the fields of the mesh are
		      reallocated to have space for the new stuff

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX old_el_nr, i;
  int err;

  old_el_nr=(*m).el_nr;

  /* loop over all elements */
  for (i=0; i< old_el_nr; i++)
    {
      err=mesh_refine_element_t1( m , i);
      FUNCTION_FAILURE_HANDLE( err, mesh_refine_element_t1, 
			       mesh_refine_uniform_t1 ); 
    }
  return SUCCESS;
}


/*FUNCTION*/
int mesh_refine_element_t1( struct mesh *m , FIDX el
/* refines the element el of the mesh m, all necessarry new parts are
   created thereby

   Input:  el       - the element/face to be refined

   In/Out: m        - the mesh, is modified accordingly, e.g. new
                      elements, faces, edges, vertices, boundary
                      elements and hierarchy entries are created,
		      the element el and the face el are thereby
		      replaced by one of their childs, the edges
		      remain to allow other elements to stay intact,
		      if necessary the fields of the mesh are
		      reallocated to have space for the new stuff

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX i, j, fc_w, el_w, eg_w, ps_w, pv_w;
  int err;


  FIDX newedge[3], midnodes[3], oldnod[3], nodedge[3][2], nodineg[3][2];
  /* nodedge[lnode][this nodes edge]=the edge (global number),
     nodineg[lnode][this nodes edge]=the node is 1st/2nd of the edge */

  fc_w=(*m).fc_w;
  el_w=(*m).el_w;
  eg_w=(*m).eg_w;
  ps_w=(*m).ps_w;
  pv_w=(*m).pv_w;

  /* clear nodedge */
  for (i=0; i<3; i++)
    for (j=0; j<2; j++)
      nodedge[i][j]=-1;

  /* get oldnod */
  for (i=0; i<3; i++)
    oldnod[i]=(*m).elem[el*el_w+MCT1ELNOD1+i];

  /* make sure the 3 edges are refined, store the midnodes, get
     nodedge */ 
  for (i=0; i<3; i++)
    {
      FIDX eg, chld;

      eg=(*m).face[el*fc_w+MCT1FCEDG1+i];

      /* make sure this edge is refined */
      if ((*m).edge[eg*eg_w+MCT1EGCHL1]==-1)
	{
	  err=mesh_split_edge_t1( m, eg);
	  FUNCTION_FAILURE_HANDLE( err, mesh_split_edge_t1,
				   mesh_refine_element_t1); 
	}
      
      /* store the midnode */
      chld=(*m).edge[eg*eg_w+MCT1EGCHL1];
      /* second node of child edge is always the midnode of the old
	 edge */ 
      midnodes[i]=(*m).edge[chld*eg_w+MCT1EGNOD1+1]; 

      for (j=0; j<2; j++)
	{
	  FIDX amnodeedge, thenode, intnode;
	  thenode=(*m).edge[eg*eg_w+MCT1EGNOD1+j];
	  intnode=0;
	  /* find the apropriate internal node for this guy */
	  while ((intnode<3)&&(oldnod[intnode]!=thenode))
	    intnode++;
	  /* if not found, cry */
	  if (intnode>=3)
	    {
	      fprintf(stderr, "internal mesh consistency error in mesh_refine_element_t1, file %s, line %d\n", __FILE__, __LINE__);
	      return FAIL;
	    }

	  /* find where in nodedge and nodineg this one goes to */
	  if (nodedge[intnode][0]==-1)
	    { amnodeedge=0; }
	  else
	    { amnodeedge=1; }
	  nodedge[intnode][amnodeedge]=eg;
	  nodineg[intnode][amnodeedge]=j;
	}
    }

  /* make sure we have enough space for the 3 new edges in the inner
     of the element, 3 new faces (the old is recycled as a 4th), 3 new
     elements (the old is recycled as a 4th), 3 new pcvol if needed
     (the old is recycled as a 4th) */ 
  if ((*m).eg_nr+3>(*m).eg_max)
    {
      err=mesh_more_edges( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_edges, mesh_refine_element_t1 );
    }
  if ((*m).fc_nr+3>(*m).fc_max)
    {
      err=mesh_more_faces( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_faces, mesh_refine_element_t1 );
    }
  if ((*m).el_nr+3>(*m).el_max)
    {
      err=mesh_more_elems( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_elems, mesh_refine_element_t1 );
    }

  if (((*m).face[el*fc_w+MCT1FCPCVL]!=-1)&&((*m).pv_nr+3>(*m).pv_max))
    {
      err=mesh_more_pcvol( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_pcvol, mesh_refine_element_t1 );
    }

  /* for all old nodes */
  for (i=0; i<3; i++)
    {
      FIDX chld0, chld1, midnod0, midnod1, newface, newel, level;

      /* get the child edges containig this node */
      chld0=(*m).edge[nodedge[i][0]*eg_w+MCT1EGCHL1+nodineg[i][0]];
      chld1=(*m).edge[nodedge[i][1]*eg_w+MCT1EGCHL1+nodineg[i][1]];

      /* the refinement level of the new edges */
      level=(*m).edge[chld0*eg_w + MCT1EGLVL];
  
      /* get the corresponding midnodes */
      midnod0=(*m).edge[chld0*eg_w+MCT1EGNOD1+1]; 
      midnod1=(*m).edge[chld1*eg_w+MCT1EGNOD1+1];
      /* second node of child edge is allway the midnode of the old
	 edge */

      /* create a new edge from the midnodes */
      newedge[i]=(*m).eg_nr;
      (*m).eg_nr++;

      (*m).edge[newedge[i]*eg_w+MCT1EGNOD1  ]=midnod0;
      (*m).edge[newedge[i]*eg_w+MCT1EGNOD1+1]=midnod1;
      (*m).edge[newedge[i]*eg_w+MCT1EGCHL1  ]=-1;
      (*m).edge[newedge[i]*eg_w+MCT1EGBND   ]=-1;
      (*m).edge[newedge[i]*eg_w+MCT1EGPCSU  ]=-1;
      (*m).edge[newedge[i]*eg_w+MCT1EGLVL   ]=level;

      /* create a new face */
      newface=(*m).fc_nr;
      (*m).fc_nr++;

      (*m).face[newface*fc_w+MCT1FCEDG1  ]=chld0;
      (*m).face[newface*fc_w+MCT1FCEDG1+1]=chld1;
      (*m).face[newface*fc_w+MCT1FCEDG1+2]=newedge[i];

      (*m).face[newface*fc_w+MCT1FCRHSF]=(*m).face[el*fc_w+MCT1FCRHSF];

      /* if necessary create a new pcvol entry */
      if ((*m).face[el*fc_w+MCT1FCPCVL]!=-1)
	{
	  FIDX pcv, new_;

	  pcv=(*m).face[el*fc_w+MCT1FCPCVL];
	  new_=(*m).pv_nr;
	  
	  (*m).pcvol[new_*pv_w+MCXXPVVOLM]=newface;
	  (*m).face[newface*fc_w+MCT1FCPCVL]=new_;
	  (*m).pcvol[new_*pv_w+MCXXPVCRIT]=(*m).pcvol[pcv*pv_w+MCXXPVCRIT];

	  (*m).pv_nr++;
	}
      else
	{
	  (*m).face[newface*fc_w+MCT1FCPCVL]=-1;
	}

      /* check if the edges of the new face are pcsurf edges, if so,
	 let them know the name of the new face */
      for (j=0; j<3; j++)
	{
	  FIDX eg, pcsu;
	  eg=(*m).face[newface*fc_w+MCT1FCEDG1+j];
	  pcsu=(*m).edge[eg*eg_w+MCT1EGPCSU];
	  if (pcsu!=-1)
	    {
	      if ((*m).pcsurf[pcsu*ps_w+MCXXPSVOLE]==-1)
		{
		  (*m).pcsurf[pcsu*ps_w+MCXXPSVOLE]=newface;
		  (*m).pcsurf[pcsu*ps_w+MCXXPSVOLE+1]=-1;
		}
	      else
		{
		  (*m).pcsurf[pcsu*ps_w+MCXXPSVOLE+1]=newface;
		}
	    }
	}

      /* create a new element */
      newel=(*m).el_nr;
      (*m).el_nr++;

      (*m).elem[newel*el_w+MCT1ELNOD1  ]=oldnod[i];
      (*m).elem[newel*el_w+MCT1ELNOD1+1]=midnod0;
      (*m).elem[newel*el_w+MCT1ELNOD1+2]=midnod1;
      
      /* next node */
    }

  /* the outer 3 elements are correctly created now,
     so lets overwrite the old element with the inner one, the 4-th */

  for (i=0; i<3; i++)
    {
      /* the face */
      (*m).face[el*fc_w+MCT1FCEDG1+i]=newedge[i];
      /* ( rhs-function and pcvol stays the same ) */
      /* the edges of the inner face can not be pcsurf edges */

      /* the element */
      (*m).elem[el*el_w+MCT1ELNOD1+i]=midnodes[i];
    }

  /* that's it */

  return SUCCESS;
}

/*FUNCTION*/
int mesh_split_edge_t1( struct mesh *m , FIDX eg
/* refines the edge eg of the mesh m, so that the first node of eg is
   the first node of the first child edge, the second node is the
   first node of the second child edge and the new middle node is the
   second node of both child edges,
   all necessarry new parts are created

   Input:  eg       - the edge to be refined

   In/Out: m        - the mesh, is modified accordingly, e.g. new
                      edges, vertices, boundary elements and hierarchy
                      entries are created, 
		      thereby the edge remains to allow neigbouring elements
		      to stay intact, 
		      if necessary the fields of the mesh are
		      reallocated to have space for the new stuff

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX i, dim, vx_w, eg_w, hi_w, bd_w, ps_w;
  int  err;
  FIDX newnode, nod1, nod2;
  FIDX neweg1, neweg2;
  FIDX oldbd, newbd1,newbd2;
  FIDX newhi, level;

  dim=(*m).dim;
  vx_w=(*m).vx_w;
  eg_w=(*m).eg_w;
  hi_w=(*m).hi_w;
  bd_w=(*m).bd_w;
  ps_w=(*m).ps_w;

  if ((*m).edge[eg*eg_w+ MCT1EGCHL1 ]!= -1)
    {
      /* nothing to do, return */
      return SUCCESS;
    }

  /* first we have to make sure we have space for one new vertex, two
     new edges, eventually 2 new boundary elements, eventually 2 new
     pcsurf elements, one hierarchy entry 
  */
  if ((*m).vx_nr+1>(*m).vx_max)
    {
      err=mesh_more_vertices( m, 1);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_vertices,  mesh_split_edge_t1 );
    }
  if ((*m).eg_nr+2>(*m).eg_max)
    {
      err=mesh_more_edges( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_edges,  mesh_split_edge_t1 );
    }
  if (((*m).edge[eg*eg_w+ MCT1EGBND]!= -1)&&((*m).bd_nr+2>(*m).bd_max))
    {
      err=mesh_more_boundary( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_boundary,  mesh_split_edge_t1 );
    }
  if (((*m).edge[eg*eg_w+ MCT1EGPCSU]!= -1)&&((*m).ps_nr+2>(*m).ps_max))
    {
      err=mesh_more_pcsurf( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_pcsurf,  mesh_split_edge_t1 );
    }
  if ((*m).hi_nr+1>(*m).hi_max)
    {
      err=mesh_more_hierarchy( m, 1);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_hierarchy,  mesh_split_edge_t1 );
    }

  /* nodes of the edge */
  nod1=(*m).edge[eg*eg_w+ MCT1EGNOD1];
  nod2=(*m).edge[eg*eg_w+ MCT1EGNOD1+1];

  /* create the new node */
  newnode= (*m).vx_nr;
  (*m).vx_nr++;
  for (i=0;i<dim; i++)
    {
      (*m).vertex[newnode*vx_w+MCT1VXSTRT+i]=
	0.5*(*m).vertex[nod1*vx_w+MCT1VXSTRT+i]+
	0.5*(*m).vertex[nod2*vx_w+MCT1VXSTRT+i];
    }

  /* refinement level of the old edge */
  level=(*m).edge[eg*eg_w+ MCT1EGLVL];
  
  /* create the new edges */
  neweg1=(*m).eg_nr;
  neweg2=(*m).eg_nr+1;
  (*m).eg_nr+=2;
  /* child edge 1 */
  (*m).edge[neweg1*eg_w+MCT1EGNOD1  ]=nod1;
  (*m).edge[neweg1*eg_w+MCT1EGNOD1+1]=newnode;
  (*m).edge[neweg1*eg_w+MCT1EGCHL1  ]=-1;
  (*m).edge[neweg1*eg_w+MCT1EGBND   ]=-1;
  (*m).edge[neweg1*eg_w+MCT1EGPCSU  ]=-1;
  (*m).edge[neweg1*eg_w+MCT1EGLVL   ]=level+1;
  /* child edge 2 */
  (*m).edge[neweg2*eg_w+MCT1EGNOD1  ]=nod2;
  (*m).edge[neweg2*eg_w+MCT1EGNOD1+1]=newnode;
  (*m).edge[neweg2*eg_w+MCT1EGCHL1  ]=-1;
  (*m).edge[neweg2*eg_w+MCT1EGBND   ]=-1;
  (*m).edge[neweg2*eg_w+MCT1EGPCSU  ]=-1;
  (*m).edge[neweg2*eg_w+MCT1EGLVL   ]=level+1;
  /* make them childs of eg */
  (*m).edge[eg*eg_w+MCT1EGCHL1  ]=neweg1;
  (*m).edge[eg*eg_w+MCT1EGCHL1+1]=neweg2;
  
  /* eventually create new boundary elements, correct boundary
     information on the childs */
  if ((*m).edge[eg*eg_w+ MCT1EGBND]!= -1)
    {
      newbd1=(*m).bd_nr;
      newbd2=(*m).bd_nr+1;
      (*m).bd_nr+=2;
      oldbd=(*m).edge[eg*eg_w+ MCT1EGBND];
      /* newbd1 */
      (*m).bound[newbd1*bd_w+MCT1BDEDGE]=neweg1;
      (*m).bound[newbd1*bd_w+MCT1BDTYPE]=
	(*m).bound[oldbd*bd_w+MCT1BDTYPE]; /* same type as old */
      (*m).bound[newbd1*bd_w+MCT1BDFNCT]=
	(*m).bound[oldbd*bd_w+MCT1BDFNCT]; /* same function data as old
					   */
      (*m).bound[newbd1*bd_w+MCT1BDORIE]=
	(*m).bound[oldbd*bd_w+MCT1BDORIE]; /* first edge has same
					      orientation as the old
					      edge 
					   */
      (*m).bound[newbd1*bd_w+MCT1BDSSEG]=
	(*m).bound[oldbd*bd_w+MCT1BDSSEG]; /* same shape variability */
      (*m).edge[neweg1*eg_w+MCT1EGBND   ]=newbd1;
      /* newbd2 */
      (*m).bound[newbd2*bd_w+MCT1BDEDGE]=neweg2;
      (*m).bound[newbd2*bd_w+MCT1BDTYPE]=
	(*m).bound[oldbd*bd_w+MCT1BDTYPE]; /* same type as old */
      (*m).bound[newbd2*bd_w+MCT1BDFNCT]=
	(*m).bound[oldbd*bd_w+MCT1BDFNCT]; /* same function data as old
					   */
      (*m).bound[newbd2*bd_w+MCT1BDORIE]=
	-(*m).bound[oldbd*bd_w+MCT1BDORIE];/* 2nd edge has reverse
					      orientation of the old
					      edge */
      (*m).bound[newbd2*bd_w+MCT1BDSSEG]=
	(*m).bound[oldbd*bd_w+MCT1BDSSEG]; /* same shape variability */
      (*m).edge[neweg2*eg_w+MCT1EGBND   ]=newbd2;
    }

  /* eventually create new pcsurf elements, correct pcsurf
     information on the childs */
  if ((*m).edge[eg*eg_w+ MCT1EGPCSU]!= -1)
    {
      FIDX oldps, newps1, newps2;
      newps1=(*m).ps_nr;
      newps2=(*m).ps_nr+1;
      (*m).ps_nr+=2;
      oldps=(*m).edge[eg*eg_w+ MCT1EGPCSU];
      /* newps1 */
      (*m).pcsurf[newps1*ps_w+MCXXPSSURF]=neweg1;
      (*m).pcsurf[newps1*ps_w+MCXXPSVOLE]=-1;/* faces are set in face
						refinement */
      (*m).pcsurf[newps1*ps_w+MCXXPSCRIT]=
	(*m).pcsurf[oldps*ps_w+MCXXPSCRIT]; /* same criterion as old */
      (*m).pcsurf[newps1*ps_w+MCXXPSORIE]=
	(*m).pcsurf[oldps*ps_w+MCXXPSORIE]; /* first edge has same
					       orientation as the old
					       edge 
					   */
      (*m).edge[neweg1*eg_w+MCT1EGPCSU  ]=newps1;
      /* newps2 */
      (*m).pcsurf[newps2*ps_w+MCXXPSSURF]=neweg2;
      (*m).pcsurf[newps2*ps_w+MCXXPSVOLE]=-1;/* faces are set in face
						refinement */
      (*m).pcsurf[newps2*ps_w+MCXXPSCRIT]=
	(*m).pcsurf[oldps*ps_w+MCXXPSCRIT]; /* same criterion as old */
      (*m).pcsurf[newps2*ps_w+MCXXPSORIE]=
	-(*m).pcsurf[oldps*ps_w+MCXXPSORIE];/* 2nd edge has reverse
					       orientation of the old
					       edge 
					   */
      (*m).edge[neweg2*eg_w+MCT1EGPCSU  ]=newps2;
    }

  /* new hierarchy entry */
  newhi = (*m).hi_nr;
  (*m).hi_nr++;
  (*m).hier[newhi*hi_w+MCT1HICHLD  ] = newnode;
  (*m).hier[newhi*hi_w+MCT1HIFAT1  ] = nod1;
  (*m).hier[newhi*hi_w+MCT1HIFAT1+1] = nod2;
  (*m).hier[newhi*hi_w+MCT1HILVL   ] = level+1;

  return SUCCESS;
}


/*FUNCTION*/
int mesh_write_solution_exp_t1( struct mesh *m, struct vector *x,
				FIDX ncompo,
				FIDX namlen, char *name
/* writes the mesh and the solution into a NAG IRIS EXPLORER readable
   file (for visualisation) 
   
   Input:  m         - the mesh
           x         - vector with data associated to the nodes
	   ncompo    - dimension of the solution vector per node which
	               shall be writen to the file, (or number of
	               solution components), the components of the
	               vector x have to be numbered such that
	               x[j*vx_nr+i] gives the j-th component in the
	               i-th node of the mesh
	   namlen    - maximal useable length of name
	   name      - basename of the files, one will be
	               <name>.pyr and the other <name>.lat

   Output: (writes the files)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  FIDX i, j, lastslash, el_w, vx_w, vx_nr;

  char *pyrname, *latname, *latbasename;
  FILE *pyrfile, *latfile;


  el_w =(*m).el_w;
  vx_w =(*m).vx_w;
  vx_nr=(*m).vx_nr;

  if (ncompo*vx_nr!=(*x).len)
    {
      fprintf( stderr,
	       "\nmesh_write_solution_exp_t1: nodenr in mesh and "
	       "dimension of vector must agree!\n");
      return FAIL;
    }

  /* prepare the names */
  TRY_MALLOC( pyrname, (namlen+4), char, 
	      mesh_write_solution_exp_t1 );
  TRY_MALLOC( latname, (namlen+4), char, 
	      mesh_write_solution_exp_t1 );
  TRY_MALLOC( latbasename, (namlen+4), char, 
	      mesh_write_solution_exp_t1 );

  strncpy( pyrname, name, namlen );
  strncpy( latname, name, namlen );
  strncpy( latbasename, name, namlen );

  /* make sure no overread happens */
  pyrname[namlen-1]='\0';
  latname[namlen-1]='\0';
  latbasename[namlen-1]='\0';

  /* cut the leading path out of latname */
  lastslash=-1;
  for (i=0; i<namlen; i++)
    if (latbasename[i]=='/') lastslash=i;
  if (lastslash>=0)
    {
      lastslash++;
      for (i=0; i<namlen-lastslash; i++)
	{
	  latbasename[i]=latbasename[i+lastslash];
	}
    }
  
  /* append ".pyr" resp. ".lat" */
  strcat( pyrname, ".pyr" );
  strcat( latname, ".lat" );
  strcat( latbasename, ".lat" );
  
  /* open the lat file */
  latfile=fopen(latname, "w");
  if (latfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_solution_exp_t1: error opening file %s\n",
	      latname);
      return FAIL;
    }

  /* write the header */
  fprintf(latfile, "#!/usr/explorer/bin/explorer cxLattice plain 1.0\n\n");
  fprintf(latfile, "# ndim\n%6d\n", 1); /* dimension of the latice */
  fprintf(latfile, "# n points\n%6d\n", (int) vx_nr); /* nr of nodes */
  fprintf(latfile, "# ndatavar\n%6d\n", (int) ncompo);/* nr data per node */
  fprintf(latfile, "# primtype (float) \n%6d\n",3); /* data type float */
  fprintf(latfile, "# coordtype cx_coord_curvilinear\n%6d\n",2); 
  fprintf(latfile, "# nsteps\n%6d\n",1); /* nr of data sets */
  fprintf(latfile, "# coords\n%6d\n\n",2); /* dimension of the coords
					      */
  /* write the node positions */
  fprintf(latfile, "# coords of the nodes\n"); 
  for (i=0; i<vx_nr; i++)
    {
      fprintf(latfile, "  %18.10e  %18.10e\n",
	      (*m).vertex[i*vx_w+MCT1VXSTRT  ],
	      (*m).vertex[i*vx_w+MCT1VXSTRT+1]);
    }

  /* write the data */
  fprintf(latfile, "\n\n# data (step= 1)\n");
  for (i=0; i<vx_nr; i++)
    {
      for (j=0; j<ncompo; j++)
	{
	  fprintf(latfile, "  %18.10e",(*x).V[j*vx_nr+i]);
	}
      fprintf(latfile, "\n");
    }
  
  fprintf(latfile, "\n\n# EOF\n");
  /* close lat file */
  fclose(latfile);


  /* open the pyr file */
  pyrfile=fopen(pyrname, "w");
  if (pyrfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_solution_exp_t1: error opening file %s\n",
	      pyrname);
      return FAIL;
    }

  /* write the header */
  fprintf(pyrfile, "#!/usr/explorer/bin/explorer cxPyramid plain 1.0\n\n");
  fprintf(pyrfile, "include %s\n\n", latbasename);
  fprintf(pyrfile, "# n layers\n%6d\n",2); /* n layers in the file */
  fprintf(pyrfile, "# compression type cx_compress_unique\n%6d\n",1);
  fprintf(pyrfile, "# compression dict cx_pyramid_dict_triangle\n%6d\n\n",2);
  fprintf(pyrfile, "# layer 1 ?????\n%6d  %6d\n\n",0,0);
  fprintf(pyrfile, "# layer 2 n elements, n connections\n%6d  %6d\n\n",
	  (int) (*m).el_nr, (int) (*m).el_nr*3 );

  /* element pointers */
  fprintf(pyrfile, "# element pointers\n");
  for (i=0; i<(*m).el_nr; i++)
    {
      fprintf(pyrfile, "%6d\n", (int) (3*i) );      
    }
  fprintf(pyrfile, "# end pointer\n%6d\n\n", (int) (3*(*m).el_nr) );

  /* connectivity */
  fprintf(pyrfile, "# connectivity\n");
  for (i=0; i<(*m).el_nr; i++)
    {
      fprintf(pyrfile, "%6d %6d %6d\n",
	      (int) (*m).elem[i*el_w+MCT1ELNOD1  ],
	      (int) (*m).elem[i*el_w+MCT1ELNOD1+1],
	      (int) (*m).elem[i*el_w+MCT1ELNOD1+2]);      
    }
  fprintf(pyrfile, "\n# no data lattice at this layer\nNULL\n");
  fprintf(pyrfile, "\n# EOF\n");

  /* close pyr file */
  fclose(pyrfile); 

  free(pyrname);
  free(latname);
  free(latbasename);

  return SUCCESS;
}

/*FUNCTION*/
int mesh_write_solution_femplot_t1( struct mesh *m, struct vector *x,
				FIDX ncompo,
				FIDX namlen, char *name
/* writes the mesh and the solution into a ?FEMplot? readable
   file for visualisation
   
   Input:  m         - the mesh
           x         - vector with data associated to the nodes
	   ncompo    - dimension of the solution vector per node which
	               shall be writen to the file, (or number of
	               solution components), the components of the
	               vector x have to be numbered such that
	               x[j*vx_nr+i] gives the j-th component in the
	               i-th node of the mesh
	   namlen    - maximal useable length of name
	   name      - basename of the file, full name will be
	               <name>.femp

   Output: (writes the file)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  FIDX i, j, lastslash, vx_w, vx_nr, el_w, el_nr;

  char *fempname;
  FILE *fempfile;


  vx_w =(*m).vx_w;
  vx_nr=(*m).vx_nr;
  el_w =(*m).el_w;
  el_nr=(*m).el_nr;

  if (ncompo*vx_nr!=(*x).len)
    {
      fprintf( stderr,
	       "\nmesh_write_solution_femplot_t1: nodenr in mesh and "
	       "dimension of vector must agree!\n");
      return FAIL;
    }

  /* prepare the names */
  TRY_MALLOC( fempname, (namlen+5), char, 
	      mesh_write_solution_femplot_t1 );

  strncpy( fempname, name, namlen );

  /* make sure no overread happens */
  fempname[namlen-1]='\0';

  /* append ".pyr" resp. ".lat" */
  strcat( fempname, ".femp" );
  
  /* open the lat file */
  fempfile=fopen(fempname, "w");
  if (fempfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_solution_femplot_t1: error opening file %s\n",
	      fempname);
      return FAIL;
    }

  /* write the header */
  fprintf(fempfile, "%8d # Nodes \n", (int) vx_nr); /* no of nodes */
  fprintf(fempfile, "%8d # Triangles \n", (int) el_nr); /* nr of elements */
  fprintf(fempfile, "%8d # components \n", (int) ncompo);/* nr data per node */

  /* write the node positions */
  fprintf(fempfile, "// Nodes\n");
  for (i=0; i<vx_nr; i++)
    {
      fprintf(fempfile, "   %18.10e  %18.10e\n",
	      (*m).vertex[i*vx_w+MCT1VXSTRT  ],
	      (*m).vertex[i*vx_w+MCT1VXSTRT+1]);
    }


  /* write the triangles */
  fprintf(fempfile, "// Triangles\n");
  for (i=0; i<(*m).el_nr; i++)
    {
      fprintf(fempfile, "   %8d %8d %8d\n",
	      (int) (*m).elem[i*el_w+MCT1ELNOD1  ],
	      (int) (*m).elem[i*el_w+MCT1ELNOD1+1],
	      (int) (*m).elem[i*el_w+MCT1ELNOD1+2]);      
    }

  /* write the data */
  fprintf(fempfile, "// data\n");
  for (i=0; i<vx_nr; i++)
    {
      for (j=0; j<ncompo; j++)
	{
	  fprintf(fempfile, "  %18.10e",(*x).V[j*vx_nr+i]);
	}
      fprintf(fempfile, "\n");
    }
  
  /* close fempfile */
  fclose(fempfile);

  free(fempname);

  return SUCCESS;
}


/*FUNCTION*/
int mesh_write_bound_gnuplot_tx( struct mesh *m, 
				FIDX namlen, char *name,
				int type
/* writes the mesh into a GNUPLOT readable file (for visualisation),
   this is done by writing a line for each edge of the finest level
   
   Input:  m         - the mesh
	   namlen    - maximal useable length of name
	   name      - basename of the files, full name will be
	               <name>.dat
           type      - type of the mesh, ==1 ==> T1, ==2 ==> T2
                       in any case only straigth lines are written,
                       ignoring the possibility of curved edges for T2
                       meshes

   Output: (writes the files)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  FIDX eg, bd;
  FIDX bd_w, bd_nr, eg_w, eg_nr, vx_w, vx_nr;
  FIDX l_mcBDEDGE, l_mcEGCHL1, l_mcEGNOD1, l_mcVXSTRT;

  char *datname;
  FILE *datfile;


  vx_w =(*m).vx_w;
  eg_w =(*m).eg_w;
  bd_w =(*m).bd_w;

  vx_nr=(*m).vx_nr;
  eg_nr=(*m).eg_nr;
  bd_nr=(*m).bd_nr;

  switch (type)
    {
    case 1: /* T1 mesh */
      l_mcVXSTRT = MCT1VXSTRT;
      l_mcEGCHL1 = MCT1EGCHL1;
      l_mcEGNOD1 = MCT1EGNOD1;
      l_mcBDEDGE = MCT1BDEDGE;
      break;
    case 2: /* T2 mesh */
      l_mcVXSTRT = MCT2VXSTRT;
      l_mcEGCHL1 = MCT2EGCHL1;
      l_mcEGNOD1 = MCT2EGNOD1; 
      l_mcBDEDGE = MCT2BDEDGE;
      break;
    default:
      fprintf(stderr,"mesh_write_bound_gnuplot_tx: "
	      "unknown type=%d\n",type);
      return FAIL;
    }

  /* prepare the names */
  TRY_MALLOC( datname, (namlen+4), char, 
	      mesh_write_bound_gnuplot_tx );

  strncpy( datname, name, namlen );

  /* make sure no overread happens */
  datname[namlen-1]='\0';

  /* append ".dat"  */
  strcat( datname, ".dat" );
  
  /* open the lat file */
  datfile=fopen(datname, "w");
  if (datfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_bound_gnuplot_tx: error opening file %s\n",
	      datname);
      return FAIL;
    }
  /* write the egdes */
  for (bd=0; bd<bd_nr; bd++)
    {
      eg=(*m).bound[bd*bd_w+l_mcBDEDGE];
      if ((*m).edge[eg*eg_w+l_mcEGCHL1]==-1) 
	{
	  /* only write edges which have not been refined further */
	  FIDX n1,n2;
	  n1=(*m).edge[eg*eg_w+l_mcEGNOD1  ];
	  n2=(*m).edge[eg*eg_w+l_mcEGNOD1+1];

	  /* write first point\n, second point\n, blank line */
	  fprintf(datfile,
		  "%18.10e  %18.10e\n"
		  "%18.10e  %18.10e\n"
		  "\n",
		  (*m).vertex[n1*vx_w+l_mcVXSTRT  ],
		  (*m).vertex[n1*vx_w+l_mcVXSTRT+1],
		  (*m).vertex[n2*vx_w+l_mcVXSTRT  ],
		  (*m).vertex[n2*vx_w+l_mcVXSTRT+1]);
	}
    }


  /* close dat file */
  fclose(datfile); 

  free(datname);

  return SUCCESS;
}



/*FUNCTION*/
int mesh_write_mesh_gnuplot_tx( struct mesh *m, 
				FIDX namlen, char *name,
				int type
/* writes the mesh into a GNUPLOT readable file (for visualisation),
   this is done by writing a line for each edge of the finest level
   
   Input:  m         - the mesh
	   namlen    - maximal useable length of name
	   name      - basename of the file, full name will be
	               <name>.dat
           type      - type of the mesh, ==1 ==> T1, ==2 ==> T2
                       in any case only straigth lines are written,
                       ignoring the possibility of curved edges for T2
                       meshes

   Output: (writes the files)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  FIDX eg, eg_w, eg_nr, vx_w, vx_nr;
  FIDX l_mcEGCHL1, l_mcEGNOD1, l_mcVXSTRT;

  char *datname;
  FILE *datfile;


  vx_w =(*m).vx_w;
  eg_w =(*m).eg_w;

  vx_nr=(*m).vx_nr;
  eg_nr=(*m).eg_nr;

  switch (type)
    {
    case 1: /* T1 mesh */
      l_mcVXSTRT = MCT1VXSTRT;
      l_mcEGCHL1 = MCT1EGCHL1;
      l_mcEGNOD1 = MCT1EGNOD1;
      break;
    case 2: /* T2 mesh */
      l_mcVXSTRT = MCT2VXSTRT;
      l_mcEGCHL1 = MCT2EGCHL1;
      l_mcEGNOD1 = MCT2EGNOD1;
      break;
    default:
      fprintf(stderr,"mesh_write_mesh_gnuplot_tx: "
	      "unknown type=%d\n",type);
      return FAIL;
    }

  /* prepare the names */
  TRY_MALLOC( datname, (namlen+4), char, 
	      mesh_write_mesh_gnuplot_tx );

  strncpy( datname, name, namlen );

  /* make sure no overread happens */
  datname[namlen-1]='\0';

  /* append ".dat"  */
  strcat( datname, ".dat" );
  
  /* open the lat file */
  datfile=fopen(datname, "w");
  if (datfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_mesh_gnuplot_tx: error opening file %s\n",
	      datname);
      return FAIL;
    }
  /* write the egdes */
  for (eg=0; eg<eg_nr; eg++)
    if ((*m).edge[eg*eg_w+l_mcEGCHL1]==-1) 
      {
	/* only write edges which have not been refined further */
	FIDX n1,n2;
	n1=(*m).edge[eg*eg_w+l_mcEGNOD1  ];
	n2=(*m).edge[eg*eg_w+l_mcEGNOD1+1];

	/* write first point\n, second point\n, blank line */
	fprintf(datfile,
		"%18.10e  %18.10e\n"
		"%18.10e  %18.10e\n"
		"\n",
		(*m).vertex[n1*vx_w+l_mcVXSTRT  ],
		(*m).vertex[n1*vx_w+l_mcVXSTRT+1],
		(*m).vertex[n2*vx_w+l_mcVXSTRT  ],
		(*m).vertex[n2*vx_w+l_mcVXSTRT+1]);
      }


  /* close dat file */
  fclose(datfile); 

  free(datname);

  return SUCCESS;
}



/*FUNCTION*/
int mesh_write_ssegs_svg_tx( struct mesh *m, 
			     FIDX namlen, char *name,
			     int type
/* writes the shape-segments into a scalable vector graphics (SVG)
   file, which can for example be read by sodipodi or inkscape
   
   Input:  m         - the mesh
	   namlen    - maximal useable length of name
	   name      - basename of the files, full name will be
	               <name>.svg
           type      - type of the mesh, ==1 ==> T1, ==2 ==> T2

   Output: (writes the file)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  int  err;
  FIDX i;
  FIDX eg, seg, paths;
  FIDX bd_w, bd_nr, eg_w, eg_nr, vx_w, vx_nr, sg_w, sg_nr, sp_w;
  FIDX l_mcBDSSEG, l_mcBDEDGE, l_mcEGCHL1, l_mcEGNOD1, l_mcVXSTRT;

  struct ilist **eglist; /* list of edges for each segment */
  FIDX *nd_nbs;   /* nd_nbs[i*2+j]=k,  j=1,2:  node i has neigbour k
		     within the current segment */
  FIDX *seg_nds;  /* first and last node for each segment:
                     seg_nds[seg*2+0]=first, seg_nds[seg*2+1]=last */
  FIDX *nd_segs;  /* for each node the segments which contain it as
		     first or last node k=nd_segs[node*2+i], i=0,1,
		     if k==-1 => no segment, else seg=k, node is
		     contained in segment seg */
  int  *seg_done; /* marks if an segment is already contained in a
		     path */

  char *datname, *datname_mrk;
  FILE *datfile, *datfile_mrk;


  vx_w =(*m).vx_w;
  eg_w =(*m).eg_w;
  bd_w =(*m).bd_w;
  sg_w =(*m).sg_w;
  sp_w =(*m).sp_w;

  vx_nr=(*m).vx_nr;
  eg_nr=(*m).eg_nr;
  bd_nr=(*m).bd_nr;
  sg_nr=(*m).sg_nr;

  switch (type)
    {
    case 1: /* T1 mesh */
      l_mcVXSTRT = MCT1VXSTRT;
      l_mcEGCHL1 = MCT1EGCHL1;
      l_mcEGNOD1 = MCT1EGNOD1;
      l_mcBDEDGE = MCT1BDEDGE;
      l_mcBDSSEG = MCT1BDSSEG;
      break;
    case 2: /* T2 mesh */
      l_mcVXSTRT = MCT2VXSTRT;
      l_mcEGCHL1 = MCT2EGCHL1;
      l_mcEGNOD1 = MCT2EGNOD1; 
      l_mcBDEDGE = MCT2BDEDGE;
      l_mcBDSSEG = MCT2BDSSEG;
      break;
    default:
      fprintf(stderr,"mesh_write_ssegs_svg_tx: "
	      "unknown type=%d\n",type);
      return FAIL;
    }

  /*************************************/
  /*                                   */
  /* stage 1: sort out segments        */
  /*                                   */
  /*************************************/
  
  TRY_MALLOC( eglist, sg_nr, struct ilist*, mesh_write_ssegs_svg_tx);
  /* init the list */
  for (i=0; i<sg_nr; i++)
    eglist[i]=NULL;
  
  /* build a list of edges for each segment */
  for (i=0; i<(*m).bd_nr; i++)
    {
      seg = m->bound[i*bd_w+l_mcBDSSEG];
      eg  = m->bound[i*bd_w+l_mcBDEDGE];

      if ((seg!=-1)&&(m->edge[eg*eg_w+l_mcEGCHL1]==-1))
	{
	  ilist_sorted_insert( &eglist[seg], eg);
	}
    }

  TRY_MALLOC( nd_nbs, 2*vx_nr, FIDX, mesh_write_ssegs_svg_tx);
  TRY_MALLOC( nd_segs, 2*vx_nr, FIDX, mesh_write_ssegs_svg_tx);
  TRY_MALLOC( seg_nds, 2*sg_nr, FIDX, mesh_write_ssegs_svg_tx);
  TRY_MALLOC( seg_done, sg_nr, int, mesh_write_ssegs_svg_tx);

  /* clear nd_segs */
  for (i=0; i<2*vx_nr; i++)
    nd_segs[i]=-1;

  for (seg=0; seg<sg_nr; seg++)
    {
      struct ilist *this_;
      FIDX tnode, pnode;      

      /**********************************************************/
      /* for each node on the edges mark its neighbouring nodes */
      /**********************************************************/
      this_=eglist[seg];

      /* clear nd_nbs */
      for (i=0; i<2*vx_nr; i++)
	nd_nbs[i]=-1;

      while (this_!=NULL)
	{
	  FIDX eg=this_->data;
	  
	  FIDX n1= m->edge[eg*eg_w+l_mcEGNOD1  ];
	  FIDX n2= m->edge[eg*eg_w+l_mcEGNOD1+1];

	  if (nd_nbs[n1*2]==-1)
	    nd_nbs[n1*2]=n2;
	  else if (nd_nbs[n1*2+1]==-1)
	    nd_nbs[n1*2+1]=n2;
	  else 
	    {
	      fprintf(stderr,"mesh_write_ssegs_svg_tx: "
		      "node multiplicity error\n");
	      return FAIL;
	    }

	  if (nd_nbs[n2*2]==-1)
	    nd_nbs[n2*2]=n1;
	  else if (nd_nbs[n2*2+1]==-1)
	    nd_nbs[n2*2+1]=n1;
	  else 
	    {
	      fprintf(stderr,"mesh_write_ssegs_svg_tx: "
		      "node multiplicity error\n");
	      return FAIL;
	    }
	  this_=this_->next;
	}

      /**********************************************************/
      /* find first and last node                               */
      /**********************************************************/
      pnode=-2;                           /* previous node */
      tnode=m->sseg[seg*sg_w+MC2XSGNOD1]; /* this node     */
      seg_nds[seg*2+0] = tnode;           /* mark the first node */
      while (tnode>=0)
	{
	  FIDX nb;

	  nb=nd_nbs[tnode*2  ];
	  /* if this neighbour is the previous node, use other
	     neighbour */
	  if (nb==pnode)
	    nb=nd_nbs[tnode*2+1];
	  
	  /* continue with the neighbour */
	  pnode=tnode;
	  tnode=nb;
	}
      tnode = pnode;
      seg_nds[seg*2+1] = tnode;           /* mark the last node */

      /* mark last and first node in nd_segs */
      /* last node: */
      if (nd_segs[tnode*2+0]==-1)
	nd_segs[tnode*2+0]=seg;
      else if (nd_segs[tnode*2+1]==-1)
	nd_segs[tnode*2+1]=seg;
      else 
	{
	  fprintf(stderr, "mesh_write_ssegs_svg_tx: "
		  "one node in >2 segments!\n");
	  return FAIL;
	}
      /* first node: */
      tnode = seg_nds[seg*2+0];
      if (nd_segs[tnode*2+0]==-1)
	nd_segs[tnode*2+0]=seg;
      else if (nd_segs[tnode*2+1]==-1)
	nd_segs[tnode*2+1]=seg;
      else 
	{
	  fprintf(stderr, "mesh_write_ssegs_svg_tx: "
		  "one node in >2 segments!\n");
	  return FAIL;
	}

      /* the edge-list is not needed anymore, free it */
      ilist_free(&eglist[seg]);
      eglist[seg]=NULL;
      /* next segment */
    }/* end loop segments */

  /*************************************/
  /*                                   */
  /* stage 2: find paths and write     */
  /*                                   */
  /*************************************/

  /* now we have begin and end nodes for all segments and the segments
     for each of these nodes, 

     we need to collect them into paths now, which are then written
     into the file 

     the paths are build by starting from one segment, going into one
     direction until the end is found or a loop detected, in either
     case the path is written to the file starting from the last
     segment of the first search, since this is an end of the path in
     any case. segments which are dealt with already are marked in
     seg_done
  */

  /* prepare the names */
  TRY_MALLOC( datname, (namlen+4), char, 
	      mesh_write_ssegs_svg_tx );
  strncpy( datname, name, namlen );
  /* make sure no overread happens */
  datname[namlen-1]='\0';
  /* append ".svg"  */
  strcat( datname, ".svg" );
  /* open the svg file */
  datfile=fopen(datname, "w");
  if (datfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_ssegs_svg_tx: error opening file %s\n",
	      datname);
      return FAIL;
    }

  /* prepare the names */
  TRY_MALLOC( datname_mrk, (namlen+4+4), char, 
	      mesh_write_ssegs_svg_tx );
  strncpy( datname_mrk, name, namlen );
  /* make sure no overread happens */
  datname_mrk[namlen-1]='\0';
  /* append "_mrk.svg"  */
  strcat( datname_mrk, "_mrk.svg" );
  /* open the marker file */
  datfile_mrk=fopen(datname_mrk, "w");
  if (datfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_ssegs_svg_tx: error opening file %s\n",
	      datname_mrk);
      return FAIL;
    }

  /* SVG file header, reverse-engineered from SODIPODI output */
  fprintf(datfile, 
	  "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n"
	  "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20010904//EN\"\n"
	  "\"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n");
  fprintf(datfile, 
	  "<svg\n"
	  "   xmlns=\"http://www.w3.org/2000/svg\"\n"
	  "   xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n"
	  "   version=\"1.0\"\n");
  fprintf(datfile, 
	  "   x=\"0\"\n"
	  "   y=\"0\"\n"
	  "   width=\"200.0\"\n"
	  "   height=\"200.0\"\n"
	  "   id=\"svg611\"\n"
	  "   xml:space=\"preserve\"><defs\n"
	  "   id=\"defs604\" />\n");


  /* mark all segments as not done yet */
  for (i=0; i<sg_nr; i++)
    seg_done[i]=0;

  seg=0;
  paths=0;
  while (seg<sg_nr)
    {
      FIDX fnode, tnode, nnode, tseg;
      int found;
      FIDX par0, par1;

      if (seg_done[seg]==0)
	{
	  /* this segment is not part of an path yet, try to find a
	     path */
	  fnode = seg_nds[seg*2+0]; /* first node in this search */
	  tnode = seg_nds[seg*2+1]; /* current end node in this search */
	  tseg  = seg;              /* current end segment in this search */
	  found = 1;
	  while (found==1) 
	    {
	      /* try to find next node on the path */
	      FIDX nseg,nnode;

	      nseg = nd_segs[tnode*2+0];
	      /* if the attempted next segment is the current one, use
		 the other one */
	      if (nseg==tseg) 
		nseg = nd_segs[tnode*2+1];

	      /* see if we have a next segment, otherwise we are done */
	      if (nseg != -1)
		{
		  nnode=seg_nds[nseg*2+0];
		  if (nnode==tnode) 
		    nnode=seg_nds[nseg*2+1];

		  /* we have found the next segment and next node, 
		     if this next node is the first, we have a loop
		     and can stop looking, otherwise we have to try
		     again */
		  tnode = nnode;
		  tseg  = nseg;
		  
		  if (tnode==fnode)
		    found = 0; /* loop, stop looking for end of path */
		  else
		    found = 1; /* continue looking for end of path */
		  
		  if (found==0) printf("loop\n");
		}
	      else
		{
		  found = 0;
		  printf("line\n");
		}
	    }

	  /* now tseg and tnode contain the end segment and end node
	     of a path, build the path from this info */
	  fnode = tnode;
	  nnode=seg_nds[tseg*2+0];
	  if (nnode==tnode) 
	    nnode=seg_nds[tseg*2+1];

	  /* write the segment to the file */
	  fprintf(datfile, "    <path\n    d=\"");

	  /* first point */
	  if (tnode==m->sseg[tseg*sg_w+MC2XSGNOD1])
	    { 
	      /* correct order of parameters */
	      par0=m->sseg[tseg*sg_w+MC2XSGPAR1+0];
	      par1=m->sseg[tseg*sg_w+MC2XSGPAR1+1];
	    }
	  else
	    { 
	      /* reverse order of parameters */
	      par0=m->sseg[tseg*sg_w+MC2XSGPAR1+2];
	      par1=m->sseg[tseg*sg_w+MC2XSGPAR1+3];
	    }
	  fprintf(datfile, "M %12.6f %12.6f ", 
		  100* m->spar[par0*sp_w], 100* m->spar[par1*sp_w]);

	  err = mesh_write_sseg_svg_file_tx(datfile, datfile_mrk,  m,
					    tseg, tnode);
	  FUNCTION_FAILURE_HANDLE(err, mesh_write_sseg_svg_file_tx,
				  mesh_write_ssegs_svg_tx);
	  seg_done[tseg]=1;

	  tnode = nnode;
	  found = 1; 
	  while (found==1) 
	    {
	      /* try to find next node on the path */
	      FIDX nseg,nnode;

	      nseg = nd_segs[tnode*2+0];
	      /* if the attempted next segment is the current one, use
		 the other one */
	      if (nseg==tseg) 
		nseg = nd_segs[tnode*2+1];

	      /* see if we have a next segment, otherwise we are done */
	      if (nseg != -1)
		{
		  nnode=seg_nds[nseg*2+0];
		  if (nnode==tnode) 
		    nnode=seg_nds[nseg*2+1];

		  /* we have found the next segment and next node, 
		     write the next segment to the file 
		     if this next node is the first, we have a loop
		     and can stop looking, otherwise we have to try
		     again */
		  err = mesh_write_sseg_svg_file_tx(datfile, datfile_mrk,
						    m, nseg, tnode);
		  FUNCTION_FAILURE_HANDLE(err, mesh_write_sseg_svg_file_tx,
					  mesh_write_ssegs_svg_tx);
		  seg_done[nseg]=1;

		  tnode = nnode;
		  tseg  = nseg;
		  
		  if (tnode==fnode)
		    found = 0; /* loop, stop looking for end of path */
		  else
		    found = 1; /* continue looking for end of path */
		}
	      else
		found = 0;
	    } /* end while found==1 */

	  /* path complete, write final properties */
	  fprintf(datfile, "\"\n"
		  "    style=\"font-size:12;fill:none;fill-rule:evenodd;"
		  "stroke:#000000;stroke-width:0.3;\"\n"
		  "    id=\"path%d\" />\n", (int) paths);
	  paths++;

	}
      /* now this segment has definitely been asigned to a path, try
	 the next one */
      seg++;
    }

  /* all paths written, write tail of the file, close it */
  fprintf(datfile, "</svg>\n");
  fclose(datfile); 
  fclose(datfile_mrk); 

  /* free local data */
  free(eglist);
  free(nd_nbs);
  free(nd_segs);
  free(seg_nds);
  free(seg_done);

  free(datname);
  free(datname_mrk);

  return SUCCESS;
}


/*FUNCTION*/
int mesh_write_sseg_svg_file_tx(FILE *datfile, FILE *datfile_mrk,
				struct mesh *m, FIDX seg, FIDX node1
/* writes the segement to the files specified by the file pointers,
   depending on the type and orientation of the segment

   input:   datfile - pointer to path file, it is written to the file in
                      this routine
	    datfile_mrk
                    - pointer to marker file, parameter markers are
                      written to the file in this routine
	    m       - a t1 or t2 mesh
            seg     - the segment to be written into the file
            node1   - first node of the segment, the orientation of
                      the segment is adjusted such that it really has
                      this node as first node

   output:  (writes to the file)

   return:  SUCCESS - success,
            FAIL    - failure, see error message 
*/
				){
  FIDX sg_w, sg_nr, sp_w;
  FIDX type;
  int orie, i;
  double lpar[8];
  double lpoint[8];
  char line_mark_color[]="#ff00ff";
  char dir_line_color[]="#0000ff";
  char dir_mark_color[]="#ff0000";
	
  sg_w =(*m).sg_w;
  sp_w =(*m).sp_w;

  sg_nr=(*m).sg_nr;

  if (seg >= sg_nr)
    {
      fprintf(stderr,"mesh_write_sseg_svg_file_tx: "
	      "seg >= sg_nr\n");
      return FAIL;
    }

  /* decide about the orientation */
  if (node1==m->sseg[seg*sg_w+MC2XSGNOD1])
    orie=1;
  else
    orie=-1;

  type = m->sseg[seg*sg_w+MC2XSGTYPE];

  switch (type)
    {
    case 1: /* line */
      for (i=0; i<4; i++)
	lpar[i]= m->spar[m->sseg[seg*sg_w+MC2XSGPAR1+i]*sp_w];

      if (orie==1)
	fprintf(datfile, "L %12.6f %12.6f ", 100*lpar[2], 100*lpar[3]);
      else
	fprintf(datfile, "L %12.6f %12.6f ", 100*lpar[0], 100*lpar[1]);

      /* independent of the orientation, both points are written as
	 marker (rectangle) to the marker file */
      fprintf(datfile_mrk, "<rect "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegrect%d\"\n"
	      "width=\"2.0\"\n"
	      "height=\"2.0\"\n"
	      "x=\"%12.6f\"\n"
	      "y=\"%12.6f\" />\n",
	      line_mark_color, (int)1+4*seg+0,
	      100*lpar[0]-1.0, -100*lpar[1]-1.0);
      fprintf(datfile_mrk, "<rect "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegrect%d\"\n"
	      "width=\"2.0\"\n"
	      "height=\"2.0\"\n"
	      "x=\"%12.6f\"\n"
	      "y=\"%12.6f\" />\n",
	      line_mark_color, (int)1+4*seg+1,
	      100*lpar[2]-1.0, -100*lpar[3]-1.0);
      break;
    case 2: /* b-spline */ 
      for (i=0; i<8; i++)
	lpar[i]= m->spar[m->sseg[seg*sg_w+MC2XSGPAR1+i]*sp_w];
	
      /* the sxi and syi need to be adjusted for easy evaluation */
      /* the first two direction need just adding the point, to give
	 control points */
      lpoint[4]=lpar[0]+lpar[4];
      lpoint[5]=lpar[1]+lpar[5];
      /* the later two require turning around and adding the point */
      lpoint[6]=lpar[2]-lpar[6];
      lpoint[7]=lpar[3]-lpar[7];
      /* the reversed direction points */
      lpoint[0]=lpar[0]-lpar[4];
      lpoint[1]=lpar[1]-lpar[5];
      lpoint[2]=lpar[2]+lpar[6];
      lpoint[3]=lpar[3]+lpar[7];

      
      if (orie==1)
	fprintf(datfile, "C %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f ",
		100*lpoint[4], 100*lpoint[5], 100*lpoint[6], 100*lpoint[7], 
		100*lpar[2], 100*lpar[3]);
      else
	fprintf(datfile, "C %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f ",
		100*lpoint[6], 100*lpoint[7], 100*lpoint[4], 100*lpoint[5],
		100*lpar[0], 100*lpar[1]);

      /* independent of the orientation, both direction markers are
	 written as to the marker file (line, diamond, circle)*/
      /***** lines *****/
      fprintf(datfile_mrk, "<path "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegline%d\"\n"
	      "d=\"M %12.6f %12.6f L %12.6f %12.6f\"\n"
	      "/>\n",
	      dir_line_color, (int)1+4*seg+0,
	      100*lpoint[0],-100*lpoint[1],
	      100*lpoint[4],-100*lpoint[5]);
      fprintf(datfile_mrk, "<path "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegline%d\"\n"
	      "d=\"M %12.6f %12.6f L %12.6f %12.6f\"\n"
	      "/>\n",
	      dir_line_color, (int)1+4*seg+1,
	      100*lpoint[2],-100*lpoint[3],
	      100*lpoint[6],-100*lpoint[7]);
      
      /***** diamonds *****/
      fprintf(datfile_mrk, "<rect "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegrect%d\"\n"
	      "width=\"2.0\"\n"
	      "height=\"2.0\"\n"
	      "x=\"%12.6f\"\n"
	      "y=\"%12.6f\"\n"
	      "transform=\"matrix(0.707,0.707,-0.707,0.707,0,0)\"\n"
	      "/>\n",
	      dir_mark_color, (int)1+4*seg+0,
	      (0.707*100*lpar[0]-0.707*100*lpar[1])-1.0,
	      -(0.707*100*lpar[0]+0.707*100*lpar[1])-1.0 );
      fprintf(datfile_mrk, "<rect "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegrect%d\"\n"
	      "width=\"2.0\"\n"
	      "height=\"2.0\"\n"
	      "x=\"%12.6f\"\n"
	      "y=\"%12.6f\"\n"
	      "transform=\"matrix(0.707,0.707,-0.707,0.707,0,0)\"\n"
	      "/>\n",
	      dir_mark_color, (int)1+4*seg+1,
	      (0.707*100*lpar[2]-0.707*100*lpar[3])-1.0,
	      -(0.707*100*lpar[2]+0.707*100*lpar[3])-1.0 );

      /***** circles *****/
      fprintf(datfile_mrk, "<path "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegcirc%d\"\n"
	      "d=\"M %12.6f %12.6f A 1 1 0 1 0 %12.6f %12.6f "
	      "A 1 1 0 1 0  %12.6f %12.6f z\"\n"
	      "/>\n",
	      dir_mark_color, (int)1+4*seg+0,
	      100*lpoint[2]+1.0,-100*lpoint[3],
	      100*lpoint[2]-1.0,-100*lpoint[3],
	      100*lpoint[2]+1.0,-100*lpoint[3] );

      fprintf(datfile_mrk, "<path "
	      "style=\"fill:none;stroke:%s;stroke-width:0.1;\"\n"
	      "id=\"ssegcirc%d\"\n"
	      "d=\"M %12.6f %12.6f A 1 1 0 1 0 %12.6f %12.6f "
	      "A 1 1 0 1 0  %12.6f %12.6f z\"\n"
	      "/>\n",
	      dir_mark_color, (int)1+4*seg+1,
	      100*lpoint[4]+1.0,-100*lpoint[5],
	      100*lpoint[4]-1.0,-100*lpoint[5],
	      100*lpoint[4]+1.0,-100*lpoint[5] );
      break;
    default:
      fprintf(stderr,"mesh_write_sseg_svg_file_tx: "
	      "unknown segment type=%d\n", (int) type);
      return FAIL;
    }

  return SUCCESS;
}



/*FUNCTION*/
int mesh_read_file_t1( struct mesh *m, char *name
/* reads a mesh file, for a descrition of the file format, see the
   example file "mesh_example.f1m"

   input:   name    - string containing the file name of the file to
                      be read
   
   output:  m       - (given by reference) a t1 mesh, it is assumed to
                      be empty and is initialised in this routine

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/
		       ){
  int err;
  long secstart;
  FIDX format, dim, problem, meshgen, vx_nr, el_nr, bd_nr, holes,
    ps_nr, pv_nr, pc_nr, fn_nr, pa_nr, sg_nr, sp_nr; 
  char buff[1024];

  FILE *in;

  mesh_init(m);

  /* open the file */
  in = fopen( name, "r" );
  if (in == NULL)
    {
      fprintf(stderr, "mesh_read_file_t1: couldn't open file: %s\n",
	      name);
      return FAIL;
    }

  /* bin all the crap before <header> */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp(buff,"<header>")!=0));
  if (strcmp(buff,"<header>")!=0)
    {
      fprintf(stderr, "mesh_read_file_t1: file has no header: %s\n",
	      name);
      return FAIL;
    }

  /* save the begin of the header to secstart */
  secstart = ftell(in);

  /* get each of the header fields */
  err=mesh_read_header_key(in, "format", &format);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  if (format!=0)
    {
      fprintf(stderr, 
	      "mesh_read_file_t1: format of file unknown: %s\n",
	      name);
      return FAIL;
    } 

  err=mesh_read_header_key(in, "dim", &dim);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "problem", &problem);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "meshgen", &meshgen);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "vertex", &vx_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "elements", &el_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "boundary", &bd_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "holes", &holes);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "pcsurf", &ps_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "pcvol", &pv_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "pccrit", &pc_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "function", &fn_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);
  fseek(in, secstart, SEEK_SET);

  err=mesh_read_header_key(in, "parameter", &pa_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);

  err=mesh_read_header_key(in, "shape_seg", &sg_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);

  err=mesh_read_header_key(in, "shape_par", &sp_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_header_key,mesh_read_file_t1);

  /* jump to </header> */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp(buff,"</header>")!=0));
  if (strcmp(buff,"</header>")!=0)
    {
      fprintf(stderr, "mesh_read_file_t1: file corrupt header: %s\n",
	      name);
      return FAIL;
    }

  /* save the end of the header to secstart */
  secstart = ftell(in);

  /* header successfully read, now set the mesh */
  (*m).dim     = dim;
  (*m).problem = problem;

  /* set width of fields */
  (*m).vx_w = MCT1VXLN;
  (*m).el_w = MCT1ELLN;
  (*m).eg_w = MCT1EGLN;
  (*m).fc_w = MCT1FCLN;
  (*m).vo_w = MCT1VOLN;
  (*m).hi_w = MCT1HILN;
  (*m).bd_w = MCT1BDLN;
  (*m).ps_w = MCXXPSLN;
  (*m).pv_w = MCXXPVLN;
  (*m).pc_w = MCXXPCLN;
  (*m).fu_w = MC2XFULN;
  (*m).pa_w = MC2XPALN;
  (*m).sg_w = MC2XSGLN;
  (*m).sp_w = MC2XSPLN;

  err=mesh_read_vertex( in, m, vx_nr, MCT1VXSTRT);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_vertex,mesh_read_file_t1);

  err=mesh_read_elements_t1( in, m, el_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_elements_t1,mesh_read_file_t1);

  err=mesh_read_boundary_t1( in, m, bd_nr, sg_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_boundary_t1,mesh_read_file_t1);

  /* save the position to secstart */
  secstart = ftell(in);
  err=mesh_read_pccrit( in, m, pc_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_pccrit,mesh_read_file_t1);

  /* rewind to read the pcsurf part */
  fseek(in, secstart, SEEK_SET);
  err=mesh_read_pcsurf( in, m, ps_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_pcsurf,mesh_read_file_t1);

  err=mesh_read_pcvol( in, m, pv_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_pcsurf,mesh_read_file_t1);

  /* read the shape segments and shape parameter */
  if (sg_nr>0)
    {
      err=mesh_read_sseg( in, m, sg_nr, sp_nr);
      FUNCTION_FAILURE_HANDLE(err, mesh_read_sseg,mesh_read_file_t1);

      err=mesh_read_spar( in, m, sp_nr);
      FUNCTION_FAILURE_HANDLE(err, mesh_read_spar,mesh_read_file_t1);
    }

  /* rewind to read the other stuff */
  fseek(in, secstart, SEEK_SET);

  /* create an empty hierarchy part */
  (*m).hi_max=10;
  (*m).hi_nr =0;
  TRY_MALLOC( (*m).hier, (*m).hi_max*(*m).hi_w, FIDX, 
	      mesh_read_file_t1);


  /* call the appropriate mesh-generator-wrapper (reading the holes
     section is left to that)  */
  switch(meshgen)
    {
    case 0: /* no mesh generator */
      if (sg_nr>0)
	{
	  /* adjust the position of nodes on the parametric boundary */
	  /*                    msh, insert, nI, dIdx, dIdF, type */
	  err=mesh_sseg_adjust_tx( m, 0,     0, NULL, NULL, 1); 
	  FUNCTION_FAILURE_HANDLE(err, mesh_sseg_adjust_tx,mesh_read_file_t1);
	}	  
      break;
    case 1: /* triangle, call the wrapper */
      if (sg_nr>0)
	{
	  /* adjust the position of nodes on the parametric boundary,
	     refine curved parts to give better resolution if necessary */
	  /*                    msh, insert, nI, dIdx, dIdF, type */
	  err=mesh_sseg_adjust_tx( m, 1,      0, NULL, NULL, 1); 
	  FUNCTION_FAILURE_HANDLE(err, mesh_sseg_adjust_tx,mesh_read_file_t1);
	}	  
      if (pv_nr!=0)
	{
	  /* if the triangle mesh generator is used no pcvol entries
	     may be defined in the file, but for each volume criterion
	     in pccrit a pcvol entry is created for each element  */
	  fprintf(stderr, "mesh_read_file_t1: "
		  "pv_nr!=0 incompatible with meshgen=%d\n",
		  (int) meshgen);
	  return FAIL;
	}
      err=mesh_read_triangle_wrapper( in, m, holes);
      FUNCTION_FAILURE_HANDLE(err, mesh_read_triangle_wrapper,mesh_read_file_t1);
      break;
    default: /* unknown, cry */
      fprintf(stderr, 
	      "mesh_read_file_t1: mesh generator unknown: %d\n",
	      (int) meshgen);
      return FAIL;
    }

  err=mesh_read_compute_conns_t1(m);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_compute_conns_t1,mesh_read_file_t1);

	    
  err=mesh_read_function( in, m, fn_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_function,mesh_read_file_t1);

  err=mesh_read_parameter( in, m, pa_nr);
  FUNCTION_FAILURE_HANDLE(err, mesh_read_parameter,mesh_read_file_t1);

  fclose(in);

  return SUCCESS;
}

/*FUNCTION*/
int mesh_read_header_key( FILE *in, char *keywd, FIDX *value
/* reads the keyword part of the header of a mesh file, for a
   descrition of the file format, see the example file
   "mesh_example.f1m" 
   here a entry 
     <keywd N > 
   is looked up, and the integer number N is returned in value 

   input:   in      - a file open for reading, reading position at the
                      begin of the header part
	    keywd   - keyword which shall be looked up
   
   output:  value   - (given by reference) the integer number found
                      with for the keyword

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  int  ihlp;
  char buff[1024];
  char *xkey;

  TRY_MALLOC(xkey, (strlen(keywd)+3), char, mesh_read_header_key);
  
  strcpy(xkey, "<");
  strcat(xkey, keywd);

  /* jump to </header> */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp(buff,xkey)!=0));
  if (strcmp(buff, xkey)!=0)
    {
      fprintf(stderr, "mesh_read_header_key: key not found: %s\n",
	      xkey);
      return FAIL;
    }
  if ((fscanf(in, "%d %s", &ihlp, buff)!=2)||(strcmp(buff,">")!=0))
    {
      fprintf(stderr, 
	      "mesh_read_header_key: key found, but unable to read "
	      "value: %s\n", xkey);
      return FAIL;
    }
  *value = (FIDX) ihlp;

  free(xkey);

  return SUCCESS;
}

/*FUNCTION*/
int mesh_read_vertex( FILE *in, struct mesh *m, FIDX vx_nr, FIDX vxstrt
/* reads the vertex part of a mesh file, for a descrition of the file
   format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            vx_nr   - the number of vertices in the file, as given by
                      the header
	    vxstrt  - the position within the vertex blocks, where the
	              coordinates of the vertex start
   
   output:  m       - (given by reference) a mesh, the vertex part is
                      allocated and defined

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  FIDX i, j;
  int  ihlp;
  double dhlp;
  char buff[1024];

  TRY_MALLOC((*m).vertex, vx_nr * (*m).vx_w , double,
	     mesh_read_vertex);
  (*m).vx_max = vx_nr;

  /* find the begin of the vertex block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<vertex>",buff)!=0));
  if (strcmp("<vertex>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_vertex: file has no <vertex>\n");
      return FAIL;
    }

  /* read all the vertices one by one */
  for (i=0; i<vx_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||(ihlp!=i))
	{
	  fprintf(stderr, "mesh_read_vertex: order of entries wrong!\n");
	  return FAIL;
	}
      for (j=0; j<(*m).dim; j++)
	{
	  if (fscanf(in, "%lg", &dhlp) !=1)
	    {
	      fprintf(stderr,
		      "mesh_read_vertex: error reading entry %d!\n",
		      (int) i);
	      return FAIL;
	    }
	  (*m).vertex[i*(*m).vx_w+vxstrt+j] = dhlp;
	}
    }
  (*m).vx_nr  = vx_nr;
  
  /* check if this is really the end of the vertex block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</vertex>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_vertex: spurious end of vertex block "
	      "(no </vertex> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}



/*FUNCTION*/
int mesh_read_elements_t1( FILE *in, struct mesh *m, FIDX el_nr
/* reads the elements part of a mesh file, for a descrition of the
   file format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            el_nr   - the number of elements in the file, as given by
                      the header
   
   output:  m       - (given by reference) a mesh, the elem and face
                      part are allocated here, but only elem set!
                      in face only the RHSF part is defined and PCVL
                      is initiated
   

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
			   ){
  int  ihlp;
  FIDX i, j;
  char buff[1024];

  TRY_MALLOC((*m).elem, el_nr * (*m).el_w , FIDX,
	     mesh_read_elements_t1);
  (*m).el_max = el_nr;

  TRY_MALLOC((*m).face, el_nr * (*m).fc_w , FIDX,
	     mesh_read_elements_t1);
  (*m).fc_max = el_nr;

  /* find the begin of the elements block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<elements>",buff)!=0));
  if (strcmp("<elements>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_elements_t1: file has no <elements>\n");
      return FAIL;
    }

  /* read all the elements one by one */
  for (i=0; i<el_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||(FIDX) (ihlp!=i))
	{
	  fprintf(stderr,
		  "mesh_read_elements_t1: order of entries wrong!\n");
	  return FAIL;
	}
      if ((fscanf(in, "%d", &ihlp) !=1)||(ihlp!=1))
	{
	  fprintf(stderr,
		  "mesh_read_elements_t1: element %d not t1!\n",(int) i);
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr,
		  "mesh_read_elements_t1: entry %d deffect\n",(int) i);
	  return FAIL;
	}
      (*m).face[i*(*m).fc_w+MCT1FCRHSF ]= (FIDX) ihlp;
      (*m).face[i*(*m).fc_w+MCT1FCPCVL ]= -1;
 
      for (j=0; j<3; j++)
	{
	  if (fscanf(in, "%d", &ihlp) !=1)
	    {
	      fprintf(stderr,
		      "mesh_read_elements_t1: error reading entry %d!\n",
		      (int) i);
	      return FAIL;
	    }
	  (*m).elem[i*(*m).el_w+MCT1ELNOD1+j] = (FIDX) ihlp;
	}
    }
  (*m).el_nr  = el_nr;
  (*m).fc_nr  = el_nr;

  /* check if this is really the end of the elements block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</elements>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_elements_t1: "
	      "spurious end of elements block (no </elements> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}


/*FUNCTION*/
int mesh_read_boundary_t1( FILE *in, struct mesh *m, FIDX bd_nr, FIDX sg_nr
/* reads the boundary part of a mesh file, for a descrition of the
   file format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            bd_nr   - the number of boundary entries in the file, as
                      given by the header
            sg_nr   - the number of shape segment entries in the file,
	              as given by the header
   
   output:  m       - (given by reference) a mesh, the edge and bound
                      part are allocated and set here, but in edge
                      only those edges are defined which are part of
                      the boundary!

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
			   ){
  int  ihlp;
  FIDX i, j;
  char buff[1024];

  TRY_MALLOC((*m).edge, bd_nr * (*m).eg_w , FIDX,
	     mesh_read_boundary_t1);
  (*m).eg_max = bd_nr;

  TRY_MALLOC((*m).bound, bd_nr * (*m).bd_w , FIDX,
	     mesh_read_boundary_t1);
  (*m).bd_max = bd_nr;

  /* find the begin of the boundary block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<boundary>",buff)!=0));
  if (strcmp("<boundary>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_boundary_t1: file has no <boundary>\n");
      return FAIL;
    }

  /* read all the boundary one by one */
  for (i=0; i<bd_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr,
		  "mesh_read_boundary_t1: order of entries wrong!\n");
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_boundary_t1: "
		  "read error bound %d \n",(int) i);
	  return FAIL;
	}
      (*m).bound[i*(*m).bd_w+MCT1BDTYPE]= (FIDX) ihlp;

      if ((fscanf(in, "%d", &ihlp) !=1)||(ihlp!=0))
	{
	  fprintf(stderr, "mesh_read_boundary_t1: "
		  "boundary element %d not t1!\n", (int) i);
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_boundary_t1: "
		  "read error bound %d \n", (int) i);
	  return FAIL;
	}
      (*m).bound[i*(*m).bd_w+MCT1BDFNCT]= (FIDX) ihlp;

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_boundary_t1: "
		  "read error bound %d \n", (int) i);
	  return FAIL;
	}
      (*m).bound[i*(*m).bd_w+MCT1BDORIE]= (FIDX) ihlp;

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_boundary_t1: "
		  "read error bound %d \n", (int) i);
	  return FAIL;
	}
      if ((ihlp!=-1)&&(ihlp>=sg_nr))
	{
	  fprintf(stderr,"mesh_read_boundary_t1: "
		  "shape segment out of bounds, bound %d \n", (int) i);
	  return FAIL;
	}
      (*m).bound[i*(*m).bd_w+MCT1BDSSEG]= (FIDX) ihlp;
 
      for (j=0; j<2; j++)
	{
	  if (fscanf(in, "%d", &ihlp) !=1)
	    {
	      fprintf(stderr, "mesh_read_boundary_t1: "
		      "error reading entry %d!\n", (int) i);
	      return FAIL;
	    }
	  (*m).edge[i*(*m).eg_w+MCT1EGNOD1+j] = (FIDX) ihlp;
	}

      /* complete the definition of the edge and bound entry */
      (*m).bound[i*(*m).bd_w+MCT1BDEDGE]=  i;
      (*m).edge[i*(*m).eg_w+MCT1EGCHL1] = -1;
      (*m).edge[i*(*m).eg_w+MCT1EGBND ] =  i;
      (*m).edge[i*(*m).eg_w+MCT1EGPCSU] = -1;
      (*m).edge[i*(*m).eg_w+MCT1EGLVL]  =  0;
    }
  (*m).eg_nr   = bd_nr;
  (*m).bd_nr   = bd_nr;
  
  /* check if this is really the end of the boundary block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</boundary>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_boundary_t1: "
	      "spurious end of boundary block (no </boundary> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}


// for triangle mesh generator
// needs to be outside function block for 'extern "C"' to work
#define REAL double
#define ANSI_DECLARATORS
extern "C" {
#include "triangle/triangle.h"
}

/*FUNCTION*/
int mesh_read_triangle_wrapper( FILE *infile, struct mesh *m,
				FIDX hole_nr
/* reads the hole part of a mesh file, then calls the mesh generator
   "triangle" to generate a mesh (mainly the elements, plus changes of
   the boundary), for a descrition of the file format, see the example
   file "mesh_example.f1m",
   for each volume criterion in pccrit a pcvol entry is created for
   each element  


   input:   infile  - a file open for reading
            hole_nr - the number of hole entries in the file, as
                      given by the header
   
   output:  m       - (given by reference) a mesh, the vertex,
                      element, edge and bound parts are reallocated
                      and reset here, according to the mesh generated
                      by "triangle",  (in edge only those edges are
                      defined which are part of the boundary!)

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
			   ){


  FIDX i, j, dim, bd_nr, bd_cnt;
  int  err;

  FIDX found[2], foundboth, k, r;

  struct triangulateio in, out;
  
  FIDX *bndtmp, *edgtmp;

  /* check if the compiled REAL size of "triangle.h" is the same as
     doulbe ! */

  dim= (*m).dim;

  /* generate data structure for "triangle" */
  /* 1. points */
  TRY_MALLOC(in.pointlist, (*m).vx_nr * dim, double,
	     mesh_read_triangle_wrapper);
  in.pointattributelist = NULL;
  in.pointmarkerlist   = NULL;
  in.numberofpointattributes = 0;
  for (i=0; i<(*m).vx_nr; i++)
    for (j=0; j<dim; j++)
      {
	in.pointlist[i*dim+j]=(*m).vertex[i*MCT1VXLN+MCT1VXSTRT+j];
      }
  in.numberofpoints    = (*m).vx_nr;
  /* 2. triangle stuff */
  in.trianglelist          = NULL;
  in.triangleattributelist = NULL;
  in.trianglearealist      = NULL;
  in.neighborlist          = NULL;
  in.numberoftriangles     = 0;
  in.numberofcorners       = 3;
  in.numberoftriangleattributes = 0;
  
  /* 3. holes, allocate space for the holes, then read them */
  TRY_MALLOC(in.holelist, hole_nr * dim, double,
	     mesh_read_triangle_wrapper);
  err=mesh_read_holes( infile, dim, hole_nr, in.holelist);
  FUNCTION_FAILURE_HANDLE( err, mesh_read_holes, 
			   mesh_read_triangle_wrapper);
  in.numberofholes = hole_nr;

  /* 4. regions */
  in.regionlist      = NULL;
  in.numberofregions = 0;

  /* 5. edges */
  in.edgelist        = NULL;
  in.edgemarkerlist  = NULL;
  in.normlist        = NULL;
  in.numberofedges   = 0;

  /* we saved the segment list to last, as it is the most work,
     6. segments,

     We need the segments which are defined by the boundary entries of
     the mesh file, and their marker. We define one marker for each
     original boundary edge (bd+2), so all (eventual) childs of that
     get the type, boundary function and shape marker copied from
     their parent, the orientation is ignored, as it is regenerated by
     triangle anyway
  */
  TRY_MALLOC( in.segmentlist, (*m).bd_nr*2, int, 
	      mesh_read_triangle_wrapper);
  TRY_MALLOC( in.segmentmarkerlist, (*m).bd_nr, int, 
	      mesh_read_triangle_wrapper);
  /* copy the boundary edges */
  bd_cnt=0;
  for (i=0; i<(*m).bd_nr; i++)
    {
      FIDX eg;
      /* copy the edge */
      eg=(*m).bound[i*(*m).bd_w+MCT1BDEDGE];

      /* only add edge if it is of the finest level */
      if ((*m).edge[eg*(*m).eg_w+MCT1EGCHL1]==-1)
	{
	  in.segmentlist[bd_cnt*2+0]=(*m).edge[eg*(*m).eg_w+MCT1EGNOD1+0];
	  in.segmentlist[bd_cnt*2+1]=(*m).edge[eg*(*m).eg_w+MCT1EGNOD1+1];
	 
	  /* define the marker */
	  in.segmentmarkerlist[bd_cnt]=i+2;
	  
	  bd_cnt++;
	}
    }
  in.numberofsegments=bd_cnt;

  /* init the out pointers */
  out.pointlist             = NULL;
  out.pointattributelist    = NULL;
  out.pointmarkerlist       = NULL;
  out.trianglelist          = NULL;
  out.triangleattributelist = NULL;
  out.trianglearealist      = NULL;
  out.neighborlist          = NULL;
  out.segmentlist           = NULL;
  out.segmentmarkerlist     = NULL;
  out.holelist              = NULL;
  out.regionlist            = NULL;
  out.edgelist              = NULL;
  out.edgemarkerlist        = NULL;
  out.normlist              = NULL;

  
  /* everything is ready for triangle, call it */
  triangulate( "-pqza.125", &in, &out, NULL);
  //fprintf(stderr,"triangle removed for AD-tests\n");
  //return FAIL;

  /* triangle should have defined a valid mesh, we need to copy it in
     our format */
  /* 1. vertices */
  TRY_REALLOC( (*m).vertex, out.numberofpoints * MCT1VXLN, double,
	       mesh_read_triangle_wrapper);
  for (i=0; i<out.numberofpoints; i++)
    for (j=0; j<dim; j++)
      {
	(*m).vertex[i*(*m).vx_w+MCT1VXSTRT+j]=out.pointlist[i*dim+j];
      }
  (*m).vx_nr=out.numberofpoints;
  (*m).vx_max=out.numberofpoints;

  /* 2. elements */
  if (out.numberofcorners!=3)
    {
      fprintf(stderr, "mesh_read_triangle_wrapper: something went " 
	      "wrong in triangle, numberofcorners=%d (!=3)\n",
	      out.numberofcorners);
      return FAIL;
    }
  TRY_REALLOC( (*m).elem, out.numberoftriangles * (*m).el_w, FIDX,
	       mesh_read_triangle_wrapper);
  TRY_REALLOC( (*m).face, out.numberoftriangles * (*m).fc_w, FIDX,
	       mesh_read_triangle_wrapper);
  (*m).el_max = out.numberoftriangles;
  (*m).fc_max = out.numberoftriangles;
  for (i=0; i<out.numberoftriangles; i++)
    {
      for (j=0; j<3; j++)
	{
	  if ( (out.trianglelist[i*3+j]<0) ||
	       (out.trianglelist[i*3+j]>=(*m).vx_nr) )
	    {
	      fprintf(stderr," mesh_read_triangle_wrapper: "
		      "triangle returned invalid vertex!\n"
		      "triangle[%d *3 + %d]=%d, vx_nr=%d\n",
		      (int) i, (int) j, 
		      out.trianglelist[i*3+j], (int) ((*m).vx_nr));
	      return FAIL;
	    }

	  (*m).elem[i*(*m).el_w+MCT1ELNOD1+j]=out.trianglelist[i*3+j];
	}
      /* face rhs and PC-vol default to zero/nothing */
      (*m).face[i*(*m).fc_w+MCT1FCRHSF ]=-1;
      (*m).face[i*(*m).fc_w+MCT1FCPCVL ]=-1;
    }
  (*m).el_nr  = out.numberoftriangles;
  (*m).fc_nr  = out.numberoftriangles;

  /* 3. boundary info + edges */
  /* save a copy of the original boundary and edge as the properties
     of the boundary edges (BC-type, BC-function, shape defining,
     pcsurf) are stored there, */
  bndtmp = (*m).bound;
  edgtmp = (*m).edge;
  (*m).bound=NULL;
  (*m).edge=NULL;

  /* (re)allocate memory for the new version */
  bd_nr = out.numberofsegments;
  TRY_MALLOC((*m).edge, bd_nr * (*m).eg_w , FIDX,
	     mesh_read_triangle_wrapper);
  (*m).eg_max = bd_nr;
  TRY_MALLOC((*m).bound, bd_nr * (*m).bd_w , FIDX,
	     mesh_read_triangle_wrapper);
  (*m).bd_max = bd_nr;

  /* copy all the boundary one by one */
  for (i=0; i<bd_nr; i++)
    {
      FIDX oldbc;

      /* the segmentmarker (-2) defines the bc entry from the file */
      oldbc=out.segmentmarkerlist[i]-2;

      /* copy the properties of the old boundary:
	 type, function, shape marker */
      (*m).bound[i*(*m).bd_w+MCT1BDTYPE]=
	bndtmp[oldbc*(*m).bd_w+MCT1BDTYPE];
      (*m).bound[i*(*m).bd_w+MCT1BDFNCT]=
	bndtmp[oldbc*(*m).bd_w+MCT1BDFNCT];
      (*m).bound[i*(*m).bd_w+MCT1BDSSEG]=
	bndtmp[oldbc*(*m).bd_w+MCT1BDSSEG];

      /* copy the edge from triangle's output */
      for (j=0; j<2; j++)
	{
	  if ( (out.segmentlist[i*2+j]<0) ||
	       (out.segmentlist[i*2+j]>=(*m).vx_nr) )
	    {
	      fprintf(stderr," mesh_read_triangle_wrapper: "
		      "triangle returned invalid vertex!\n"
		      "segment[%d *2 + %d]=%d, vx_nr=%d\n",
		      (int) i, (int) j, 
		      out.segmentlist[i*2+j], (int) ((*m).vx_nr));
	      return FAIL;
	    }
	  (*m).edge[i*(*m).eg_w+MCT1EGNOD1+j] = 
	    out.segmentlist[i*2+j];
	}

      /* orientation is defined by the orientation of the edge in its
	 element (triangle defines the nodes in counterclockwise
	 order) */
      /* find the edge in an triangle */
      k=0;
      foundboth=0;
      while ( (k<(*m).el_nr)&&(foundboth==0) )
	{
	  for(j=0; j<2; j++)
	    {
	      FIDX node;
	      node= out.segmentlist[i*2+j];
	      found[j]=-1;
	      for (r=0; r<3; r++)
		{
		  if ( (*m).elem[k*(*m).el_w+MCT1ELNOD1+r]==node )
		    {
		      found[j]=r;
		    }
		}
	    }
	  if ((found[0]!=-1)&&(found[1]!=-1))
	    {
	      int bma, n;
	      foundboth=1;
	      /* found it, now work out what orientation it defines */
	      bma=found[1]-found[0];
	      n=1;
	      if (bma<0)
		{
		  n=-n;
		  bma=-bma;
		}
	      /* now bma>0, test if >=2 */
	      if (bma>=2) n=-n;
	      (*m).bound[i*(*m).bd_w+MCT1BDORIE]=n;
	    }
	  k++;
	}
      if (foundboth !=1)
	{
	  fprintf(stderr,"mesh_read_triangle_wrapper: "
		  "triangel output segment not found in elements!\n"
		  "segment[%d]= %d %d \n", (int) i,
		  out.segmentlist[i*2+0], out.segmentlist[i*2+1]);
	  return FAIL;
	}

      /* complete the definition of the edge and bound entry */
      (*m).bound[i*(*m).bd_w+MCT1BDEDGE]=  i;
      (*m).edge[i*(*m).eg_w+MCT1EGCHL1] = -1;
      (*m).edge[i*(*m).eg_w+MCT1EGBND ] =  i;
      (*m).edge[i*(*m).eg_w+MCT1EGLVL]  =  0;

      /* copy pcsurf info */
      (*m).edge[i*(*m).eg_w+MCT1EGPCSU] =
	edgtmp[oldbc*(*m).eg_w+MCT1EGPCSU];
      /* if there was pcsurf info: */
      if ((*m).edge[i*(*m).eg_w+MCT1EGPCSU]!=-1)
	{
	  FIDX oldpcs;
	  oldpcs=(*m).edge[i*(*m).eg_w+MCT1EGPCSU];
	  /* check if the edge has changed name */
	  if(i!= oldbc)
	    {
	      /* if so, we need to create a new pcsurf entry */
	      FIDX newpcs;
	      newpcs=(*m).ps_nr;

	      if( (*m).ps_max< (*m).ps_nr+1 )
		{
		  err=err=mesh_more_pcsurf( m, 1);
		  FUNCTION_FAILURE_HANDLE( err, mesh_more_pcsurf, 
					   mesh_read_triangle_wrapper );
		}

	      /* copy the data of the old to the new entry */
	      (*m).pcsurf[newpcs*(*m).ps_w+MCXXPSCRIT]= 
		(*m).pcsurf[oldpcs*(*m).ps_w+MCXXPSCRIT];

	      /* the adjacent vols are determined later, together with
		 the connectivity */
	      (*m).pcsurf[newpcs*(*m).ps_w+MCXXPSVOLE+0]= -1;
	      (*m).pcsurf[newpcs*(*m).ps_w+MCXXPSVOLE+1]= -1;

	      /* point it to the new edge and vice versa */
	      (*m).pcsurf[newpcs*(*m).ps_w+MCXXPSSURF]=i;
	      (*m).edge[i*(*m).eg_w+MCT1EGPCSU]=newpcs;
	      
	      (*m).ps_nr++;
	      oldpcs=newpcs;
	    }
	  /* orientation is that of the boundary entry */
	  (*m).pcsurf[oldpcs*(*m).ps_w+MCXXPSORIE]=
	    (*m).bound[(*m).edge[i*(*m).eg_w+MCT1EGBND]
		       *(*m).bd_w+MCT1BDORIE];
	}
    }
  (*m).eg_nr   = bd_nr;
  (*m).bd_nr   = bd_nr;

  /* 4. for each volume type pccrit create an pcvol entry for each
     element */
  {
    FIDX cr, pccrit_vol, *is_vol; 

    TRY_MALLOC(is_vol, (*m).pc_nr, FIDX, mesh_read_triangle_wrapper);

    /* count the volume type pccrit entries */
    pccrit_vol=0;
    for (cr=0; cr<(*m).pc_nr; cr++)
      {
	FIDX type=(FIDX) (*m).pccrit[cr*(*m).pc_w+MCXXPCTYPE];

	/* set marker is_vol for this criterion */
	switch (type)
	  {
	  case 0: /* squared deviation from given function */
	    is_vol[cr]=1;
	    break;
	  case 1: /* surface force (stokes+navsto) */
	    is_vol[cr]=0;
	    break;
	  case 2: /* energy dissipation */
	    is_vol[cr]=1;
	    break;
	  case 3: /* mesh volume */
	    is_vol[cr]=1;
	    break;
	  case 4: /* mean value */
	    is_vol[cr]=1;
	    break;
	  case 5: /* square deviation from mean value */
	    is_vol[cr]=1;
	    break;
	  default:
	    fprintf(stderr, "mesh_read_triangle_wrapper: "
		    "unknown pccrit type: type=%d\n", (int) type);
	    return FAIL;
	  }
	
	/* count if it is */
	pccrit_vol+=is_vol[cr];
      }

    if (pccrit_vol>1)
      {
	fprintf(stderr,"mesh_read_triangle_wrapper: "
		" >1 volume criteria in the mesh,\n"
		"that should cause problems as only one space in "
		"MCT2FCPCVL\n");
	return FAIL;
      }

    /* now allocate memory for pccrit_vol*fc_nr more pcvol entries */
    if (pccrit_vol>0)
      {
	FIDX new_pv_max;
	new_pv_max = ((*m).pv_nr+(*m).fc_nr*pccrit_vol);
	TRY_REALLOC( (*m).pcvol,
		     (*m).pv_w*new_pv_max,
		     FIDX, mesh_read_triangle_wrapper);
	(*m).pv_max=new_pv_max;

	/* generate an entry for each element and each volume criterion */
	for (cr=0; cr<(*m).pc_nr; cr++)
	  {
	    if (is_vol[cr]==1)
	      {
		for (i=0; i<(*m).fc_nr; i++)
		  {
		    (*m).pcvol[(*m).pv_nr*(*m).pv_w+MCXXPVCRIT] = cr;
		    (*m).pcvol[(*m).pv_nr*(*m).pv_w+MCXXPVVOLM] =  i;
		    (*m).face[i*(*m).fc_w+MCT1FCPCVL] = (*m).pv_nr;
		    (*m).pv_nr++;
		  }
	      }
	  }
      }
    
    /* free temporary data */
    free(is_vol);
  }

  
  /* free the copy of the boundary field */
  free(bndtmp);
  free(edgtmp);

  /* free triangles data */
  free(in.pointlist);
  free(in.pointattributelist);
  free(in.pointmarkerlist);
  free(in.trianglelist);
  free(in.triangleattributelist);
  free(in.trianglearealist);
  free(in.neighborlist);
  free(in.segmentlist);
  free(in.segmentmarkerlist);
  free(in.holelist);
  free(in.regionlist);
  free(in.edgelist);
  free(in.edgemarkerlist);
  free(in.normlist);

  free(out.pointlist);
  free(out.pointattributelist);
  free(out.pointmarkerlist);
  free(out.trianglelist);
  free(out.triangleattributelist);
  free(out.trianglearealist);
  free(out.neighborlist);
  free(out.segmentlist);
  free(out.segmentmarkerlist);
  free(out.edgelist);
  free(out.edgemarkerlist);
  /* these are said to may be ignored 
     free(out.holelist);
     free(out.regionlist);
     free(out.normlist); */
  
  return SUCCESS;
}


/*FUNCTION*/
int mesh_read_holes( FILE *in, FIDX dim, FIDX hole_nr, 
		     double *holes
/* reads the holes part of a mesh file, for a descrition of the file
   format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            dim     - dimension of the hole entries
            hole_nr - number of holes, as specified in the header

   output:  holes   - vector of the points inside the holes,
                      holes[i*dim+j] gives the j-th component of the
                      i-th hole point

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  int  ihlp;
  FIDX i, j;
  double dhlp;
  char buff[1024];

  /* find the begin of the holes block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<holes>",buff)!=0));
  if (strcmp("<holes>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_holes: file has no <holes>\n");
      return FAIL;
    }

  /* read all the points one by one */
  for (i=0; i<hole_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr, "mesh_read_holes: order of entries wrong!\n");
	  return FAIL;
	}
      for (j=0; j<dim; j++)
	{
	  if (fscanf(in, "%lg", &dhlp) !=1)
	    {
	      fprintf(stderr, "mesh_read_holes: "
		      "error reading entry %d!\n", (int) i);
	      return FAIL;
	    }
	  holes[i*dim+j] = dhlp;
	}
    }
  
  /* check if this is really the end of the holes block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</holes>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_holes: spurious end of holes block "
	      "(no </holes> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}


/*FUNCTION*/
int mesh_read_pcsurf( FILE *in, struct mesh *m, FIDX ps_nr
/* reads the performance criteria surfaces of a mesh file, for a
   descrition of the file format, see the example file
   "mesh_example.f1m"

   input:   in      - a file open for reading
            ps_nr   - the number of pcsurf entries in the file, as
                      given by the header
   
   output:  m       - (given by reference) a mesh, the pcsurf part is
                      allocated and set here, the edge and pccrit
                      parts have to be set properly, as they will be
                      referenced 

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
			   ){
  int  ihlp, ihlp2;
  FIDX i, j, edge, found;
  char buff[1024];

  if ((*m).dim!=2)
    {
      fprintf(stderr, "mesh_read_pcsurf: only for dim=2 yet!\n");
      return FAIL;
    }

  TRY_MALLOC((*m).pcsurf, ps_nr * (*m).ps_w , FIDX,
	     mesh_read_pcsurf);
  (*m).ps_max = ps_nr;

  /* find the begin of the pcsurf block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<pcsurf>",buff)!=0));
  if (strcmp("<pcsurf>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_pcsurf: file has no <pcsurf>\n");
      return FAIL;
    }

  /* read all the pcsurf one by one */
  for (i=0; i<ps_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||( (FIDX) ihlp!=i))
	{
	  fprintf(stderr,
		  "mesh_read_pcsurf: order of entries wrong!\n");
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_pcsurf: "
		  "read error pcsurf %d \n", (int) i);
	  return FAIL;
	}
      else if ((ihlp<0)||((FIDX) ihlp>=(*m).pc_nr))
	{
	  fprintf(stderr, "mesh_read_pcsurf: "
		  "pcsurf %d, crit_id=%d out of range!\n",
		  (int) i, ihlp);
	  return FAIL;
	}
      (*m).pcsurf[i*(*m).ps_w+MCXXPSCRIT]= (FIDX) ihlp;

      if ((fscanf(in, "%d", &ihlp) !=1)||((ihlp!=1)&&(ihlp!=-1)))
	{
	  fprintf(stderr, "mesh_read_pcsurf: "
		  "pcsurf %d orientation nonsense!\n", (int) i);
	  return FAIL;
	}
      (*m).pcsurf[i*(*m).ps_w+MCXXPSORIE]= (FIDX) ihlp;

      if (fscanf(in, "%d %d", &ihlp, &ihlp2) !=2)
	{
	  fprintf(stderr, "mesh_read_pcsurf: "
		  "read error pcsurf %d \n", (int) i);
	  return FAIL;
	}

      /* search a edge consisting of these 2 nodes */
      found=-1;
      j=0;
      while ((found==-1)&&(j<(*m).eg_nr))
	{
	  FIDX nod1, nod2;
	  nod1= (*m).edge[j*(*m).eg_w+MCT1EGNOD1  ];
	  if (( nod1 != (FIDX) ihlp) && (nod1 != (FIDX) ihlp2))
	    {
	      /* edge can not be the same, try next */
	      j++;
	    }
	  else
	    {
	      nod2= (*m).edge[j*(*m).eg_w+MCT1EGNOD1+1];
	      if (( nod2 != (FIDX) ihlp) && (nod2 != (FIDX) ihlp2))
		{
		  /* edge can not be the same, try next */
		  j++;
		}
	      else if (( nod1 != (FIDX) ihlp) && (nod2 != (FIDX) ihlp2))
		{
		  /* edge found and has same orientation */
		  found=j;
		}
	      else if (( nod1 != (FIDX) ihlp2) && (nod2 != (FIDX) ihlp))
		{
		  /* edge found but has oposite orientation */
		  found=j;
		  (*m).pcsurf[i*(*m).ps_w+MCXXPSORIE]*=-1;
		}
	    }
	} /* end search edge for this pcsurf */

      if (found==-1)
	{
	  fprintf(stderr,
		  "mesh_read_pcsurf: no edge found with specified "
		  "nodes, entry %d \n", (int) i);
	  return FAIL;
	}

      edge = found;

      /* the vols connected to the pcsurf are determined by the
	 connectivity_init */
      /* complete the definition of the entry */
      (*m).pcsurf[i*(*m).ps_w+MCXXPSSURF]=   edge;
      (*m).edge[edge*(*m).eg_w+MCT1EGPCSU]= i;

    }
  (*m).ps_nr  = ps_nr;
  
  /* check if this is really the end of the pcsurf block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</pcsurf>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_pcsurf: spurious end of pcsurf block "
	      "(no </pcsurf> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}
    


/*FUNCTION*/
int mesh_read_pcvol( FILE *in, struct mesh *m, FIDX pv_nr
/* reads the performance criteria volumes of a mesh file, for a
   descrition of the file format, see the example file
   "mesh_example.f1m",
   if the triangle mesh generator is used no pcvol entries may be
   defined in the file, but for each volume criterion in pccrit a
   pcvol entry is created for each element 

   input:   in      - a file open for reading
            pv_nr   - the number of pcvol entries in the file, as
                      given by the header
   
   output:  m       - (given by reference) a mesh, the pcvol part is
                      allocated and set here, the pccrit and face
                      parts have to be set properly, as they will be
                      referenced 

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
			   ){
  int  ihlp;
  FIDX i;
  char buff[1024];

  if ((*m).dim!=2)
    {
      fprintf(stderr, "mesh_read_pcvol: only for dim=2 yet!\n");
      return FAIL;
    }

  TRY_MALLOC((*m).pcvol, pv_nr * (*m).pv_w , FIDX,
	     mesh_read_pcvol);
  (*m).pv_max = pv_nr;

  /* find the begin of the pcvol block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<pcvol>",buff)!=0));
  if (strcmp("<pcvol>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_pcvol: file has no <pcvol>\n");
      return FAIL;
    }

  /* read all the pcvol one by one */
  for (i=0; i<pv_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr,
		  "mesh_read_pcvol: order of entries wrong!\n");
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr,
		  "mesh_read_pcvol: read error pcvol %d \n",(int) i);
	  return FAIL;
	}
      else if ((ihlp<0)||((FIDX) ihlp>=(*m).pc_nr))
	{
	  fprintf(stderr,
		  "mesh_read_pcvol: pcvol %d, crit_id=%d out of "
		  "range!\n",(int) i, ihlp);
	  return FAIL;
	}
      (*m).pcvol[i*(*m).pv_w+MCXXPVCRIT]= (FIDX) ihlp;

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr,
		  "mesh_read_pcvol: read error pcvol %d \n",(int) i);
	  return FAIL;
	}
      else if ((ihlp<0)||((FIDX) ihlp>=(*m).fc_nr))
	{
	  fprintf(stderr,
		  "mesh_read_pcvol: element out of range, "
		  "entry %d\n",(int) i);
	  return FAIL;
	}

      /* complete the definition of the entry */
      (*m).pcvol[i*(*m).pv_w+MCXXPVVOLM]  = (FIDX) ihlp;
      (*m).face[ihlp*(*m).fc_w+MCT1FCPCVL] = i;
    }
  (*m).pv_nr  = pv_nr;
  
  /* check if this is really the end of the pcvol block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</pcvol>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_pcvol: spurious end of pcvol block "
	      "(no </pcvol> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}
    


/*FUNCTION*/
int mesh_read_pccrit( FILE *in, struct mesh *m, FIDX pc_nr
/* reads the performance function part of a mesh file, for a
   descrition of the file format, see the example file
   "mesh_example.f1m"

   input:   in      - a file open for reading
            pc_nr   - the number of performance functions in the file,
                      as given by the header
   
   output:  m       - (given by reference) a mesh, the pccrit part is
                      allocated and defined

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  int  ihlp;
  FIDX i, k, type, fnr;
  double dhlp;
  char buff[1024];


  if ((*m).dim!=2)
    {
      fprintf(stderr, "mesh_read_pccrit: only for dim=2 yet!\n");
      return FAIL;
    }

  TRY_MALLOC((*m).pccrit, pc_nr * (*m).pc_w , double,
	     mesh_read_pccrit);
  (*m).pc_max = pc_nr;


  /* find the begin of the pccrit block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<pccrit>",buff)!=0));
  if (strcmp("<pccrit>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_pccrit: file has no <pccrit>\n");
      return FAIL;
    }

  /* read all the pccrit entries one by one */
  for (i=0; i<pc_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr, "mesh_read_pccrit: order of entries wrong!\n");
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_pccrit: error reading type!"
		  "(entry %d)\n", (int) i);
	  return FAIL;
	}
      type = ihlp;

      /* decode the type */
      /* type of components */
      switch (type)
	{
	case 0:  /* squared deviation from given function, 
		    data= component_nr, function
		    component_nr - the component for which the square
		                   deviation is to be considered
		    function     - the reference function, pointer
                                   into the function block of the mesh
		 */
	  fnr=2;
	  break;
	case 1:  /* surface force (combinations of drag and lift),
		    handled by stokes_perfcrit_t21 in file stokes_aux.c,
		    data= weight_0, ..., weight_dim
		    weight_i - weight for the i-th component of the
		               force vector, e.g. if the weight vector
		               is orthogonal to the mainstream
		               velocity this is +-lift, if it is
		               parallel it is +-drag
		 */
	  fnr=(*m).dim;
	  break;
 	case 2:  /* energy dissipation,
		    handled by stokes_perfcrit_t21 in file stokes_aux.c,
		    no data required
		 */
	  fnr=0;
	  break;
	case 3:  /* volume of the mesh,
		    no data required */
	  fnr=0;
	  break;
	case 4:  /* mean value, 
		    data= component_nr, coef
		    component_nr - the component for which the mean
		                   value is to be considered
		    coef         - a coefficient by which the result is
                                   multiplied
		 */
	  fnr=2;
	  break;
	case 5:  /* squared deviation from mean value, 
		    data= component_nr, coef
		    component_nr - the component for which the square
		                   deviation is to be considered
		    coef         - a coefficient by which the result is
                                   multiplied
		 */
	  fnr=2;
	  break;
	case 6:  /* squared deviation from reference solution, 
		    data= component_nr1, component_nr2
		    component_nr*- the components for which the square
		                   deviation is to be considered,
		                   i.e. components 
				   component_nr1 to component_nr2
		 */
	  fnr=2;
	  break;
	default: 
	  fprintf(stderr, "mesh_read_pccrit: "
		  "unknown pccrit type!(entry %d)\n", (int) i);
	  return FAIL;
	}
      (*m).pccrit[i*(*m).pc_w+MCXXPCTYPE] = type;

      for (k=0; k<fnr; k++)
	{
	  if (fscanf(in, "%lg", &dhlp) !=1)
	    {
	      fprintf(stderr, "mesh_read_pccrit: "
		      "error reading entry %d!\n", (int) i);
	      return FAIL;
	    }
	  (*m).pccrit[i*(*m).pc_w+MCXXPCDAT1+k] = dhlp;
	}
    }
  (*m).pc_nr  = pc_nr;
  
  /* check if this is really the end of the pccrit block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</pccrit>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_pccrit: spurious end of pccrit block "
	      "(no </pccrit> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}


/*FUNCTION*/
int mesh_read_function( FILE *in, struct mesh *m, FIDX fu_nr
/* reads the function part of a mesh file, for a descrition of the file
   format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            fu_nr   - the number of functions in the file, as given by
                      the header
   
   output:  m       - (given by reference) a mesh, the function part is
                      allocated and defined

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  int  ihlp;
  FIDX i, j, k, type, fdim, fnr;
  double dhlp;
  char buff[1024];

  TRY_MALLOC((*m).func, fu_nr * (*m).fu_w , double,
	     mesh_read_function);
  (*m).fu_max = fu_nr;

  if ((*m).dim!=2)
    {
      fprintf(stderr, "mesh_read_function: only for dim=2 yet!\n");
      return FAIL;
    }


  /* find the begin of the function block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<function>",buff)!=0));
  if (strcmp("<function>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_function: file has no <function>\n");
      return FAIL;
    }

  /* read all the function entries one by one */
  for (i=0; i<fu_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr, "mesh_read_function: order of entries wrong!\n");
	  return FAIL;
	}

      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_function: error reading type!"
		  "(entry %d)\n", (int) i);
	  return FAIL;
	}
      type = (FIDX) ihlp;

      /* decode the type */
      /* scalar / vector */
      if (type>=100)
	{ fdim=(*m).dim; }
      else
	{ fdim=1; }
      /* type of components */
      switch (type%100)
	{
	case 0:
	  fnr=1;
	  break;
	case 1:
	  fnr=(*m).dim+1;
	  break;
	case 2: 
	  fnr=6;
	  break;
	case 10: 
	  fnr=5;
	  break;
	default: 
	  fprintf(stderr, "mesh_read_function: "
		  "unknown function type! (entry %d)\n", (int) i);
	  return FAIL;
	}
      (*m).func[i*(*m).fu_w+MC2XFUTYPE] = type;

      for (j=0; j<fdim; j++)
	{
	  for (k=0; k<fnr; k++)
	    {
	      if (fscanf(in, "%lg", &dhlp) !=1)
		{
		  fprintf(stderr, "mesh_read_function: "
			  "error reading entry %d!\n", (int) i);
		  return FAIL;
		}
	      (*m).func[i*(*m).fu_w+MC2XFUDAT1+j*MC2XFUDATX+k] = dhlp;
	    }
	}
    }
  (*m).fu_nr  = fu_nr;
  
  /* check if this is really the end of the function block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</function>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_function: spurious end of function block "
	      "(no </function> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}



/*FUNCTION*/
int mesh_read_parameter( FILE *in, struct mesh *m, FIDX pa_nr
/* reads the parameter part of a mesh file, for a descrition of the file
   format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            pa_nr   - the number of parameters in the file, as given by
                      the header
   
   output:  m       - (given by reference) a mesh, the parameter part is
                      allocated and defined

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  FIDX i;
  double dhlp;
  char buff[1024];

  TRY_MALLOC((*m).para, (*m).pa_w , double,
	     mesh_read_parameter);
  (*m).pa_max = (*m).pa_w;

  if (pa_nr>(*m).pa_w)
    {
      fprintf(stderr, "mesh_read_parameter: "
	      "number of parameter to large!\n");
      return FAIL;
    }


  /* find the begin of the parameter block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<parameter>",buff)!=0));
  if (strcmp("<parameter>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_parameter: file has no <parameter>\n");
      return FAIL;
    }

  /* read all the parameter data one by one */
  for (i=0; i<pa_nr; i++)
    {
      if (fscanf(in, "%lg", &dhlp) !=1)
	{
	  fprintf(stderr, "mesh_read_parameter: "
		  "error reading datum %d!\n", (int) i);
	  return FAIL;
	}
      (*m).para[i] = dhlp;
    }
  (*m).pa_nr  = pa_nr;
  
  /* check if this is really the end of the parameter block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</parameter>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_parameter: spurious end of parameter block "
	      "(no </parameter> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}





/*FUNCTION*/
int mesh_read_sseg( FILE *in, struct mesh *m, FIDX sg_nr, FIDX sp_nr
/* reads the elements part of a mesh file, for a descrition of the
   file format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            sg_nr   - the number of shape segments in the file, as
                      given by the header
            sp_nr   - the number of shape parameters in the file, as
                      given by the header
   
   output:  m       - (given by reference) a mesh, the sseg
                      part is allocated and set here

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
			   ){
  int  ihlp;
  FIDX i, j, type, type_par;
  char buff[1024];

  TRY_MALLOC((*m).sseg, sg_nr * (*m).sg_w , FIDX,
	     mesh_read_sseg);
  (*m).sg_max = sg_nr;

  /* find the begin of the shape segments block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<shape_seg>",buff)!=0));
  if (strcmp("<shape_seg>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_sseg: file has no <shape_seg>\n");
      return FAIL;
    }

  /* read all the entries one by one */
  for (i=0; i<sg_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr,
		  "mesh_read_sseg: order of entries wrong!\n");
	  return FAIL;
	}
   
      if (fscanf(in, "%d", &ihlp) !=1)
	{
	  fprintf(stderr, "mesh_read_sseg: "
		  "error reading type shape_seg %d\n", (int) i);
	  return FAIL;
	}
      type = ihlp;

      switch (type)
	{
	case 1: /* line */
	  type_par = 4; /* 4 parameters, x0,y0, x1,y1 */
	  break;
	case 2: /* b-spline */
	  type_par = 8; /* 8 parameters, x0,y0, x1,y1, sx0,sy0, sx1,sy1 */
	  break;
	default: /* error */
	  fprintf(stderr, "mesh_read_sseg: "
		  "unknown type, shape_seg %d\n", (int) i);
	  return FAIL;
	}
      (*m).sseg[i*(*m).sg_w+MC2XSGTYPE]=type;


      if ((fscanf(in, "%d", &ihlp) !=1)||
	  ((ihlp<0)||((FIDX) ihlp>=(*m).vx_nr)))
	{
	  fprintf(stderr, "mesh_read_sseg: "
		  "starting vertex out of range, shape_seg %d\n", (int) i);
	  return FAIL;
	}
      (*m).sseg[i*(*m).sg_w+MC2XSGNOD1]= (FIDX) ihlp;



      if (type_par>MC2XSGPMAX)
	{
	  fprintf(stderr, "mesh_read_sseg: "
		  "type_par>=MC2XGPMAX would trouble!\n");
	  return FAIL;
	}

      /* read the parameters */
      for (j=0; j<type_par; j++)
	{
	  if (fscanf(in, "%d", &ihlp) !=1)
	    {
	      fprintf(stderr,"mesh_read_sseg: "
		      "error reading entry %d\n", (int) i);
	      return FAIL;
	    }
	  if ((ihlp<0)||((FIDX) ihlp>=sp_nr))
	    {
	      fprintf(stderr,"mesh_read_sseg: "
		      "parameter out of range, shape_seg entry %d\n",
		      (int) i);
	      return FAIL;
	    }
	  (*m).sseg[i*(*m).sg_w+MC2XSGPAR1+j]= (FIDX) ihlp;
	}
    }
  (*m).sg_nr  = sg_nr;

  /* check if this is really the end of the shape_seg block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</shape_seg>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_sseg: spurious end of shape_seg block "
	      "(no </shape_seg> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}
    




/*FUNCTION*/
int mesh_read_spar( FILE *in, struct mesh *m, FIDX sp_nr
/* reads the shape_par part of a mesh file, for a descrition of the file
   format, see the example file "mesh_example.f1m"

   input:   in      - a file open for reading
            sp_nr   - the number of shape parameters in the file, as
                      given by the header
   
   output:  m       - (given by reference) a mesh, the shape_par part is
                      allocated and defined

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  int  ihlp;
  FIDX i;
  double dhlp;
  char buff[1024];

  TRY_MALLOC((*m).spar, sp_nr*(*m).sp_w , double,
	     mesh_read_spar);
  (*m).sp_max = (*m).sp_nr;

  /* find the begin of the shape_par block */
  while ((fscanf(in, "%1023s", buff)==1)&&(strcmp("<shape_par>",buff)!=0));
  if (strcmp("<shape_par>",buff)!=0)
    {
      fprintf(stderr, "mesh_read_spar: file has no <shape_par>\n");
      return FAIL;
    }

  /* read all the shape_par data one by one */
  for (i=0; i<sp_nr; i++)
    {
      if ((fscanf(in, "%d", &ihlp) !=1)||((FIDX) ihlp!=i))
	{
	  fprintf(stderr, "mesh_read_spar: order of entries wrong!\n");
	  return FAIL;
	}

      if (fscanf(in, "%lg", &dhlp) !=1)
	{
	  fprintf(stderr, "mesh_read_spar: "
		  "error reading shape parameter %d!\n", (int) i);
	  return FAIL;
	}
      (*m).spar[i] = dhlp;
    }
  (*m).sp_nr  = sp_nr;
  
  /* check if this is really the end of the shape_par block */
  if ((fscanf(in, "%1023s", buff)!=1)||(strcmp("</shape_par>",buff)!=0))
    {
      fprintf(stderr, "mesh_read_spar: spurious end of shape_par block "
	      "(no </shape_par> ?)\n");
      return FAIL;
    }
  
  return SUCCESS;
}






/*FUNCTION*/
int mesh_read_compute_conns_t1( struct mesh *m
/* builds the part of the connectivity data of the internal mesh
   format that is not covered by the mesh files (see the example file
   "mesh_example.f1m")

   In/Out:  m       - (given by reference) a mesh, the edge and face
                      parts are completed

   retun:   SUCCESS - success,
            FAIL    - failure, see error message 
*/			 
      		      ){
  FIDX i, j, el, eg, nod1, nod2, found;
  int  err;

  struct ilist *vx2eg; /* an array of list anchors, where each list
			  entry is the id of an edge the vertex of
			  the lists anchor belongs to */
  struct ilist *pntr, *hlp;
                       /* auxiliary list pointers */

  TRY_MALLOC(vx2eg, (*m).vx_nr , struct ilist, 
	     mesh_read_compute_conns_t1);

  /* initialise all the lists as empty */
  for (i=0; i<(*m).vx_nr; i++)
    {
      vx2eg[i].next=NULL;
      vx2eg[i].data=-1;
    }

  /* take all the known edges (from the boundary) into the list */
  for (i=0; i<(*m).eg_nr; i++)
    {
      /* both vertices of the edge */
      for (j=0; j<2; j++)
	{
	  pntr=&vx2eg[(*m).edge[i*(*m).eg_w+MCT1EGNOD1+j]];
	  /* find the end of the list of this vertex */
	  while (pntr->next!=NULL) pntr=pntr->next;
	  /* append an entry to the list */
	  TRY_MALLOC(pntr->next, 1, struct ilist,  
		     mesh_read_compute_conns_t1);
	  pntr       = pntr->next;
	  pntr->data = i;
	  pntr->next = NULL;
	}
    }

  /* now for each element: check if the 3 edges exist, if not create
     them, build a face from these edges */
  for (el=0; el<(*m).el_nr; el++)
    {
      for (j=0; j<3; j++)
	{
	  nod1=(*m).elem[el*(*m).el_w+MCT1ELNOD1+ j     ];
	  nod2=(*m).elem[el*(*m).el_w+MCT1ELNOD1+(j+1)%3];

	  /* check if an edge eg with nod1 and nod2 exists */
	  found=0;
	  pntr=vx2eg[nod1].next;
	  hlp =&vx2eg[nod1];
	  while ((found==0)&&(pntr!=NULL))
	    {
	      eg=pntr->data;
	      if ((nod2==(*m).edge[eg*(*m).eg_w+MCT1EGNOD1  ])||
		  (nod2==(*m).edge[eg*(*m).eg_w+MCT1EGNOD1+1]))
		{
		  found=1;
		}
	      hlp  = pntr;
	      pntr = pntr->next;
	    }
	  
	  /* if not, create one */
	  if (found==0)
	    {
	      /* check if enough space is available */
	      if ((*m).eg_max<(*m).eg_nr+1) 
		{ 
		  err=mesh_more_edges( m, 3);
		  FUNCTION_FAILURE_HANDLE( err, mesh_more_edges,
					   mesh_read_compute_conns_t1);
		}
	      /* create the edge */
	      eg=(*m).eg_nr;
	      (*m).eg_nr++;

	      (*m).edge[eg*(*m).eg_w+MCT1EGNOD1  ]=nod1;
	      (*m).edge[eg*(*m).eg_w+MCT1EGNOD1+1]=nod2;
	      (*m).edge[eg*(*m).eg_w+MCT1EGCHL1  ]= -1;
	      (*m).edge[eg*(*m).eg_w+MCT1EGBND   ]= -1;
	      (*m).edge[eg*(*m).eg_w+MCT1EGPCSU  ]= -1;
	      (*m).edge[eg*(*m).eg_w+MCT1EGLVL   ]=  0;

	      /* register the new edge in vx2eg */
	      /* append an entry to the end of the nod1 list */
	      TRY_MALLOC(hlp->next, 1, struct ilist,  
			 mesh_read_compute_conns_t1);
	      pntr       = hlp->next;
	      pntr->data = eg;
	      pntr->next = NULL;

	      /* append an entry to the end of the nod2 list */
	      pntr=&vx2eg[nod2];
	      /* find the end of the list of this vertex */
	      while (pntr->next!=NULL) pntr=pntr->next;
	      /* append an entry to the list */
	      TRY_MALLOC(pntr->next, 1, struct ilist,  
			 mesh_read_compute_conns_t1);
	      pntr       = pntr->next;
	      pntr->data = eg;
	      pntr->next = NULL;
	      
	    }

	  /* now we have the edge eg which holds both nodes, make it
	     edge of the face corresponding to the element el */
	  (*m).face[el*(*m).fc_w+MCT1FCEDG1+j]=eg;
	}
    }

  /* check the connectivity of the pcsurf entries, which faces they
     belong to */
  for (i=0; i<(*m).ps_nr; i++)
    {
      FIDX edge, found;

      edge= (*m).pcsurf[i*(*m).ps_w+MCXXPSSURF];

      /* search elements containing this edge */
      (*m).pcsurf[i*(*m).ps_w+MCXXPSVOLE+0]= -1;
      (*m).pcsurf[i*(*m).ps_w+MCXXPSVOLE+1]= -1;
      found=0;
      j=0;
      while ((found<2)&&(j<(*m).fc_nr))
	{
	  FIDX eledge;
	  for (eledge=0; eledge<3; eledge++)
	    {
	      if (edge==(*m).face[j*(*m).fc_w+MCT1FCEDG1+eledge])
		{
		  /* this face contains the edge */
		  (*m).pcsurf[i*(*m).ps_w+MCXXPSVOLE]= j;
		  found++;
		}
	    }
	  j++;
	} /* end search element for this edge */

      if (found==-1)
	{
	  fprintf(stderr,
		  "mesh_read_compute_conns_t1: no face found with "
		  "psurf edge, pcsruf entry %d \n", (int) i);
	  return FAIL;
	}
    }

  /* clean up */
  /* empty the list */
  for (i=0; i<(*m).vx_nr; i++)
    {
      pntr=vx2eg[i].next;
      while (pntr!=NULL)
	{
	  hlp  = pntr;
	  pntr = pntr->next;
	  free(hlp);
	}
    }
  free(vx2eg);

  return SUCCESS;
}


/********************************************************************/
/*                                                                  */
/*         T2 element stuff                                         */
/*                                                                  */
/********************************************************************/


/*FUNCTION*/
int mesh_refine_uniform_t2( struct mesh *m 
/* refines the all elements of the mesh m, all necessarry new parts are
   created thereby

   In/Out: m        - the mesh, is modified accordingly, e.g. new
                      elements, faces, edges, vertices, boundary
                      elements and hierarchy entries are created,
		      the element el and the face el are thereby
		      replaced by one of their childs, the edges
		      remain to allow other elements to stay intact,
		      if necessary the fields of the mesh are
		      reallocated to have space for the new stuff

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX old_el_nr, i;
  int  err;

  old_el_nr=(*m).el_nr;

  /* loop over all elements */
  for (i=0; i< old_el_nr; i++)
    {
      err=mesh_refine_element_t2( m , i);
      FUNCTION_FAILURE_HANDLE( err, mesh_refine_element_t2, 
			       mesh_refine_uniform_t2 ); 
    }
  return SUCCESS;
}


/*FUNCTION*/
int mesh_refine_element_t2( struct mesh *m , FIDX el
/* refines the element el of the mesh m, all necessarry new parts are
   created thereby

   Input:  el       - the element/face to be refined

   In/Out: m        - the mesh, is modified accordingly, e.g. new
                      elements, faces, edges, vertices, boundary
                      elements and hierarchy entries are created,
		      the element el and the face el are thereby
		      replaced by one of their childs, the edges
		      remain to allow other elements to stay intact,
		      if necessary the fields of the mesh are
		      reallocated to have space for the new stuff

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX i, j, vx_w, fc_w, el_w, eg_w, ps_w, pv_w, hi_w, eh_w, elhi, eho;
  int  err;


  FIDX newedge[3], midnodes[3], oldnod[3], nodedge[3][2], nodineg[3][2];
  /* nodedge[lnode][this nodes edge]=the edge (global number),
     nodineg[lnode][this nodes edge]=the node is 1st/2nd of the edge */

  vx_w=(*m).vx_w;
  fc_w=(*m).fc_w;
  el_w=(*m).el_w;
  eg_w=(*m).eg_w;
  ps_w=(*m).ps_w;
  pv_w=(*m).pv_w;
  hi_w=(*m).hi_w;
  eh_w=(*m).eh_w;

  /* clear nodedge */
  for (i=0; i<3; i++)
    for (j=0; j<2; j++)
      nodedge[i][j]=-1;

  /* get oldnod */
  for (i=0; i<3; i++)
    oldnod[i]=(*m).elem[el*el_w+MCT2ELNOD1+i];

  /* make sure the 3 edges are refined, store the midnodes, get
     nodedge */ 
  for (i=0; i<3; i++)
    {
      FIDX eg, chld;

      eg=(*m).face[el*fc_w+MCT2FCEDG1+i];

      /* make sure this edge is refined */
      if ((*m).edge[eg*eg_w+MCT2EGCHL1]==-1)
	{
	  err=mesh_split_edge_t2( m, eg);
	  FUNCTION_FAILURE_HANDLE( err, mesh_split_edge_t2,
				   mesh_refine_element_t2); 
	}
      
      /* store the midnode */
      chld=(*m).edge[eg*eg_w+MCT2EGCHL1];
      /* second node of child edge is always the midnode of the old
	 edge */ 
      midnodes[i]=(*m).edge[chld*eg_w+MCT2EGNOD1+1]; 

      for (j=0; j<2; j++)
	{
	  FIDX amnodeedge, thenode, intnode;
	  thenode=(*m).edge[eg*eg_w+MCT2EGNOD1+j];
	  intnode=0;
	  /* find the apropriate internal node for this guy */
	  while ((intnode<3)&&(oldnod[intnode]!=thenode))
	    intnode++;
	  /* if not found, cry */
	  if (intnode>=3)
	    {
	      fprintf(stderr, "internal mesh consistency error in "
		      "mesh_refine_element_t2, file %s, line %d\n",
		      __FILE__, __LINE__);
	      return FAIL;
	    }

	  /* find where in nodedge and nodineg this one goes to */
	  if (nodedge[intnode][0]==-1)
	    { amnodeedge=0; }
	  else
	    { amnodeedge=1; }
	  nodedge[intnode][amnodeedge]=eg;
	  nodineg[intnode][amnodeedge]=j;
	}
    }

  /* make sure we have enough space for the 3 new edges in the inner
     of the element, the 3 new vertices and their hierarchy entires
     who come with the new edges, 3 new faces (the old is recycled as
     a 4th), 3 new elements (the old is recycled as a 4th), 3 new
     pcvol if needed (the old is recycled as a 4th), one element
     hierarchy entry */ 
  if ((*m).eg_nr+3>(*m).eg_max)
    {
      err=mesh_more_edges( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_edges, mesh_refine_element_t2 );
    }
  if ((*m).vx_nr+3>(*m).vx_max)
    {
      err=mesh_more_vertices( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_vertices,
			       mesh_refine_element_t2 );
    }
  if ((*m).hi_nr+3>(*m).hi_max)
    {
      err=mesh_more_hierarchy( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_hierarchy,
			       mesh_refine_element_t2 );
    }
  if ((*m).fc_nr+3>(*m).fc_max)
    {
      err=mesh_more_faces( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_faces, mesh_refine_element_t2 );
    }
  if ((*m).el_nr+3>(*m).el_max)
    {
      err=mesh_more_elems( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_elems, mesh_refine_element_t2 );
    }
  if (((*m).face[el*fc_w+MCT2FCPCVL]!=-1)&&((*m).pv_nr+3>(*m).pv_max))
    {
      err=mesh_more_pcvol( m, 3);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_pcvol, mesh_refine_element_t2 );
    }
  if ((*m).eh_nr+1>(*m).eh_max)
    {
      err=mesh_more_elhiers( m, 1);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_elhiers,
			       mesh_refine_element_t2 );
    }

  /* create the new elhier */
  elhi=(*m).eh_nr;
  (*m).eh_nr++;
  /* level = level of the new child edges +1 (=lvl of edge hierarchy entry */
  (*m).elhier[elhi*eh_w+MCT2EHLVL]=
    (*m).edge[(*m).edge[nodedge[0][0]*eg_w+MCT2EGCHL1]*eg_w + MCT2EGLVL]+1;

  /* for all old corner nodes */
  for (i=0; i<3; i++)
    {
      FIDX chld0, chld1, midnod0, midnod1, newface, newel, newnode,
	old1, old2, level, newhier;

      /* check if the first or the second edge points to the successor
	 node */
      if ((nodedge[i][0]==nodedge[(i+1)%3][0])||
	  (nodedge[i][0]==nodedge[(i+1)%3][1]))
	{ /* first edge positive */
	  eho=0; }
      else
	{ /* second edge must be positive */
	  eho=1; }

      /* get the child edges containig this node */
      chld0=(*m).edge[nodedge[i][0]*eg_w+MCT2EGCHL1+nodineg[i][0]];
      chld1=(*m).edge[nodedge[i][1]*eg_w+MCT2EGCHL1+nodineg[i][1]];

      /* the refinement level of the new edges */
      level=(*m).edge[chld0*eg_w + MCT2EGLVL];
  
      /* get the corresponding midnodes */
      midnod0=(*m).edge[chld0*eg_w+MCT2EGNOD1+1]; 
      midnod1=(*m).edge[chld1*eg_w+MCT2EGNOD1+1];
      /* second node of child edge is allways the midnode of the old
	 edge */

      /* create a new edge from the midnodes */
      newedge[i]=(*m).eg_nr;
      (*m).eg_nr++;

      (*m).edge[newedge[i]*eg_w+MCT2EGNOD1  ]=midnod0;
      (*m).edge[newedge[i]*eg_w+MCT2EGNOD1+1]=midnod1;
      (*m).edge[newedge[i]*eg_w+MCT2EGCHL1  ]=-1;
      (*m).edge[newedge[i]*eg_w+MCT2EGBND   ]=-1;
      (*m).edge[newedge[i]*eg_w+MCT2EGPCSU  ]=-1;
      (*m).edge[newedge[i]*eg_w+MCT2EGLVL   ]=level;

      /* create the midnode of the new edge */
      newnode=(*m).vx_nr;
      (*m).vx_nr++;
      
      old1= (*m).edge[newedge[i]*eg_w + MCT2EGNOD1  ];
      old2= (*m).edge[newedge[i]*eg_w + MCT2EGNOD1+1];

      (*m).vertex[newnode*vx_w+MCT2VXSTRT  ]= 0.5*
	( (*m).vertex[old1*vx_w+MCT2VXSTRT  ]
	  + (*m).vertex[old2*vx_w+MCT2VXSTRT  ]);
      (*m).vertex[newnode*vx_w+MCT2VXSTRT+1]= 0.5*
	( (*m).vertex[old1*vx_w+MCT2VXSTRT+1]
	  + (*m).vertex[old2*vx_w+MCT2VXSTRT+1]);
      
      (*m).edge[newedge[i]*eg_w + MCT2EGNODM]=newnode;

      /* the acording hierarchy entry to the midnode */
      newhier=(*m).hi_nr;
      (*m).hi_nr++;

      (*m).hier[newhier*hi_w+ MCT2HICHLD  ]=newnode;
      (*m).hier[newhier*hi_w+ MCT2HIFAT1  ]=old1;
      (*m).hier[newhier*hi_w+ MCT2HIFAT1+1]=old2;
      (*m).hier[newhier*hi_w+ MCT2HILVL   ]=level+1;

      /* create a new face */
      newface=(*m).fc_nr;
      (*m).fc_nr++;

      (*m).face[newface*fc_w+MCT2FCEDG1  ]=chld0;
      (*m).face[newface*fc_w+MCT2FCEDG1+1]=chld1;
      (*m).face[newface*fc_w+MCT2FCEDG1+2]=newedge[i];

      (*m).face[newface*fc_w+MCT2FCRHSF]=(*m).face[el*fc_w+MCT2FCRHSF];

      /* if necessary create a new pcvol entry */
      if ((*m).face[el*fc_w+MCT2FCPCVL]!=-1)
	{
	  FIDX pcv, new_;

	  pcv=(*m).face[el*fc_w+MCT2FCPCVL];
	  new_=(*m).pv_nr;
	  
	  (*m).pcvol[new_*pv_w+MCXXPVVOLM]=newface;
	  (*m).face[newface*fc_w+MCT2FCPCVL]=new_;
	  (*m).pcvol[new_*pv_w+MCXXPVCRIT]=(*m).pcvol[pcv*pv_w+MCXXPVCRIT];

	  (*m).pv_nr++;
	}
      else
	{
	  (*m).face[newface*fc_w+MCT2FCPCVL]=-1;
	}

      /* check if the edges of the new face are pcsurf edges, if so,
	 let them know the name of the new face */
      for (j=0; j<3; j++)
	{
	  FIDX eg, pcsu;
	  eg=(*m).face[newface*fc_w+MCT2FCEDG1+j];
	  pcsu=(*m).edge[eg*eg_w+MCT2EGPCSU];
	  if (pcsu!=-1)
	    {
	      if ((*m).pcsurf[pcsu*ps_w+MCXXPSVOLE]==-1)
		{
		  (*m).pcsurf[pcsu*ps_w+MCXXPSVOLE]=newface;
		  (*m).pcsurf[pcsu*ps_w+MCXXPSVOLE+1]=-1;
		}
	      else
		{
		  (*m).pcsurf[pcsu*ps_w+MCXXPSVOLE+1]=newface;
		}
	    }
	}


      /* create a new element */
      newel=(*m).el_nr;
      (*m).el_nr++;

      err=mesh_t2_get_elem_from_face( m, newel);
      FUNCTION_FAILURE_HANDLE( err, mesh_t2_get_elem_from_face,
			       mesh_refine_element_t2 );

      /* define this node part of the new elhier */
      /* fathers */
      (*m).elhier[elhi*eh_w+MCT2EHFAT1+i]=(*m).elem[el*el_w+MCT2ELNOD1+i];
      (*m).elhier[elhi*eh_w+MCT2EHFAT1+3+(i+    eho*2)%3]=midnod0;
      (*m).elhier[elhi*eh_w+MCT2EHFAT1+3+(i+(1-eho)*2)%3]=midnod1;
      /* children */
      (*m).elhier[elhi*eh_w+MCT2EHCHL1+i*3+    eho*2]=
	(*m).edge[chld0*eg_w+MCT2EGNODM];
      (*m).elhier[elhi*eh_w+MCT2EHCHL1+i*3+(1-eho)*2]=
	(*m).edge[chld1*eg_w+MCT2EGNODM];
      (*m).elhier[elhi*eh_w+MCT2EHCHL1+i*3+        1]=newnode;

      /* next node */
    }

  /* the outer 3 elements are correctly created now,
     so lets overwrite the old element with the inner one, the 4-th */

  for (i=0; i<3; i++)
    {
      /* the face */
      (*m).face[el*fc_w+MCT2FCEDG1+i]=newedge[i];
      /* ( rhs-function, pcvol stays the same ) */
      /* the edges of the inner face can not be pcsurf edges */
    }
  /* the element */
  err=mesh_t2_get_elem_from_face( m, el);
  FUNCTION_FAILURE_HANDLE( err, mesh_t2_get_elem_from_face,
			   mesh_refine_element_t2 );

  /* that's it */

  return SUCCESS;
}

/*FUNCTION*/
int mesh_split_edge_t2( struct mesh *m , FIDX eg
/* refines the edge eg of the mesh m, so that the first node of eg is
   the first node of the first child edge, the second node is the
   first node of the second child edge and the new middle node is the
   second node of both child edges,
   all necessarry new parts are created

   Input:  eg       - the edge to be refined

   In/Out: m        - the mesh, is modified accordingly, e.g. new
                      edges, vertices, boundary elements and hierarchy
                      entries are created, 
		      thereby the edge remains to allow neigbouring elements
		      to stay intact, 
		      if necessary the fields of the mesh are
		      reallocated to have space for the new stuff

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX i, dim, vx_w, eg_w, hi_w, bd_w, ps_w;
  int  err;
  FIDX newm1, newm2, nod1, nod2, nodm;
  FIDX neweg1, neweg2;
  FIDX oldbd, newbd1,newbd2;
  FIDX newhi1, newhi2, level;

  dim=(*m).dim;
  vx_w=(*m).vx_w;
  eg_w=(*m).eg_w;
  hi_w=(*m).hi_w;
  bd_w=(*m).bd_w;
  ps_w=(*m).ps_w;

  if ((*m).edge[eg*eg_w+ MCT2EGCHL1 ]!= -1)
    {
      /* nothing to do, return */
      return SUCCESS;
    }

  /* first we have to make sure we have space for two new vertices, two
     new edges, eventually new boundary elements, two hierarchy entries
  */
  if ((*m).vx_nr+2>(*m).vx_max)
    {
      err=mesh_more_vertices( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_vertices,  mesh_split_edge_t2 );
    }
  if ((*m).eg_nr+2>(*m).eg_max)
    {
      err=mesh_more_edges( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_edges,  mesh_split_edge_t2 );
    }
  if (((*m).edge[eg*eg_w+ MCT2EGBND]!= -1)&&((*m).bd_nr+2>(*m).bd_max))
    {
      err=mesh_more_boundary( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_boundary,  mesh_split_edge_t2 );
    }
  if (((*m).edge[eg*eg_w+ MCT2EGPCSU]!= -1)&&((*m).ps_nr+2>(*m).ps_max))
    {
      err=mesh_more_pcsurf( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_pcsurf,  mesh_split_edge_t2 );
    }
  if ((*m).hi_nr+2>(*m).hi_max)
    {
      err=mesh_more_hierarchy( m, 2);
      FUNCTION_FAILURE_HANDLE( err, mesh_more_hierarchy,  mesh_split_edge_t2 );
    }

  /* nodes of the edge */
  nod1=(*m).edge[eg*eg_w+ MCT2EGNOD1];
  nod2=(*m).edge[eg*eg_w+ MCT2EGNOD1+1];
  nodm=(*m).edge[eg*eg_w+ MCT2EGNODM];

  /* create the new nodes */
  newm1= (*m).vx_nr;
  (*m).vx_nr++;
  for (i=0;i<dim; i++)
    {
      (*m).vertex[newm1*vx_w+MCT2VXSTRT+i]=
	0.5*(*m).vertex[nod1*vx_w+MCT2VXSTRT+i]+
	0.5*(*m).vertex[nodm*vx_w+MCT2VXSTRT+i];
    }
  newm2= (*m).vx_nr;
  (*m).vx_nr++;
  for (i=0;i<dim; i++)
    {
      (*m).vertex[newm2*vx_w+MCT2VXSTRT+i]=
	0.5*(*m).vertex[nod2*vx_w+MCT2VXSTRT+i]+
	0.5*(*m).vertex[nodm*vx_w+MCT2VXSTRT+i];
    }

  /* refinement level of the old edge */
  level=(*m).edge[eg*eg_w+ MCT2EGLVL];
  
  /* create the new edges */
  neweg1=(*m).eg_nr;
  neweg2=(*m).eg_nr+1;
  (*m).eg_nr+=2;
  /* child edge 1 */
  (*m).edge[neweg1*eg_w+MCT2EGNOD1  ]=nod1;
  (*m).edge[neweg1*eg_w+MCT2EGNOD1+1]=nodm;
  (*m).edge[neweg1*eg_w+MCT2EGNODM  ]=newm1;
  (*m).edge[neweg1*eg_w+MCT2EGCHL1  ]=-1;
  (*m).edge[neweg1*eg_w+MCT2EGBND   ]=-1;
  (*m).edge[neweg1*eg_w+MCT2EGPCSU  ]=-1;
  (*m).edge[neweg1*eg_w+MCT2EGLVL   ]=level+1;
  /* child edge 2 */
  (*m).edge[neweg2*eg_w+MCT2EGNOD1  ]=nod2;
  (*m).edge[neweg2*eg_w+MCT2EGNOD1+1]=nodm;
  (*m).edge[neweg2*eg_w+MCT2EGNODM  ]=newm2;
  (*m).edge[neweg2*eg_w+MCT2EGCHL1  ]=-1;
  (*m).edge[neweg2*eg_w+MCT2EGBND   ]=-1;
  (*m).edge[neweg2*eg_w+MCT2EGPCSU  ]=-1;
  (*m).edge[neweg2*eg_w+MCT2EGLVL   ]=level+1;
  /* make them childs of eg */
  (*m).edge[eg*eg_w+MCT2EGCHL1  ]=neweg1;
  (*m).edge[eg*eg_w+MCT2EGCHL1+1]=neweg2;
  
  /* eventually create new boundary elements, correct boundary
     information on the childs */
  if ((*m).edge[eg*eg_w+ MCT2EGBND]!= -1)
    {
      newbd1=(*m).bd_nr;
      newbd2=(*m).bd_nr+1;
      (*m).bd_nr+=2;
      oldbd=(*m).edge[eg*eg_w+ MCT2EGBND];
      /* newbd1 */
      (*m).bound[newbd1*bd_w+MCT2BDEDGE]=neweg1;
      (*m).bound[newbd1*bd_w+MCT2BDTYPE]=
	(*m).bound[oldbd*bd_w+MCT2BDTYPE]; /* same type as old */
      (*m).bound[newbd1*bd_w+MCT2BDFNCT]=
	(*m).bound[oldbd*bd_w+MCT2BDFNCT]; /* same function data as old
					   */
      (*m).bound[newbd1*bd_w+MCT2BDORIE]=
	(*m).bound[oldbd*bd_w+MCT2BDORIE]; /* first edge has same
					      orientation as the old
					      edge */
      (*m).bound[newbd1*bd_w+MCT2BDSSEG]=
	(*m).bound[oldbd*bd_w+MCT2BDSSEG]; /* same shape variability */
      (*m).edge[neweg1*eg_w+MCT2EGBND   ]=newbd1;
      /* newbd2 */

      (*m).bound[newbd2*bd_w+MCT2BDEDGE]=neweg2;
      (*m).bound[newbd2*bd_w+MCT2BDTYPE]=
	(*m).bound[oldbd*bd_w+MCT2BDTYPE]; /* same type as old */
      (*m).bound[newbd2*bd_w+MCT2BDFNCT]=
	(*m).bound[oldbd*bd_w+MCT2BDFNCT]; /* same function data as old
					   */
      (*m).bound[newbd2*bd_w+MCT2BDORIE]=
	-(*m).bound[oldbd*bd_w+MCT2BDORIE];/* second edge has reverse
					      orientation as the old
					      edge */
      (*m).bound[newbd2*bd_w+MCT2BDSSEG]=
	(*m).bound[oldbd*bd_w+MCT2BDSSEG]; /* same shape variability */
      (*m).edge[neweg2*eg_w+MCT2EGBND   ]=newbd2;
    }


  /* eventually create new pcsurf elements, correct pcsurf
     information on the childs */
  if ((*m).edge[eg*eg_w+ MCT2EGPCSU]!= -1)
    {
      FIDX oldps, newps1, newps2;
      newps1=(*m).ps_nr;
      newps2=(*m).ps_nr+1;
      (*m).ps_nr+=2;
      oldps=(*m).edge[eg*eg_w+ MCT2EGPCSU];
      /* newps1 */
      (*m).pcsurf[newps1*ps_w+MCXXPSSURF]=neweg1;
      (*m).pcsurf[newps1*ps_w+MCXXPSVOLE]=-1;/* faces are set in face
						refinement */
      (*m).pcsurf[newps1*ps_w+MCXXPSCRIT]=
	(*m).pcsurf[oldps*ps_w+MCXXPSCRIT]; /* same criterion as old */
      (*m).pcsurf[newps1*ps_w+MCXXPSORIE]=
	(*m).pcsurf[oldps*ps_w+MCXXPSORIE]; /* first edge has same
					       orientation as the old
					       edge 
					   */
      (*m).edge[neweg1*eg_w+MCT2EGPCSU  ]=newps1;
      /* newps2 */
      (*m).pcsurf[newps2*ps_w+MCXXPSSURF]=neweg2;
      (*m).pcsurf[newps2*ps_w+MCXXPSVOLE]=-1;/* faces are set in face
						refinement */
      (*m).pcsurf[newps2*ps_w+MCXXPSCRIT]=
	(*m).pcsurf[oldps*ps_w+MCXXPSCRIT]; /* same criterion as old */
      (*m).pcsurf[newps2*ps_w+MCXXPSORIE]=
	-(*m).pcsurf[oldps*ps_w+MCXXPSORIE];/* 2nd edge has reverse
					       orientation of the old
					       edge 
					   */
      (*m).edge[neweg2*eg_w+MCT2EGPCSU  ]=newps2;
    }


  /* new hierarchy entries */
  newhi1 = (*m).hi_nr;
  (*m).hi_nr++;
  (*m).hier[newhi1*hi_w+MCT2HICHLD  ] = newm1;
  (*m).hier[newhi1*hi_w+MCT2HIFAT1  ] = nod1;
  (*m).hier[newhi1*hi_w+MCT2HIFAT1+1] = nodm;
  (*m).hier[newhi1*hi_w+MCT2HILVL   ] = level+2;
  newhi2 = (*m).hi_nr;
  (*m).hi_nr++;
  (*m).hier[newhi2*hi_w+MCT2HICHLD  ] = newm2;
  (*m).hier[newhi2*hi_w+MCT2HIFAT1  ] = nod2;
  (*m).hier[newhi2*hi_w+MCT2HIFAT1+1] = nodm;
  (*m).hier[newhi2*hi_w+MCT2HILVL   ] = level+2;

  return SUCCESS;
}




/*FUNCTION*/
int mesh_t2_get_elem_from_face( struct mesh *m , FIDX i
/* extracts the nodes of the element i from the data in m.face and
   m.edge and stores the element i in m

   Input:  i        - the element to be build

   In/Out: m        - the mesh, is modified accordingly, 
                      only the elem part is modified

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  FIDX j, fc_w, el_w, eg_w;
  FIDX nodes[3], edges[3];

  fc_w=(*m).fc_w; el_w=(*m).el_w; eg_w=(*m).eg_w;

  /* the nodes of the element have to be ordered such that 0,1,2
     are the corners of the triangle, 3 is between 0 and 1, 4
     between 1 and 2, 5 between 2 and 0
  */


  for (j=0; j<3; j++)
    edges[j]=(*m).face[i*fc_w+MCT2FCEDG1+j];


  /* first, get the edges into order,
     first edge stays first and defines the first two points
  */
  nodes[0]=(*m).edge[edges[0]*eg_w+MCT2EGNOD1];
  nodes[1]=(*m).edge[edges[0]*eg_w+MCT2EGNOD1+1];
  
  /* now see which of the two other edges has node[1] as well */
  if ((nodes[1]!=(*m).edge[edges[1]*eg_w+MCT2EGNOD1])&&
      (nodes[1]!=(*m).edge[edges[1]*eg_w+MCT2EGNOD1+1]))
    {
      /* edges 2 and 3 have to be swaped */
      j=edges[2];
      edges[2]=edges[1];
      edges[1]=j;
    }
  /* get the 3rd node */
  if (nodes[1]==(*m).edge[edges[1]*eg_w+MCT2EGNOD1  ])
    {
      /* ok, the second node of the second edge is the 3rd node of
	 the face
      */
      nodes[2]=(*m).edge[edges[1]*eg_w+MCT2EGNOD1+1];
    }
  else if (nodes[1]==(*m).edge[edges[1]*eg_w+MCT2EGNOD1+1])
    {
      /* ok, the first node of the second edge is the 3rd node of
	 the face
      */
      nodes[2]=(*m).edge[edges[1]*eg_w+MCT2EGNOD1  ];
    }
  else
    { 
      /* cry */
      fprintf(stderr,
      "mesh_t2_get_elem_from_face: face with one node only in one edge!\n");
      return FAIL;
    }
  
  for (j=0; j<3; j++)
    {
      /* nodes 0,1,2 */
      (*m).elem[i*el_w + j]=nodes[j];
      /* nodes 3,4,5 are midnodes of the edges */
      (*m).elem[i*el_w+3+j]=(*m).edge[edges[j]*eg_w+MCT2EGNODM];
    }

  return SUCCESS;
}


/*FUNCTION*/
void multilvl_free( struct multilvl *ml
/* empties the multilvl data structure, such that it can be freed

   Output: ml      - internal memory is released,
*/
	       ){
  (*ml).vx_nr = 0; 
  (*ml).dim   = 0; 
  (*ml).lmax  = 0; 

  free((*ml).nlevl);
  (*ml).nlevl=NULL;

  free((*ml).levlvx);
  (*ml).levlvx=NULL;

  free((*ml).levlhindx);
  (*ml).levlhindx=NULL;

  return;
}


/*FUNCTION*/
void multilvl_null_def( struct multilvl *ml
/* defines the values in multilvl data structure, such that it can be
   given to multilvl_free and give no errors, even before
   multilvl_init is called

   Output: ml      - is defined to be empty,
*/
	       ){
  (*ml).vx_nr = 0; 
  (*ml).dim   = 0; 
  (*ml).lmax  = 0; 

  (*ml).nlevl=NULL;
  (*ml).levlvx=NULL;
  (*ml).levlhindx=NULL;

  return;
}


/*FUNCTION*/
int multilvl_init_t2( struct mesh *m, FIDX datdim, struct multilvl *ml
/* initialises the multilvl data structure
   
   Input:  m       - the mesh (T2) to which the multilvl data shall
                     correspond (the important bit is the hierarchy
                     data) 
	   datdim  - the number of dofs per node for which the
	             multilvl data will be created, on each of the
	             levels dim*(this levels vx_nr) components will be
	             available

   Output: ml      - internal data is initialised, memory allocated
                     for the arrays

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		 ){
  int err;

  /* call the general code for type=2 */
  err=multilvl_init_tx( m, datdim, ml, 2); 
  FUNCTION_FAILURE_HANDLE( err, multilvl_init_tx, multilvl_init_t2);
  
  return SUCCESS;
}




/*FUNCTION*/
int multilvl_init_tx( struct mesh *m, FIDX datdim, struct multilvl *ml,
		      int type
/* initialises the multilvl data structure
   
   Input:  m       - the mesh to which the multilvl data shall
                     correspond (the important bit is the hierarchy
                     data) 
	   datdim  - the number of dofs per node for which the
	             multilvl data will be created, on each of the
	             levels dim*(this levels vx_nr) components will be
	             available
           type    - the type of the mesh, type=1 linear, =2 quadratic

   Output: ml      - internal data is initialised, memory allocated
                     for the arrays

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		 ){
  FIDX i, dim, vx_nr, hi_w, hi_nr, eh_w, eh_nr, eg_w, eg_nr;
  FIDX level, lvl; 

  FIDX l_mctxhilvl, l_mctxhichld,  l_mctxhifat1, l_mctxehlvl;

  vx_nr= (*m).vx_nr;
  dim  = (*m).dim;
  eg_nr= (*m).eg_nr;
  eg_w = (*m).eg_w;
  hi_nr= (*m).hi_nr;
  hi_w = (*m).hi_w;
  eh_nr= (*m).eh_nr;
  eh_w = (*m).eh_w;

  /* basic init */
  (*ml).vx_nr = vx_nr; 
  (*ml).dim   = datdim; 

  switch (type)
    {
    case 1: /* T1 mesh */
      l_mctxhilvl  = MCT1HILVL;
      l_mctxhichld = MCT1HICHLD;
      l_mctxhifat1 = MCT1HIFAT1;
      l_mctxehlvl  = 0;
      break;
    case 2: /* T2 mesh */
      l_mctxhilvl  = MCT2HILVL;
      l_mctxhichld = MCT2HICHLD;
      l_mctxhifat1 = MCT2HIFAT1;
      l_mctxehlvl  = MCT2EHLVL;
      break;
    default:
      fprintf(stderr,"multilvl_init_tx: unknown type=%d\n",type);
      return FAIL;
    }

  (*ml).type=type;
    

  /* check for the highest level */
  (*ml).lmax=0; 
  for (i=0; i<hi_nr; i++)
    {
      level=(*m).hier[i*hi_w+l_mctxhilvl];
      if (level>(*ml).lmax) (*ml).lmax=level;
      if (level<0) 
	{
	  fprintf(stderr,"multilvl_init_tx: level<0 not allowed!\n");
	  return FAIL;
	}
    }

  /* allocate memory for nlevl and levlhindx */
  TRY_MALLOC( (*ml).nlevl, (*ml).lmax+2, FIDX, multilvl_init_tx);
  TRY_MALLOC( (*ml).levlvx, (*ml).lmax+1, FIDX, multilvl_init_tx);
  TRY_MALLOC((*ml).levlhindx, (*ml).lmax+2, FIDX, multilvl_init_tx);
  

  /* search the highest nodennumber in each level */
  for (i=0; i<=(*ml).lmax+1; i++) (*ml).nlevl[i]=-1;
  for (i=0; i<hi_nr; i++)
    {
      level=(*m).hier[i*hi_w + l_mctxhilvl];
      if ((*ml).nlevl[level]<(*m).hier[i*hi_w + l_mctxhichld])
	(*ml).nlevl[level]=(*m).hier[i*hi_w + l_mctxhichld];
      if ((*ml).nlevl[level]<(*m).hier[i*hi_w + l_mctxhifat1])
	(*ml).nlevl[level]=(*m).hier[i*hi_w + l_mctxhifat1];
      if ((*ml).nlevl[level]<(*m).hier[i*hi_w + l_mctxhifat1+1])
	(*ml).nlevl[level]=(*m).hier[i*hi_w + l_mctxhifat1+1];

      if (level>=1)
	{
	  if ((*ml).nlevl[level-1]<(*m).hier[i*hi_w + l_mctxhifat1])
	    (*ml).nlevl[level-1]=(*m).hier[i*hi_w + l_mctxhifat1];
	  if ((*ml).nlevl[level-1]<(*m).hier[i*hi_w + l_mctxhifat1+1])
	    (*ml).nlevl[level-1]=(*m).hier[i*hi_w + l_mctxhifat1+1];
	}
    }
  /* if no hierarchy given, use current vertex number */
  if (hi_nr==0) (*ml).nlevl[0]=(*m).vx_nr-1;

  /* now correct from highest index to number of entries and copy it
     to levlvx */
  for (i=0; i<=(*ml).lmax+1; i++)
    {
      (*ml).nlevl[i]+=1;
    }
  for (i=0; i<=(*ml).lmax; i++)
    {
      (*ml).levlvx[i]=(*ml).nlevl[i];
    }

  
  if ((*ml).nlevl[(*ml).lmax]!=(*m).vx_nr)
    {
      fprintf(stderr, 
	      "multilvl_init_tx: something is wrong with vx_nr=%d, "
	      "max_vx=%d\n", (int) (*m).vx_nr,
	      (int) (*ml).nlevl[(*ml).lmax]);
      return FAIL;
    }

  /* now define the end positions of each level in the help vectors,
     the finest level lmax comes first, then..., end position of one
     vector=start position of the next */
  for (i=(*ml).lmax;   i>=0; i--) (*ml).nlevl[i]*=datdim;
  for (i=(*ml).lmax-1; i>=0; i--) (*ml).nlevl[i]+=(*ml).nlevl[i+1];
  (*ml).nlevl[(*ml).lmax+1]=0;


  /* set levelhindx for each level */
  (*ml).levlhindx[0]=0;
  for (lvl=0; lvl<=(*ml).lmax; lvl++)
    {
      (*ml).levlhindx[lvl+1]=0;
      if (type==2) /* element hierarchy used for T2 meshes */
	{    
	  for (i=0; i<eh_nr; i++)
	    {
	      if ((*m).elhier[i*eh_w+l_mctxehlvl]==lvl)
		{
		  (*ml).levlhindx[lvl+1]=i+1;
		}
	    }
	}
      else if (type==1) /* hierarchy used for T1 meshes */
	{    
	  for (i=0; i<hi_nr; i++)
	    {
	      if ((*m).hier[i*hi_w+MCT1HILVL]==lvl)
		{
		  (*ml).levlhindx[lvl+1]=i+1;
		}
	    }
	}
    }
  return SUCCESS;
}




/*FUNCTION*/
void bpx_free( struct bpxdata *bpx
/* empties the data structure for the BPX preconditioner, such that it
   can be freed

   Output: bpx     - internal memory is released,
*/
	       ){
  free((*bpx).hvec);
  (*bpx).hvec=NULL;

  if ((*bpx).cmat!=NULL) 
    coarse_mat_free((*bpx).cmat);
  free((*bpx).cmat);
  (*bpx).cmat=NULL;  

  /* it is not upon us to delete the mesh, the ml, or the projector
     but we can forget about them */
  (*bpx).msh=NULL;
  (*bpx).ml=NULL;
  (*bpx).P=NULL;

  return;
}





/*FUNCTION*/
int bpx_init_tx( struct bpxdata *bpx, struct multilvl *ml
/* initialises the data structure for the BPX preconditioner
   
   Input:  msh     - the mesh to which the BPX data shall correspond
                     (the important bit is the hierarchy data)
           ml      - multilevel data struct

   Output: bpx     - internal data is initialised, memory allocated
                     for the arrays

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		 ){
  /* allocate memory for the help vectors */
  TRY_MALLOC((*bpx).hvec,  (*ml).nlevl[0], double,
	     bpx_init);

  /* initialise the coarse grid as non-existend */
  (*bpx).cmat=NULL;

  /* the mesh, if needed, should be set elsewere */
  (*bpx).msh =NULL; 

  /* the multilevel struct */
  (*bpx).ml  =ml; 

  /* the projector, if needed, should be set elsewere */
  (*bpx).P =NULL; 

  return SUCCESS;
}



/*FUNCTION*/
void mg_free( struct mgdata *mg
/* empties the multilvl data structure, such that it can be freed

   Output: mg      - internal memory is released,
*/
	       ){
  (*mg).vx_nr=0; 
  (*mg).dim=0; 

  free((*mg).xl);
  (*mg).xl=NULL;

  free((*mg).dxl);
  (*mg).dxl=NULL;

  free((*mg).bl);
  (*mg).bl=NULL;

  free((*mg).sorterl);
  (*mg).sorterl=NULL;

  (*mg).bas_n=0;
  (*mg).chl_n=0;

  free((*mg).phih);
  (*mg).phih=NULL;

  free((*mg).done);
  (*mg).done=NULL;

  free((*mg).invdiag);
  (*mg).invdiag=NULL;

  /* it is not upon us to delete msh, ml, Ks, P, as we only have
     copies of each of them, but set them to NULL */
  (*mg).msh  = NULL;
  (*mg).ml   = NULL;
  (*mg).Ks   = NULL;
  (*mg).P    = NULL;

  if ((*mg).cmat != NULL)
    {
      coarse_mat_free((*mg).cmat);
      free((*mg).cmat);
    }
  (*mg).cmat = NULL;


  return;
}


/*FUNCTION*/
void mg_null_def( struct mgdata *mg
/* defines the multilvl data structure, such that a call ti mg_free
   prodruces no errors even though mg_init has not been called yet

   Output: mg      - is defined to be empty,
*/
	       ){
  (*mg).vx_nr=0; 
  (*mg).dim=0; 

  (*mg).xl=NULL;
  (*mg).dxl=NULL;
  (*mg).bl=NULL;
  (*mg).sorterl=NULL;

  (*mg).bas_n=0;
  (*mg).chl_n=0;

  (*mg).phih=NULL;
  (*mg).done=NULL;
  (*mg).invdiag=NULL;
  
  (*mg).msh  = NULL;
  (*mg).ml   = NULL;
  (*mg).Ks   = NULL;
  (*mg).P    = NULL;
  (*mg).cmat = NULL;

  (*mg).CGC_scale= 1.0;

  return;
}


#include "elements.h"
/*FUNCTION*/
int mg_init_tx( struct sparse *Ks, struct mesh *m, struct multilvl *ml,
		struct mgdata *mg, struct projector1 *P
/* initialises the data structure for the multigrid preconditioner
   
   Input:  Ks      - array of sparse matrices, K[i] is the stiffness
                     matrix for the mesh on the i-th level
           m       - the mesh (T2) to which the mgdata shall correspond
                     (the important bit is the element hierarchy data)
	   ml      - multilevel data
	   P       - Projector for boundary conditions, give P=NULL if
	             not wanted

   Output: mg      - internal data is initialised, memory allocated
                     for the arrays

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		){
  FIDX i;
  int  err;
  FIDX vx_nr;
  FIDX lvl, lvl_vx;

  double *entry;
  double *invdiag;

  FIDX bas_n;
  double *phih, *gradphi, *hessphi;
  double points[9*2]={ 0.25, 0.0,
		       0.25, 0.25,
		       0.0,  0.25,
		       0.75, 0.25,
		       0.5,  0.25,
		       0.75, 0.0,
		       0.0,  0.75,
		       0.25, 0.5,
		       0.25, 0.75 };
  FIDX *done;

  vx_nr=(*m).vx_nr;
  (*mg).vx_nr=(*m).vx_nr;
  (*mg).dim=(*ml).dim;

  if ((*ml).type==2)
    {
      /* initialise interpolation info */
      err= eval_basis( 2, tria, 2,
		       9, points,
		       &bas_n, &phih, &gradphi, &hessphi);
      FUNCTION_FAILURE_HANDLE( err, eval_basis, mg_init_tx);
      free( gradphi );
      free( hessphi );
      (*mg).phih=phih;
      (*mg).bas_n=bas_n;
      (*mg).chl_n=9;
    }
  else
    {
      /* linear elements don't use the element hierarchy but the edge
	 hierarchy */
      (*mg).phih=NULL;
      (*mg).bas_n=0;
      (*mg).chl_n=0;
    }


  /* allocate mg.xl, mg.dxl, mg.bl */
  TRY_MALLOC((*mg).xl,  (*ml).nlevl[0], double, mg_init_tx);
  TRY_MALLOC((*mg).dxl, (*ml).nlevl[0], double, mg_init_tx);
  TRY_MALLOC((*mg).bl,  (*ml).nlevl[0], double, mg_init_tx);

  /* allocate mg.done */
  TRY_MALLOC((*mg).done, (*mg).dim*vx_nr, FIDX, mg_init_tx);
  done=(*mg).done;

  /* if the matrices are given, set invdiag */
  if (Ks != NULL)
    {
      /* allocate mg.invdiag */
      TRY_MALLOC((*mg).invdiag, (*ml).nlevl[0], double, mg_init_tx);
      invdiag=(*mg).invdiag;
      
      /* for each level, finest to coarsest */
      for (lvl=(*ml).lmax; lvl>=0; lvl--)
	{
	  lvl_vx=(*ml).nlevl[lvl]-(*ml).nlevl[lvl+1];
	  /* printf("lvl=%d  lvl_vx=%4d\n", (int) lvl, (int) lvl_vx); /* */
	  for (i=0; i<lvl_vx; i++)
	    {
	      err=sparse_get_entry( &Ks[lvl], i, i, 0, &entry);
	      FUNCTION_FAILURE_HANDLE(err, sparse_get_entry, mg_init_tx);

	      invdiag[i+(*ml).nlevl[lvl+1]]=1.0/(*entry); 
	    }
	  /* if the projector is given project invdiag */
	  if (P!=NULL)
	    {
	      for (i=0; i<(*P).len; i++)
		{
		  FIDX child,dof;
		  dof=(*P).V[i];
		  MLVLFINDENTRY(child, dof, lvl, *ml);
		  if (child >=0)
		    {
		      invdiag[child]=0.0;
		    }
		}
	    }
	}/* end of loop over levels */
    }

  (*mg).sorterl=NULL;

  (*mg).msh  = m;
  (*mg).ml   = ml;
  (*mg).Ks   = Ks;
  (*mg).P    = P;
  (*mg).cmat = NULL;

  (*mg).CGC_scale= 1.0;

  return SUCCESS;
}

/*FUNCTION*/
int mg_restrict_t2( struct mgdata *mg,
		    FIDX lvlin, FIDX lvlout, double *mlv
/* restricts the vector lvlin part of mlv to lvlout, by applying the
   transpose operation of interpolation
   
   Input:  lvlin   - the level from which the restriction shall start
	   lvlout  - the level to which the restriction shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin>lvl>=lvlout are
	             overwritten by the restriction from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		){
  FIDX d, i, j, k;
  FIDX dim, vx_nr, eh_w, eh_nr;
  FIDX node, fath, fathl, fathlp1, chldlp1;
  FIDX lvl;

  FIDX bas_n, chl_n;
  double *phih;
  FIDX *done;

  struct mesh *m;
  struct multilvl *ml;
  
  m     = (*mg).msh;
  ml    = (*mg).ml;

  vx_nr = (*mg).vx_nr;
  dim   = (*mg).dim;
  eh_w  = (*m).eh_w;
  eh_nr = (*m).eh_nr;
  phih  = (*mg).phih;
  bas_n = (*mg).bas_n;
  chl_n = (*mg).chl_n;
  done  = (*mg).done;

  if (vx_nr!=(*m).vx_nr)
    {
      fprintf(stderr,
	      "mg_restrict_t2: mesh doesn't match mgdata (vx_nr test)!\n");
      return FAIL;
    }

  if (lvlout>lvlin)
    {
      fprintf(stderr,
	      "mg_restrict_t2: %d=lvlout>lvlin=%d, makes no sense!\n",
	      (int) lvlout, (int) lvlin);
      return FAIL;
    }

  if ((lvlin>(*ml).lmax)||(lvlout<0))
    {
      fprintf(stderr,
	      "mg_restrict_t2: lvlout=%d or lvlin=%d, out of range "
	      "(0<lvl<%d=maxl)\n", (int) lvlout, (int) lvlin,
	      (int) (*ml).lmax);
      return FAIL;
    }
  
      
  /* now, for each lvl, lvlin-1 to lvlout */
  for (lvl=lvlin-1; lvl>=lvlout; lvl--)
    {
      /* mark all nodes as undone */
      for (i=0; i<dim*vx_nr; i++) done[i]=0;

      /* compute mlv[lvl]= (level interpolation)^T * mlv[lvl+1] */
      /* copy the father nodes */
      for (i=(*ml).levlhindx[lvl+1]; i<(*ml).levlhindx[lvl+2]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).elhier[i*eh_w+MCT2EHLVL]!=lvl+1)
	    {
	      fprintf(stderr,
		      "mg_restrict_t2: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  for (j=0; j<bas_n; j++)
	    for (d=0; d<dim; d++)
	      {
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(fathl, node, lvl, *ml);
		    MLVLFINDENTRY(fathlp1, node, lvl+1, *ml);
		    /* printf("node=%3d   fathl=%3d    fathlp1=%3d\n",
		       node, fathl, fathlp1); /* */
		    mlv[fathl]=mlv[fathlp1];
		  }
	      }
	}
      for (i=(*ml).levlhindx[lvl+1]; i<(*ml).levlhindx[lvl+2]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).elhier[i*eh_w+MCT2EHLVL]!=lvl+1)
	    {
	      fprintf(stderr,
		      "mg_restrict_t2: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  for (j=0; j<9; j++)
	    for (d=0; d<dim; d++)
	      {
		/* check out the child nodes */
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHCHL1+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(chldlp1, node, lvl+1, *ml);
		    for (k=0; k<bas_n; k++)
		      {
			/* father += alpha*child */
			fath=d*vx_nr
			  +(*m).elhier[i*eh_w+MCT2EHFAT1+k];
			MLVLFINDENTRY(fathl, fath, lvl, *ml);
			mlv[fathl]+=phih[k+bas_n*j]*mlv[chldlp1];
		      }
		  }
		/* end check out child nodes */
	      }
	}
    }/* end of loop over levels */

  return SUCCESS;
}


/*FUNCTION*/
int mg_interpolate_t2( struct mgdata *mg,
		       FIDX lvlin, FIDX lvlout, double *mlv
/* interpolates the vector lvlin part of mlv to lvlout, 
   
   Input:  lvlin   - the level from which the interpolation shall start
	   lvlout  - the level to which the interpolation shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin<lvl<=lvlout are
	             overwritten by the interpolation from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		){
  FIDX d, i, j, k;
  FIDX dim, vx_nr, eh_w, eh_nr;
  FIDX node, fath, fathl, fathlm1, chldl;
  FIDX lvl;

  FIDX bas_n, chl_n;
  double *phih;
  FIDX *done;

  struct mesh *m;
  struct multilvl *ml;
  
  m     = (*mg).msh;
  ml    = (*mg).ml;

  vx_nr = (*mg).vx_nr;
  dim   = (*mg).dim;
  eh_w  = (*m).eh_w;
  eh_nr = (*m).eh_nr;
  phih  = (*mg).phih;
  bas_n = (*mg).bas_n;
  chl_n = (*mg).chl_n;
  done  = (*mg).done;


  if (vx_nr!=(*m).vx_nr)
    {
      fprintf(stderr,
	      "mg_interpolate_t2: mesh doesn't match mgdata (vx_nr test)!\n");
      return FAIL;
    }

  if (lvlout<lvlin)
    {
      fprintf(stderr,
	      "mg_interpolate_t2: %d=lvlout<lvlin=%d, makes no sense!\n",
	      (int) lvlout, (int) lvlin);
      return FAIL;
    }

  if ((lvlout>(*ml).lmax)||(lvlin<0))
    {
      fprintf(stderr, "mg_interpolate_t2: "
	      "lvlout=%d or lvlin=%d, out of range (0<lvl<%d=lmax)\n",
	      (int) lvlout, (int) lvlin, (int) (*ml).lmax);
      return FAIL;
    }
  
  /* now, for each lvl, lvlin+1 to lvlout */
  for (lvl=lvlin+1; lvl<=lvlout; lvl++)
    {
      /* mark all nodes as undone */
      for (i=0; i<dim*vx_nr; i++) done[i]=0;

      /* compute mlv[lvl]= (level interpolation) * mlv[lvl-1] */
      for (i=(*ml).levlhindx[lvl]; i<(*ml).levlhindx[lvl+1]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).elhier[i*eh_w+MCT2EHLVL]!=lvl)
	    {
	      fprintf(stderr,
		      "mg_interpolate_t2: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  /* copy the father nodes */
	  for (j=0; j<bas_n; j++)
	    for (d=0; d<dim; d++)
	      {
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(fathl, node, lvl, *ml);
		    MLVLFINDENTRY(fathlm1, node, lvl-1, *ml);
		    mlv[fathl]=mlv[fathlm1];
		  }
	      }
	  for (j=0; j<9; j++)
	    for (d=0; d<dim; d++)
	      {
		/* check out the child nodes */
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHCHL1+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(chldl, node, lvl, *ml);
		    mlv[chldl]=0.0;
		    for (k=0; k<bas_n; k++)
		      {
			/* father += alpha*child */
			fath=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+k];
			MLVLFINDENTRY(fathlm1, fath, lvl-1, *ml);
			mlv[chldl]+=phih[k+bas_n*j]*mlv[fathlm1];
		      }
		  }
		/* end check out child nodes */
	      }
	}
    }/* end of loop over levels */

  return SUCCESS;
}





/*FUNCTION*/
int mg_restrict_t1( struct mgdata *mg, FIDX lvlin, FIDX lvlout, double *mlv
/* restricts the vector lvlin part of mlv to lvlout, by applying the
   transpose operation of interpolation
   
   Input:  lvlin   - the level from which the restriction shall start
	   lvlout  - the level to which the restriction shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin>lvl>=lvlout are
	             overwritten by the restriction from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		    ){
  FIDX i, j, d, lvl, child, fath1, fath2;
  FIDX hi_w, mldim, vx_nr;

  struct mesh *m;
  struct multilvl *ml;
  
  m     = (*mg).msh;
  ml    = (*mg).ml;

  mldim = (*ml).dim;
  vx_nr = (*mg).vx_nr;

  hi_w  = (*m).hi_w;

  
  if (vx_nr!=(*m).vx_nr)
    {
      fprintf(stderr,
	      "mg_restrict_t1: mesh doesn't match mgdata (vx_nr test)!\n");
      return FAIL;
    }

  if (lvlout>lvlin)
    {
      fprintf(stderr,
	      "mg_restrict_t1: %d=lvlout>lvlin=%d, makes no sense!\n",
	      (int) lvlout, (int) lvlin);
      return FAIL;
    }

  if ((lvlin>(*ml).lmax)||(lvlout<0))
    {
      fprintf(stderr,
	      "mg_restrict_t1: lvlout=%d or lvlin=%d, out of range "
	      "(0<lvl<%d=maxl)\n", (int) lvlout, (int) lvlin, 
	      (int) (*ml).lmax);
      return FAIL;
    }

  /* now, for each lvl, lvlin-1 to lvlout */
  for (lvl=lvlin-1; lvl>=lvlout; lvl--)
    {
      FIDX l_vx_nr=((*ml).nlevl[lvl+1]-(*ml).nlevl[lvl+2])/mldim;
      
      /* initialise the new (coarser) level */
      for (j=0; j<l_vx_nr; j++)
	for (d=0; d<mldim; d++)
	  {
	    /* get the entry number of the coarser level */
	    child = j+d*vx_nr;
	    MLVLFINDENTRY(fath2, child, lvl, *ml);
	    if (fath2>=0) 
	      {
		MLVLFINDENTRY(child, child, lvl+1, *ml);
		/* the I (identity) part, r_l = r_(l+1) */
		mlv[fath2]=mlv[child];
	      }
	  }
      
      /* compute mlv[lvl]= (level interpolation)^T * mlv[lvl+1] */
      /* copy the father nodes */
      for (i=(*ml).levlhindx[lvl+1]; i<(*ml).levlhindx[lvl+2]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).hier[i*hi_w+MCT1HILVL]!=lvl+1)
	    {
	      fprintf(stderr,
		      "mg_restrict_t1: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif

	  for (d=0; d<mldim; d++)
	    {
	      child=(*m).hier[i*hi_w+ MCT1HICHLD  ]+d*vx_nr;
	      fath1=(*m).hier[i*hi_w+ MCT1HIFAT1  ]+d*vx_nr;
	      fath2=(*m).hier[i*hi_w+ MCT1HIFAT1+1]+d*vx_nr;

	      MLVLFINDENTRY(fath1, fath1, lvl, *ml);
	      MLVLFINDENTRY(fath2, fath2, lvl, *ml);
	      MLVLFINDENTRY(child, child, lvl+1,   *ml);

	      /* the interpolation part */
	      mlv[fath1] += 0.5*mlv[child];
	      mlv[fath2] += 0.5*mlv[child]; 
	    }
	} /* end loop hierarchy entries */
    } /* end loop levels */
  
  return SUCCESS;
}


/*FUNCTION*/
int mg_interpolate_t1( struct mgdata *mg,
		       FIDX lvlin, FIDX lvlout, double *mlv
/* interpolates the vector lvlin part of mlv to lvlout, 
   
   Input:  lvlin   - the level from which the interpolation shall start
	   lvlout  - the level to which the interpolation shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin<lvl<=lvlout are
	             overwritten by the interpolation from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		       ){
  FIDX i, j, d, lvl, child, fath1, fath2;
  FIDX hi_w, vx_nr, mldim;

  struct mesh *m;
  struct multilvl *ml;
  
  m     = (*mg).msh;
  ml    = (*mg).ml;

  mldim = (*ml).dim;
  vx_nr = (*mg).vx_nr;

  hi_w  = (*m).hi_w;

  if (vx_nr!=(*m).vx_nr)
    {
      fprintf(stderr, "mg_interpolate_t1: "
	      "mesh doesn't match mgdata (vx_nr test)!\n");
      return FAIL;
    }

  if (lvlout<lvlin)
    {
      fprintf(stderr, "mg_interpolate_t1: "
	      "%d=lvlout<lvlin=%d, makes no sense!\n",
	      (int) lvlout, (int) lvlin);
      return FAIL;
    }

  if ((lvlout>(*ml).lmax)||(lvlin<0))
    {
      fprintf(stderr, "mg_interpolate_t1: "
	      "lvlout=%d or lvlin=%d, out of range (0<lvl<%d=lmax)\n",
	      (int) lvlout, (int) lvlin, (int) (*ml).lmax);
      return FAIL;
    }
  
  /* now, for each lvl, lvlin+1 to lvlout */
  for (lvl=lvlin+1; lvl<=lvlout; lvl++)
    {
      FIDX lm1_vx_nr=((*ml).nlevl[lvl-1]-(*ml).nlevl[lvl])/mldim;
      
      /* initialise the new (finer) level */
      for (j=0; j<lm1_vx_nr; j++)
	for (d=0; d<mldim; d++)
	  {
	    /* get the entry number of the coarser level */
	    fath2 = j+d*vx_nr;
	    MLVLFINDENTRY(child, fath2, lvl, *ml);
	    MLVLFINDENTRY(fath2, fath2, lvl-1, *ml);
	    /* the I (identity) part */
	    mlv[child]=mlv[fath2];
	  }

      /* compute mlv[lvl]= (level interpolation) * mlv[lvl-1] */
      for (i=(*ml).levlhindx[lvl]; i<(*ml).levlhindx[lvl+1]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).hier[i*hi_w+MCT1HILVL]!=lvl)
	    {
	      fprintf(stderr,
		      "mg_interpolate_t1: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  for (d=0; d<mldim; d++)
	    {
	      child=(*m).hier[i*hi_w+MCT1HICHLD  ]+d*vx_nr;
	      fath1=(*m).hier[i*hi_w+MCT1HIFAT1  ]+d*vx_nr;
	      fath2=(*m).hier[i*hi_w+MCT1HIFAT1+1]+d*vx_nr;
      
	      MLVLFINDENTRY(fath1, fath1, lvl-1, *ml);
	      MLVLFINDENTRY(fath2, fath2, lvl-1, *ml);
	      MLVLFINDENTRY(child, child, lvl,   *ml);

	      /* the interpolation part */
	      mlv[child] =0.5*( mlv[fath1] + mlv[fath2] );
	    }
	} /* end loop hierarchy entries */
    } /* end loop levels */
  
  return SUCCESS;
}






/*FUNCTION*/
int mg_restrict_tx( struct mgdata *mg, FIDX lvlin, FIDX lvlout, double *mlv
/* restricts the vector lvlin part of mlv to lvlout, by applying the
   transpose operation of interpolation
   
   Input:  lvlin   - the level from which the restriction shall start
	   lvlout  - the level to which the restriction shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin>lvl>=lvlout are
	             overwritten by the restriction from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		    ){
  int err;

  if ( (*(*mg).ml).type == 1 )
    {
      err = mg_restrict_t1( mg, lvlin, lvlout, mlv);
      FUNCTION_FAILURE_HANDLE(err, mg_restrict_t1, mg_restrict_tx);
    }
  else if ( (*(*mg).ml).type == 2 )
    {
      err = mg_restrict_t2( mg, lvlin, lvlout, mlv);
      FUNCTION_FAILURE_HANDLE(err, mg_restrict_t2, mg_restrict_tx);
    }
  else
    {
      fprintf(stderr,"mg_restrict_tx: unknown multilevel type!\n");
      return FAIL;
    }

  return SUCCESS;
}


/*FUNCTION*/
int mg_interpolate_tx( struct mgdata *mg, FIDX lvlin, FIDX lvlout,
		       double *mlv
/* interpolates the vector lvlin part of mlv to lvlout
   
   Input:  lvlin   - the level from which the interpolation shall start
	   lvlout  - the level to which the interpolation shall take
	             place, lvlout >= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin>lvl>=lvlout are
	             overwritten by the interpolateion from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		    ){
  int err;

  if ( (*(*mg).ml).type == 1 )
    {
      err = mg_interpolate_t1( mg, lvlin, lvlout, mlv);
      FUNCTION_FAILURE_HANDLE(err, mg_interpolate_t1, mg_interpolate_tx);
    }
  else if ( (*(*mg).ml).type == 2 )
    {
      err = mg_interpolate_t2( mg, lvlin, lvlout, mlv);
      FUNCTION_FAILURE_HANDLE(err, mg_interpolate_t2, mg_interpolate_tx);
    }
  else
    {
      fprintf(stderr,"mg_interpolate_tx: unknown multilevel type!\n");
      return FAIL;
    }

  return SUCCESS;
}






/****************************************************************************/












/*FUNCTION*/
int mg_restrict_t1_t2( struct mgdata *mg,
		       FIDX lvlin, FIDX lvlout, double *mlv
/* restricts the vector lvlin part of mlv to lvlout, by applying the
   transpose operation of interpolation,
   using the T1 hiriarchy of the T1 mesh associated with the T2 mesh

   
   Input:  lvlin   - the level from which the restriction shall start
	   lvlout  - the level to which the restriction shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin>lvl>=lvlout are
	             overwritten by the restriction from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		){
  FIDX d, i, j, k;
  FIDX dim, vx_nr, eh_w, eh_nr;
  FIDX node, fath, fathl, fathlp1, chldlp1;
  FIDX lvl;

  FIDX *done;

  struct mesh *m;
  struct multilvl *ml;
  
  m     = (*mg).msh;
  ml    = (*mg).ml;

  vx_nr = (*mg).vx_nr;
  dim   = (*mg).dim;
  eh_w  = (*m).eh_w;
  eh_nr = (*m).eh_nr;
  done  = (*mg).done;

  if (vx_nr!=(*m).vx_nr)
    {
      fprintf(stderr, "mg_restrict_t1_t2: "
	      "mesh doesn't match mgdata (vx_nr test)!\n");
      return FAIL;
    }

  if (lvlout>lvlin)
    {
      fprintf(stderr,
	      "mg_restrict_t1_t2: %d=lvlout>lvlin=%d, makes no sense!\n",
	      (int) lvlout, (int) lvlin);
      return FAIL;
    }

  if ((lvlin>(*ml).lmax)||(lvlout<0))
    {
      fprintf(stderr, "mg_restrict_t1_t2: "
	      "lvlout=%d or lvlin=%d, out of range (0<lvl<%d=maxl)\n",
	      (int) lvlout, (int) lvlin, (int) (*ml).lmax);
      return FAIL;
    }
  
      
  /* now, for each lvl, lvlin-1 to lvlout */
  for (lvl=lvlin-1; lvl>=lvlout; lvl--)
    {
      /* mark all nodes as undone */
      for (i=0; i<dim*vx_nr; i++) done[i]=0;

      /* compute mlv[lvl]= (level interpolation)^T * mlv[lvl+1] */
      /* copy the father nodes */
      for (i=(*ml).levlhindx[lvl+1]; i<(*ml).levlhindx[lvl+2]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).elhier[i*eh_w+MCT2EHLVL]!=lvl+1)
	    {
	      fprintf(stderr,
		      "mg_restrict_t2_t1: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  for (j=0; j<3; j++)
	    for (d=0; d<dim; d++)
	      {
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(fathl, node, lvl, *ml);
		    MLVLFINDENTRY(fathlp1, node, lvl+1, *ml);
		    /* printf("node=%3d   fathl=%3d    fathlp1=%3d\n",
		       node, fathl, fathlp1); /* */
		    mlv[fathl]=mlv[fathlp1];
		  }
	      }
	}
      for (i=(*ml).levlhindx[lvl+1]; i<(*ml).levlhindx[lvl+2]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).elhier[i*eh_w+MCT2EHLVL]!=lvl+1)
	    {
	      fprintf(stderr,
		      "mg_restrict_t2_t1: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  for (j=0; j<3; j++)
	    for (d=0; d<dim; d++)
	      {
		/* check out the child nodes */
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+3+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(chldlp1, node, lvl+1, *ml);
		    for (k=0; k<2; k++)
		      {
			/* father += alpha*child */
			fath=d*vx_nr
			  +(*m).elhier[i*eh_w+MCT2EHFAT1+
				       (j+k)%3];
			MLVLFINDENTRY(fathl, fath, lvl, *ml);
			mlv[fathl]+= 0.5*mlv[chldlp1];
		      }
		  }
		/* end check out child nodes */
	      }
	}
    }/* end of loop over levels */

  return SUCCESS;
}


/*FUNCTION*/
int mg_interpolate_t1_t2( struct mgdata *mg,
			  FIDX lvlin, FIDX lvlout, double *mlv
/* interpolates the vector lvlin part of mlv to lvlout, 
   using the T1 hiriarchy of the T1 mesh associated with the T2 mesh
   
   Input:  lvlin   - the level from which the interpolation shall start
	   lvlout  - the level to which the interpolation shall take
	             place, lvlout <= lvlin


   In/Out: mg      - multigrid data, the done part is used and therby
                     overwritten, other parts may be used read only
                     (phih)
	   mlv     - a multilevel vector, the lvlin part is read, and
	             all levels lvl with lvlin<lvl<=lvlout are
	             overwritten by the interpolation from lvlin

   Return: SUCCESS - success,
           FAIL    - failure, see error message
*/
		){
  FIDX d, i, j, k;
  FIDX dim, vx_nr, eh_w, eh_nr;
  FIDX node, fath, fathl, fathlm1, chldl;
  FIDX lvl;

  FIDX *done;

  struct mesh *m;
  struct multilvl *ml;
  
  m     = (*mg).msh;
  ml    = (*mg).ml;

  vx_nr = (*mg).vx_nr;
  dim   = (*mg).dim;
  eh_w  = (*m).eh_w;
  eh_nr = (*m).eh_nr;
  done  = (*mg).done;


  if (vx_nr!=(*m).vx_nr)
    {
      fprintf(stderr, "mg_interpolate_t1_t2: "
	      "mesh doesn't match mgdata (vx_nr test)!\n");
      return FAIL;
    }

  if (lvlout<lvlin)
    {
      fprintf(stderr, "mg_interpolate_t1_t2: "
	      "%d=lvlout<lvlin=%d, makes no sense!\n",
	      (int) lvlout, (int) lvlin);
      return FAIL;
    }

  if ((lvlout>(*ml).lmax)||(lvlin<0))
    {
      fprintf(stderr, "mg_interpolate_t1_t2: "
	      "lvlout=%d or lvlin=%d, out of range (0<lvl<%d=lmax)\n",
	      (int) lvlout, (int) lvlin, (int) (*ml).lmax);
      return FAIL;
    }
  
  /* now, for each lvl, lvlin+1 to lvlout */
  for (lvl=lvlin+1; lvl<=lvlout; lvl++)
    {
      /* mark all nodes as undone */
      for (i=0; i<dim*vx_nr; i++) done[i]=0;

      /* compute mlv[lvl]= (level interpolation) * mlv[lvl-1] */
      for (i=(*ml).levlhindx[lvl]; i<(*ml).levlhindx[lvl+1]; i++)
	{
#ifdef DEBUFEINS
	  if ((*m).elhier[i*eh_w+MCT2EHLVL]!=lvl)
	    {
	      fprintf(stderr,
		      "mg_interpolate_t1_t2: ml.levlhindx wrong?\n");
	      return FAIL;
	    }
#endif
	  /* copy the father nodes */
	  for (j=0; j<3; j++)
	    for (d=0; d<dim; d++)
	      {
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(fathl, node, lvl, *ml);
		    MLVLFINDENTRY(fathlm1, node, lvl-1, *ml);
		    mlv[fathl]=mlv[fathlm1];
		  }
	      }
	  for (j=0; j<3; j++)
	    for (d=0; d<dim; d++)
	      {
		/* check out the child nodes */
		node=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1+3+j];
		if (done[node]==0)
		  {
		    done[node]=1;
		    MLVLFINDENTRY(chldl, node, lvl, *ml);
		    mlv[chldl]=0.0;
		    for (k=0; k<2; k++)
		      {
			/* father += alpha*child */
			fath=d*vx_nr+(*m).elhier[i*eh_w+MCT2EHFAT1
						 +(j+k)%3];
			MLVLFINDENTRY(fathlm1, fath, lvl-1, *ml);
			mlv[chldl]+= 0.5 * mlv[fathlm1];
		      }
		  }
		/* end check out child nodes */
	      }
	}
    }/* end of loop over levels */

  return SUCCESS;
}














/****************************************************************************/





















/*FUNCTION*/
int mesh_write_solution_exp_t2( struct mesh *m, struct vector *x,
				FIDX ncompo,
				FIDX namlen, char *name
/* writes the mesh and the solution into a NAG IRIS EXPLORER readable
   file (for visualisation) 
   
   Input:  m         - the mesh
           x         - vector with data associated to the nodes,
                       usually only input, but if (xtype==2) then the
                       pressure values are corrected, such that the
		       edge-mid-nodes have propper values
	   ncompo    - dimension of the solution vector per node which
	               shall be writen to the file, (or number of
	               solution components), the components of the
	               vector x have to be numbered such that
	               x[j*vx_nr+i] gives the j-th component in the
	               i-th node of the mesh
	   namlen    - maximal useable length of name
	   name      - basename of the files, one will be
	               <name>.pyr and the other <name>.lat
           
   Output: (writes the files)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
				){
  int err;
  struct mesh m_t1;

  if ((*x).len!=ncompo*(*m).vx_nr)
    {
      /* cry */
      fprintf(stderr, "mesh_write_solution_exp_t2: "
	      "size of x doesn't match ncompo=%d and vx_nr=%d!\n",
	      (int) ncompo, (int) (*m).vx_nr );
      return FAIL;
    }

  mesh_init(& m_t1);

  err=mesh_t2_to_t1( m, &m_t1 );
  FUNCTION_FAILURE_HANDLE( err, mesh_t2_to_t1,
			   mesh_write_solution_exp_t2);

  err=mesh_write_solution_exp_t1( &m_t1, x, ncompo, namlen, name );
  FUNCTION_FAILURE_HANDLE( err, mesh_write_solution_exp_t1,
			   mesh_write_solution_exp_t2);

  mesh_free( &m_t1 );

  return SUCCESS;
}



/*FUNCTION*/
int mesh_t1_to_t2( struct mesh *m1, struct mesh *m2
/* creates the T2 mesh by converting the T1 mesh m1 into a T2 mesh,
   Input:  m1         - T1 mesh
   Output: m2         - T2 mesh, (empty mesh given via reference,
                        interior is set up)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX somany, i, j, len, insane;
  int  err;

  FIDX vx_w1, el_w1, eg_w1, fc_w1, hi_w1, bd_w1,
    ps_w1, pv_w1, pc_w1, fu_w1, pa_w1, sg_w1, sp_w1;

  const FIDX   vx_w2 = MCT2VXLN, eg_w2 = MCT2EGLN, 
    fc_w2 = MCT2FCLN, el_w2 = MCT2ELLN, vo_w2 = MCT2VOLN,
    eh_w2 = MCT2EHLN, hi_w2 = MCT2HILN, 
    bd_w2 = MCT2BDLN, ps_w2 = MCXXPSLN, pv_w2 = MCXXPVLN,
    pc_w2 = MCXXPCLN, fu_w2 = MC2XFULN, pa_w2 = MC2XPALN,
    sg_w2 = MC2XSGLN, sp_w2 = MC2XSPLN;

  /* set dimension */
  (*m2).dim = (*m1).dim;

  /* set width of fields */
  (*m2).vx_w = vx_w2;
  (*m2).el_w = el_w2;
  (*m2).eg_w = eg_w2;
  (*m2).fc_w = fc_w2;
  (*m2).vo_w = vo_w2;
  (*m2).hi_w = hi_w2;
  (*m2).eh_w = eh_w2;
  (*m2).bd_w = bd_w2;
  (*m2).ps_w = ps_w2;
  (*m2).pv_w = pv_w2;
  (*m2).pc_w = pc_w2;
  (*m2).fu_w = fu_w2;
  (*m2).pa_w = pa_w2;
  (*m2).sg_w = sg_w2;
  (*m2).sp_w = sp_w2;

  vx_w1 = (*m1).vx_w; el_w1 = (*m1).el_w; eg_w1 = (*m1).eg_w;
  fc_w1 = (*m1).fc_w; hi_w1 = (*m1).hi_w; bd_w1 = (*m1).bd_w;
  ps_w1 = (*m1).ps_w; pv_w1 = (*m1).pv_w; pc_w1 = (*m1).pc_w;
  fu_w1 = (*m1).fu_w; pa_w1 = (*m1).pa_w; sg_w1 = (*m1).sg_w;
  sp_w1 = (*m1).sp_w;

  /* initialise arrays */
  somany=(*m1).vx_nr+(*m1).eg_nr+100;
  TRY_MALLOC( (*m2).vertex, (*m2).vx_w * somany , double,
	      mesh_t1_to_t2 );
  (*m2).vx_max=somany;
  (*m2).vx_nr=0;

  somany=(*m1).el_nr+100;
  TRY_MALLOC( (*m2).elem, (*m2).el_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).el_max=somany;
  (*m2).el_nr=0;

  somany=(*m1).eg_nr+100;
  TRY_MALLOC( (*m2).edge, (*m2).eg_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).eg_max=somany;
  (*m2).eg_nr=0;

  somany=(*m1).fc_nr+100;
  TRY_MALLOC( (*m2).face, (*m2).fc_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).fc_max=somany;
  (*m2).fc_nr=0;

  (*m2).vols= NULL;
  (*m2).vo_max=0;
  (*m2).vo_nr=0;

  somany=(*m1).hi_nr+(*m1).eg_nr+100;
  TRY_MALLOC( (*m2).hier, (*m2).hi_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).hi_max=somany;
  (*m2).hi_nr=0;

  somany=(*m1).fc_nr;
  TRY_MALLOC( (*m2).elhier, (*m2).eh_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).eh_max=somany;
  (*m2).eh_nr=0;

  somany=(*m1).bd_nr+100;
  TRY_MALLOC( (*m2).bound, (*m2).bd_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).bd_max=somany;
  (*m2).bd_nr=0;

  somany=(*m1).ps_nr+100;
  TRY_MALLOC( (*m2).pcsurf, (*m2).ps_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).ps_max=somany;
  (*m2).ps_nr=0;

  somany=(*m1).pv_nr+100;
  TRY_MALLOC( (*m2).pcvol, (*m2).pv_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).pv_max=somany;
  (*m2).pv_nr=0;

  somany=(*m1).pc_nr+100;
  TRY_MALLOC( (*m2).pccrit, (*m2).pc_w * somany , double,
	      mesh_t1_to_t2 );
  (*m2).pc_max=somany;
  (*m2).pc_nr=0;

  somany=(*m1).fu_nr+8;
  TRY_MALLOC( (*m2).func, (*m2).fu_w * somany , double,
	      mesh_t1_to_t2 );
  (*m2).fu_max=somany;
  (*m2).fu_nr=0;

  somany=(*m1).pa_nr;
  TRY_MALLOC( (*m2).para, (*m2).pa_w * somany , double,
	      mesh_t1_to_t2 );
  (*m2).pa_max=somany;
  (*m2).pa_nr=0;

  somany=(*m1).sg_nr;
  TRY_MALLOC( (*m2).sseg, (*m2).sg_w * somany , FIDX,
	      mesh_t1_to_t2 );
  (*m2).sg_max=somany;
  (*m2).sg_nr=0;

  somany=(*m1).sp_nr;
  TRY_MALLOC( (*m2).spar, (*m2).sp_w * somany , double,
	      mesh_t1_to_t2 );
  (*m2).sp_max=somany;
  (*m2).sp_nr=0;

  /* for debuging purposes, initialise the arrays with insane data */
  insane=-1523123;

#ifdef DEBUGFEINS
  len=(*m2).vx_w*(*m2).vx_max;
  for (i=0;i<len; i++)
    (*m2).vertex[i]=(double)insane;

  len=(*m2).el_w*(*m2).el_max;
  for (i=0;i<len; i++)
    (*m2).elem[i]=insane;

  len=(*m2).eg_w*(*m2).eg_max;
  for (i=0;i<len; i++)
    (*m2).edge[i]=insane;

  len=(*m2).fc_w*(*m2).fc_max;
  for (i=0;i<len; i++)
    (*m2).face[i]=insane;

  len=(*m2).hi_w*(*m2).hi_max;
  for (i=0;i<len; i++)
    (*m2).hier[i]=insane;

  len=(*m2).eh_w*(*m2).eh_max;
  for (i=0;i<len; i++)
    (*m2).elhier[i]=insane;

  len=(*m2).bd_w*(*m2).bd_max;
  for (i=0;i<len; i++)
    (*m2).bound[i]=insane;

  len=(*m2).fu_w*(*m2).fu_max;
  for (i=0;i<len; i++)
    (*m2).func[i]=(double)insane;

  len=(*m2).pa_w*(*m2).pa_max;
  for (i=0;i<len; i++)
    (*m2).para[i]=(double)insane;

  len=(*m2).sg_w*(*m2).sg_max;
  for (i=0;i<len; i++)
    (*m2).sseg[i]=insane;

  len=(*m2).sp_w*(*m2).sp_max;
  for (i=0;i<len; i++)
    (*m2).spar[i]=(double)insane;
#endif
  
  /* now copy the mesh */

  /* vertices */
  for(i=0;i<(*m1).vx_nr;i++)
    {
      (*m2).vertex[i*vx_w2+MCT2VXSTRT  ] =
	(*m1).vertex[i*vx_w1+MCT1VXSTRT  ];
      (*m2).vertex[i*vx_w2+MCT2VXSTRT+1] =
	(*m1).vertex[i*vx_w1+MCT1VXSTRT+1];
    }
  (*m2).vx_nr=(*m1).vx_nr;

  /* hierarchy */
  for (i=0; i<(*m1).hi_nr; i++)
    {
      (*m2).hier[i*hi_w2 + MCT2HICHLD] =
	(*m1).hier[i*hi_w1 + MCT1HICHLD];

      (*m2).hier[i*hi_w2 + MCT2HIFAT1  ] =
	(*m1).hier[i*hi_w1 + MCT1HIFAT1  ];
      (*m2).hier[i*hi_w2 + MCT2HIFAT1+1] =
	(*m1).hier[i*hi_w1 + MCT1HIFAT1+1];

      (*m2).hier[i*hi_w2 + MCT2HILVL] =
	(*m1).hier[i*hi_w1 + MCT1HILVL];
    }
  (*m2).hi_nr=(*m1).hi_nr;

  /* no element hierarchy to be copied */

  /* edges */
  /* fist copy as much as possible */
  for (i=0; i<(*m1).eg_nr; i++)
    {
      /* both end nodes of the edge */
      (*m2).edge[i*eg_w2 + MCT2EGNOD1   ] =
	(*m1).edge[i*eg_w1 + MCT1EGNOD1   ];
      (*m2).edge[i*eg_w2 + MCT2EGNOD1+1 ] =
	(*m1).edge[i*eg_w1 + MCT1EGNOD1+1 ];
      /* child data */
      (*m2).edge[i*eg_w2 + MCT2EGCHL1   ] =
	(*m1).edge[i*eg_w1 + MCT1EGCHL1   ];
      (*m2).edge[i*eg_w2 + MCT2EGCHL1+1 ] =
	(*m1).edge[i*eg_w1 + MCT1EGCHL1+1 ];
      /* boundary data */
      (*m2).edge[i*eg_w2 + MCT2EGBND    ] =
	(*m1).edge[i*eg_w1 + MCT1EGBND    ];
      /* pcsurf data */
      (*m2).edge[i*eg_w2 + MCT2EGPCSU   ] =
	(*m1).edge[i*eg_w1 + MCT1EGPCSU   ];
      /* level data */
      (*m2).edge[i*eg_w2 + MCT2EGLVL    ] =
	(*m1).edge[i*eg_w1 + MCT1EGLVL    ];
    }
  (*m2).eg_nr=(*m1).eg_nr;
  /* now build the midnodes and the hierarchy entires for them */
  for (i=0; i<(*m2).eg_nr; i++)
    {
      /* if the T1 edge had no child we have to create the midnode */
      if ((*m2).edge[i*eg_w2 + MCT2EGCHL1]==-1)
	{
	  FIDX old1, old2, newnode, newhier, level;

	  old1= (*m2).edge[i*eg_w2 + MCT2EGNOD1  ];
	  old2= (*m2).edge[i*eg_w2 + MCT2EGNOD1+1];

	  /* level of the edge */
	  level= (*m2).edge[i*eg_w2 + MCT2EGLVL];

	  /* create new vertex, enough space is assured (vx_nr1+eg_nr1) */
	  newnode=(*m2).vx_nr;
	  (*m2).vx_nr++;

	  (*m2).vertex[newnode*vx_w2+MCT2VXSTRT  ]= 0.5*
	    ( (*m2).vertex[old1*vx_w2+MCT2VXSTRT  ]
	      + (*m2).vertex[old2*vx_w2+MCT2VXSTRT  ]);
	  (*m2).vertex[newnode*vx_w2+MCT2VXSTRT+1]= 0.5*
	    ( (*m2).vertex[old1*vx_w2+MCT2VXSTRT+1]
	      + (*m2).vertex[old2*vx_w2+MCT2VXSTRT+1]);

	  (*m2).edge[i*eg_w2 + MCT2EGNODM]=newnode;

	  /* create new hierarchy entry */
	  newhier=(*m2).hi_nr;
	  (*m2).hi_nr++;

	  (*m2).hier[newhier*hi_w2+MCT2HICHLD  ]=newnode;
	  (*m2).hier[newhier*hi_w2+MCT2HIFAT1  ]=old1;
	  (*m2).hier[newhier*hi_w2+MCT2HIFAT1+1]=old2;
	  (*m2).hier[newhier*hi_w2+MCT2HILVL   ]=level+1;
	}
      else
	{
	  FIDX chld;
	  /* the midnode is the second node of the child edge */
	  chld=(*m2).edge[i*eg_w2 + MCT2EGCHL1];
	  (*m2).edge[i*eg_w2 + MCT2EGNODM]=
	    (*m2).edge[chld*eg_w2 + MCT2EGNOD1+1];
	}
    }
  
  /* faces */
  /* just copy, no difference */
  for (i=0; i<(*m1).fc_nr; i++)
    {
      for (j=0; j<3; j++)
	{
	  (*m2).face[i*fc_w2 + MCT2FCEDG1+j]=
	    (*m1).face[i*fc_w1 + MCT1FCEDG1+j];
	}
      (*m2).face[i*fc_w2 + MCT2FCRHSF]=
	(*m1).face[i*fc_w1 + MCT1FCRHSF];
      (*m2).face[i*fc_w2 + MCT2FCPCVL]=
	(*m1).face[i*fc_w1 + MCT1FCPCVL];
    }
  (*m2).fc_nr=(*m1).fc_nr;

  /* no volumes */

  /* boundary */
  /* just copy */
  for (i=0; i<(*m1).bd_nr; i++)
    {
      (*m2).bound[i*bd_w2 + MCT2BDEDGE]=
	(*m1).bound[i*bd_w1 + MCT1BDEDGE];
      (*m2).bound[i*bd_w2 + MCT2BDTYPE]=
	(*m1).bound[i*bd_w1 + MCT1BDTYPE];
      (*m2).bound[i*bd_w2 + MCT2BDFNCT]=
	(*m1).bound[i*bd_w1 + MCT1BDFNCT];
      (*m2).bound[i*bd_w2 + MCT2BDORIE]=
	(*m1).bound[i*bd_w1 + MCT1BDORIE];
      (*m2).bound[i*bd_w2 + MCT2BDSSEG]=
	(*m1).bound[i*bd_w1 + MCT1BDSSEG];
    }
  (*m2).bd_nr=(*m1).bd_nr;


  /* pcsurf */
  /* just copy */
  for (i=0; i<(*m1).ps_nr; i++)
    {
      for (j=0; j<MCXXPSLN; j++)
	{
	  (*m2).pcsurf[i*ps_w2 + j] = (*m1).pcsurf[i*ps_w1 + j];
	}
    }
  (*m2).ps_nr=(*m1).ps_nr;

  /* pcvol */
  /* just copy */
  for (i=0; i<(*m1).pv_nr; i++)
    {
      for (j=0; j<MCXXPVLN; j++)
	{
	  (*m2).pcvol[i*pv_w2 + j] = (*m1).pcvol[i*pv_w1 + j];
	}
    }
  (*m2).pv_nr=(*m1).pv_nr;

  /* pccrit */
  /* just copy */
  for (i=0; i<(*m1).pc_nr; i++)
    {
      for (j=0; j<MCXXPCLN; j++)
	{
	  (*m2).pccrit[i*pc_w2 + j] = (*m1).pccrit[i*pc_w1 + j];
	}
    }
  (*m2).pc_nr=(*m1).pc_nr;

  /* function, just copy */
  for (i=0; i<(*m1).fu_nr; i++)
    {
      for (j=0; j<MC2XFULN; j++)
	{
	  (*m2).func[i*fu_w2 + j] = (*m1).func[i*fu_w1 + j];
	}
    }
  (*m2).fu_nr=(*m1).fu_nr;


  /* parameter, just copy */
  for (i=0; i<(*m1).pa_nr; i++)
    {
      for (j=0; j<MC2XPALN; j++)
	{
	  (*m2).para[i*pa_w2 + j] = (*m1).para[i*pa_w1 + j];
	}
    }
  (*m2).pa_nr=(*m1).pa_nr;

  /* shape segments, just copy */
  for (i=0; i<(*m1).sg_nr; i++)
    {
      for (j=0; j<MC2XSGLN; j++)
	{
	  (*m2).sseg[i*sg_w2 + j] = (*m1).sseg[i*sg_w1 + j];
	}
    }
  (*m2).sg_nr=(*m1).sg_nr;


  /* shape parameters, just copy */
  for (i=0; i<(*m1).sp_nr; i++)
    {
      for (j=0; j<MC2XSPLN; j++)
	{
	  (*m2).spar[i*sp_w2 + j] = (*m1).spar[i*sp_w1 + j];
	}
    }
  (*m2).sp_nr=(*m1).sp_nr;


  /* elements */
  for(i=0; i<(*m1).el_nr; i++)
    {
      err=mesh_t2_get_elem_from_face( m2, i );
      FUNCTION_FAILURE_HANDLE( err, mesh_t2_get_elem_from_face,
			       mesh_t1_to_t2);
    }
  (*m2).el_nr = (*m1).el_nr;


  return SUCCESS;
}





/*FUNCTION*/
int mesh_t2_to_t1( struct mesh *m2, struct mesh *m1
/* creates the T1 mesh by converting the T2 mesh m2 into a T1 mesh,
   where one element/face in m2 creates 4 elements/faces in m1, such
   that they have the same number of vertices

   Input:  m2         - T2 mesh
   Output: m1         - T1 mesh, (empty mesh given via reference,
                        interior is set up)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX somany, i, j, elj;

  FIDX   vx_w2, el_w2, eg_w2, fc_w2, hi_w2, bd_w2,
    ps_w2, pv_w2, pc_w2, fu_w2, pa_w2, sg_w2, sp_w2;

  const FIDX   vx_w1 = MCT1VXLN, eg_w1 = MCT1EGLN, 
    fc_w1 = MCT1FCLN, el_w1 = MCT1ELLN, vo_w1 = MCT1VOLN,
    eh_w1 = MCT1EHLN, hi_w1 = MCT1HILN, 
    bd_w1 = MCT1BDLN, ps_w1 = MCXXPSLN, pv_w1 = MCXXPVLN,
    pc_w1 = MCXXPCLN, fu_w1 = MC2XFULN, pa_w1 = MC2XPALN,
    sg_w1 = MC2XSGLN, sp_w1 = MC2XSPLN;

  /* set dimension */
  (*m1).dim = (*m2).dim;

  /* set width of fields */
  (*m1).vx_w = vx_w1;
  (*m1).el_w = el_w1;
  (*m1).eg_w = eg_w1;
  (*m1).fc_w = fc_w1;
  (*m1).vo_w = vo_w1;
  (*m1).hi_w = hi_w1;
  (*m1).eh_w = eh_w1;
  (*m1).bd_w = bd_w1;
  (*m1).ps_w = ps_w1;
  (*m1).pv_w = pv_w1;
  (*m1).pc_w = pc_w1;
  (*m1).fu_w = fu_w1;
  (*m1).pa_w = pa_w1;
  (*m1).sg_w = sg_w1;
  (*m1).sp_w = sp_w1;

  vx_w2 = (*m2).vx_w; el_w2 = (*m2).el_w; eg_w2 = (*m2).eg_w;
  fc_w2 = (*m2).fc_w; hi_w2 = (*m2).hi_w; bd_w2 = (*m2).bd_w;
  ps_w2 = (*m2).ps_w; pv_w2 = (*m2).pv_w; pc_w2 = (*m2).pc_w;
  fu_w2 = (*m2).fu_w; pa_w2 = (*m2).pa_w; sg_w2 = (*m2).sg_w;
  sp_w2 = (*m2).sp_w;

  /* initialise arrays */
  somany=(*m2).vx_nr;
  TRY_MALLOC( (*m1).vertex, (*m1).vx_w * somany , double,
	      mesh_t2_to_t1 );
  (*m1).vx_max=somany;
  (*m1).vx_nr=0;

  somany=4*(*m2).el_nr; /* one face to 4 */
  TRY_MALLOC( (*m1).elem, (*m1).el_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).el_max=somany;
  (*m1).el_nr=0;

  somany=3*(*m2).eg_nr+3*(*m2).fc_nr; /* all edges without children
					 have to be split + each face
					 generates 3 child edges */
  TRY_MALLOC( (*m1).edge, (*m1).eg_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).eg_max=somany;
  (*m1).eg_nr=0;

  somany=4*(*m2).fc_nr; /* one face to 4 */
  TRY_MALLOC( (*m1).face, (*m1).fc_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).fc_max=somany;
  (*m1).fc_nr=0;

  (*m1).vols= NULL;
  (*m1).vo_max=0;
  (*m1).vo_nr=0;

  somany=(*m2).hi_nr+(*m2).eg_nr;
  TRY_MALLOC( (*m1).hier, (*m1).hi_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).hi_max=somany;
  (*m1).hi_nr=0;

  somany=3*(*m2).bd_nr; /* due to split edges */
  TRY_MALLOC( (*m1).bound, (*m1).bd_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).bd_max=somany;
  (*m1).bd_nr=0;

  somany=3*(*m2).ps_nr; /* due to split edges */
  TRY_MALLOC( (*m1).pcsurf, (*m1).ps_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).ps_max=somany;
  (*m1).ps_nr=0;

  somany=4*(*m2).pv_nr; /* one face to 4 */
  TRY_MALLOC( (*m1).pcvol, (*m1).pv_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).pv_max=somany;
  (*m1).pv_nr=0;

  somany=(*m2).pc_nr;
  TRY_MALLOC( (*m1).pccrit, (*m1).pc_w * somany , double,
	      mesh_t2_to_t1 );
  (*m1).pc_max=somany;
  (*m1).pc_nr=0;

  somany=(*m2).fu_nr;
  TRY_MALLOC( (*m1).func, (*m1).fu_w * somany , double,
	      mesh_t2_to_t1 );
  (*m1).fu_max=somany;
  (*m1).fu_nr=0;

  somany=(*m2).pa_nr;
  TRY_MALLOC( (*m1).para, (*m1).pa_w * somany , double,
	      mesh_t2_to_t1 );
  (*m1).pa_max=somany;
  (*m1).pa_nr=0;

  somany=(*m2).sg_nr;
  TRY_MALLOC( (*m1).sseg, (*m1).sg_w * somany , FIDX,
	      mesh_t2_to_t1 );
  (*m1).sg_max=somany;
  (*m1).sg_nr=0;

  somany=(*m2).sp_nr;
  TRY_MALLOC( (*m1).spar, (*m1).sp_w * somany , double,
	      mesh_t2_to_t1 );
  (*m1).sp_max=somany;
  (*m1).sp_nr=0;

  /* now copy the mesh */

  /* vertices */
  for(i=0;i<(*m2).vx_nr;i++)
    {
      (*m1).vertex[i*vx_w1+MCT1VXSTRT  ] =
	(*m2).vertex[i*vx_w2+MCT2VXSTRT  ];
      (*m1).vertex[i*vx_w1+MCT1VXSTRT+1] =
	(*m2).vertex[i*vx_w2+MCT2VXSTRT+1];
    }
  (*m1).vx_nr=(*m2).vx_nr;


  /* hierarchy */
  for (i=0; i<(*m2).hi_nr; i++)
    {
      (*m1).hier[i*hi_w1 + MCT1HICHLD] =
	(*m2).hier[i*hi_w2 + MCT2HICHLD];

      (*m1).hier[i*hi_w1 + MCT1HIFAT1  ] =
	(*m2).hier[i*hi_w2 + MCT2HIFAT1  ];
      (*m1).hier[i*hi_w1 + MCT1HIFAT1+1] =
	(*m2).hier[i*hi_w2 + MCT2HIFAT1+1];

      (*m1).hier[i*hi_w1 + MCT1HILVL] =
	(*m2).hier[i*hi_w2 + MCT2HILVL];
    }
  (*m1).hi_nr=(*m2).hi_nr;

  /* boundary, just copy */
  for (i=0; i<(*m2).bd_nr; i++)
    {
      (*m1).bound[i*bd_w1 + MCT1BDEDGE]=
	(*m2).bound[i*bd_w2 + MCT2BDEDGE];
      (*m1).bound[i*bd_w1 + MCT1BDTYPE]=
	(*m2).bound[i*bd_w2 + MCT2BDTYPE];
      (*m1).bound[i*bd_w1 + MCT1BDFNCT]=
	(*m2).bound[i*bd_w2 + MCT2BDFNCT];
      (*m1).bound[i*bd_w1 + MCT1BDORIE]=
	(*m2).bound[i*bd_w2 + MCT2BDORIE];
      (*m1).bound[i*bd_w1 + MCT1BDSSEG]=
	(*m2).bound[i*bd_w2 + MCT2BDSSEG];
    }
  (*m1).bd_nr=(*m2).bd_nr;

  /* pcsurf, just copy */
  for (i=0; i<(*m2).ps_nr; i++)
    {
      for (j=0; j<MCXXPSLN; j++)
	{
	  (*m1).pcsurf[i*ps_w1 + j]= (*m2).pcsurf[i*ps_w2 + j];
	}
    }
  (*m1).ps_nr=(*m2).ps_nr;

  /* edges */
  /* fist copy as much as possible */
  for (i=0; i<(*m2).eg_nr; i++)
    {
      /* both end nodes of the edge */
      (*m1).edge[i*eg_w1 + MCT1EGNOD1   ] =
	(*m2).edge[i*eg_w2 + MCT2EGNOD1   ];
      (*m1).edge[i*eg_w1 + MCT1EGNOD1+1 ] =
	(*m2).edge[i*eg_w2 + MCT2EGNOD1+1 ];
      /* child data */
      (*m1).edge[i*eg_w1 + MCT1EGCHL1   ] =
	(*m2).edge[i*eg_w2 + MCT2EGCHL1   ];
      (*m1).edge[i*eg_w1 + MCT1EGCHL1+1 ] =
	(*m2).edge[i*eg_w2 + MCT2EGCHL1+1 ];
      /* boundary data */
      (*m1).edge[i*eg_w1 + MCT1EGBND    ] =
	(*m2).edge[i*eg_w2 + MCT2EGBND    ];
      /* pcsurf data */
      (*m1).edge[i*eg_w1 + MCT1EGPCSU   ] =
	(*m2).edge[i*eg_w2 + MCT2EGPCSU   ];
      /* level data */
      (*m1).edge[i*eg_w1 + MCT1EGLVL    ] =
	(*m2).edge[i*eg_w2 + MCT2EGLVL    ];
    }
  (*m1).eg_nr=(*m2).eg_nr;
  /* now handle the midnodes */
  for (i=0; i<(*m2).eg_nr; i++)
    {
      /* if the T2 edge had no child we have to create two */
      if ((*m2).edge[i*eg_w2 + MCT2EGCHL1]==-1)
	{
	  FIDX newe1, newe2, level;
	  newe1=(*m1).eg_nr; (*m1).eg_nr++;
	  newe2=(*m1).eg_nr; (*m1).eg_nr++;

	  level=(*m2).edge[i*eg_w2 + MCT2EGLVL];

	  (*m1).edge[newe1*eg_w1+MCT1EGNOD1  ]=
	    (*m2).edge[i*eg_w2 + MCT2EGNOD1  ]; /* 1st child = 1st
						   node */
	  (*m1).edge[newe1*eg_w1+MCT1EGNOD1+1]=
	    (*m2).edge[i*eg_w2 + MCT2EGNODM  ]; /* 2n node = midnode */

	  (*m1).edge[newe2*eg_w1+MCT1EGNOD1  ]=
	    (*m2).edge[i*eg_w2 + MCT2EGNOD1+1]; /* 2nd child = 2nd
						   node */
	  (*m1).edge[newe2*eg_w1+MCT1EGNOD1+1]=
	    (*m2).edge[i*eg_w2 + MCT2EGNODM  ]; /* 2n node = midnode */

	  (*m1).edge[newe1*eg_w1+MCT1EGCHL1]=-1; /* no childs */
	  (*m1).edge[newe2*eg_w1+MCT1EGCHL1]=-1; /* no childs */

	  (*m1).edge[i*eg_w1+MCT1EGCHL1  ]=newe1; /* the childs */
	  (*m1).edge[i*eg_w1+MCT1EGCHL1+1]=newe2; 

	  (*m1).edge[newe1*eg_w1+MCT1EGBND]=-1; /* no boundary */
	  (*m1).edge[newe2*eg_w1+MCT1EGBND]=-1; /* by default  */
	  /* if boundary .. */
	  if ((*m2).edge[i*eg_w2+MCT2EGBND]!=-1)
	    {
	      FIDX newbd1, newbd2, oldbd;

	      oldbd=(*m2).edge[i*eg_w2+MCT2EGBND];

	      /* create 2 new boundary entries */
	      newbd1=(*m1).bd_nr;
	      newbd2=(*m1).bd_nr+1;
	      (*m1).bd_nr+=2;

	      /* bd -> edge */
	      (*m1).bound[newbd1*bd_w1+MCT1BDEDGE]=newe1;
	      (*m1).bound[newbd2*bd_w1+MCT1BDEDGE]=newe2;
	      /* bd -> type */
	      (*m1).bound[newbd1*bd_w1+MCT1BDTYPE]=
		(*m1).bound[oldbd*bd_w1+MCT1BDTYPE];
	      (*m1).bound[newbd2*bd_w1+MCT1BDTYPE]=
		(*m1).bound[oldbd*bd_w1+MCT1BDTYPE];
	      /* bd -> func */
	      (*m1).bound[newbd1*bd_w1+MCT1BDFNCT]=
		(*m1).bound[oldbd*bd_w1+MCT1BDFNCT];
	      (*m1).bound[newbd2*bd_w1+MCT1BDFNCT]=
		(*m1).bound[oldbd*bd_w1+MCT1BDFNCT];
	      /* bd -> orie */
	      (*m1).bound[newbd1*bd_w1+MCT1BDORIE]=
		(*m1).bound[oldbd*bd_w1+MCT1BDORIE];
	      (*m1).bound[newbd2*bd_w1+MCT1BDORIE]=
		-(*m1).bound[oldbd*bd_w1+MCT1BDORIE];
	      /* bd -> sseg */
	      (*m1).bound[newbd1*bd_w1+MCT1BDSSEG]=
		(*m1).bound[oldbd*bd_w1+MCT1BDSSEG];
	      (*m1).bound[newbd2*bd_w1+MCT1BDSSEG]=
		(*m1).bound[oldbd*bd_w1+MCT1BDSSEG];

	      (*m1).edge[newe1*eg_w1+MCT1EGBND]=newbd1; /* set to valid */
	      (*m1).edge[newe2*eg_w1+MCT1EGBND]=newbd2; 
	    }
	  (*m1).edge[newe1*eg_w1+MCT1EGLVL]=level+1; /* one level */
	  (*m1).edge[newe2*eg_w1+MCT1EGLVL]=level+1; /* higher    */

	  (*m1).edge[newe1*eg_w1+MCT1EGPCSU]=-1; /* no pcsurf */
	  (*m1).edge[newe2*eg_w1+MCT1EGPCSU]=-1; /* by default  */
	  /* if pcsurf .. */
	  if ((*m2).edge[i*eg_w2+MCT2EGPCSU]!=-1)
	    {
	      FIDX newps1, newps2, oldps;

	      oldps=(*m2).edge[i*eg_w2+MCT2EGPCSU];

	      /* create 2 new pcsurf entries */
	      newps1=(*m1).ps_nr;
	      newps2=(*m1).ps_nr+1;
	      (*m1).ps_nr+=2;

	      /* ps -> surf */
	      (*m1).pcsurf[newps1*ps_w1+MCXXPSSURF]=newe1;
	      (*m1).pcsurf[newps2*ps_w1+MCXXPSSURF]=newe2;
	      /* ps -> vole is set later*/
	      (*m1).pcsurf[newps1*ps_w1+MCXXPSVOLE]=-1;
	      (*m1).pcsurf[newps2*ps_w1+MCXXPSVOLE]=-1;
	      /* ps -> orie */
	      (*m1).pcsurf[newps1*ps_w1+MCXXPSORIE]=
		(*m2).pcsurf[oldps*ps_w2+MCXXPSORIE];
	      (*m1).pcsurf[newps2*ps_w1+MCXXPSORIE]=
		-(*m2).pcsurf[oldps*ps_w2+MCXXPSORIE];
	      /* ps -> crit */
	      (*m1).pcsurf[newps1*ps_w1+MCXXPSCRIT]=
		(*m2).pcsurf[oldps*ps_w2+MCXXPSCRIT];
	      (*m1).pcsurf[newps2*ps_w1+MCXXPSCRIT]=
		(*m2).pcsurf[oldps*ps_w2+MCXXPSCRIT];

	      (*m1).edge[newe1*eg_w1+MCT1EGPCSU]=newps1; /* set to valid */
	      (*m1).edge[newe2*eg_w1+MCT1EGPCSU]=newps2;
	    }
	}
    }
  
  /* pcvol, just copy */
  for (i=0; i<(*m2).pv_nr; i++)
    {
      for (j=0; j<MCXXPVLN; j++)
	{
	  (*m1).pcvol[i*pv_w1 + j]= (*m2).pcvol[i*pv_w2 + j];
	}
    }
  (*m1).pv_nr=(*m2).pv_nr;

  /* faces and elements */
  for (i=0; i<(*m2).fc_nr; i++)
    {
      FIDX nodes[3], edges[3], nodiedge[3], newedges[3];

      for (j=0; j<3; j++)
	edges[j]=(*m2).face[i*fc_w2+MCT2FCEDG1+j];
      
      /* first, get the edges into order,
	 first edge stays first and defines the first two points
      */
      nodes[0]=(*m2).edge[edges[0]*eg_w2+MCT2EGNOD1];
      nodes[1]=(*m2).edge[edges[0]*eg_w2+MCT2EGNOD1+1];
      nodiedge[0]=0; /* edge 0 is first node first */
      
      /* now see which of the two other edges has node[1] as well */
      if ((nodes[1]!=(*m2).edge[edges[1]*eg_w2+MCT2EGNOD1])&&
	  (nodes[1]!=(*m2).edge[edges[1]*eg_w2+MCT2EGNOD1+1]))
	{
	  /* edges 2 and 3 have to be swaped */
	  j=edges[2];
	  edges[2]=edges[1];
	  edges[1]=j;
	}
      /* get the 3rd node */
      if (nodes[1]==(*m2).edge[edges[1]*eg_w2+MCT2EGNOD1  ])
	{
	  nodiedge[1]=0; /* edge 1 is first node first */
	  /* ok, the second node of the second edge is the 3rd node of
	     the face
	  */
	  nodes[2]=(*m2).edge[edges[1]*eg_w2+MCT2EGNOD1+1];
	}
      else if (nodes[1]==(*m2).edge[edges[1]*eg_w2+MCT2EGNOD1+1])
	{
	  nodiedge[1]=1; /* edge 1 is second node first */
	  /* ok, the first node of the second edge is the 3rd node of
	     the face
	  */
	  nodes[2]=(*m2).edge[edges[1]*eg_w2+MCT2EGNOD1  ];
	}
      else
	{ 
	  /* cry */
	  fprintf(stderr,
		  "mesh_t2_to_t1: face with one node only in one edge!\n");
	  return FAIL;
	}

      if (nodes[2]==(*m2).edge[edges[2]*eg_w2+MCT2EGNOD1  ])
	{
	  nodiedge[2]=0; /* edge 2 is first node first */
	}
      else if (nodes[2]==(*m2).edge[edges[2]*eg_w2+MCT2EGNOD1+1])
	{
	  nodiedge[2]=1; /* edge 2 is second node first */
	}
      else
	{ 
	  /* cry */
	  fprintf(stderr,
		  "mesh_t2_to_t1: face with 3rd node not in face!\n");
	  return FAIL;
	}

      /* build first 3 faces + elements (outer 3) */
      for (j=0; j<3; j++)
	{
	  FIDX newedge,old1,old2,chld1,chld2;

	  /* outer edges */
	  old1=edges[j]; old2=edges[(j+2)%3];
	  chld1=(*m1).edge[old1*eg_w1+MCT1EGCHL1+ nodiedge[j]];
	  chld2=(*m1).edge[old2*eg_w1+MCT1EGCHL1+ nodiedge[(j+2)%3]];

	  /* new inner edge */
	  newedge=(*m1).eg_nr; (*m1).eg_nr++;
	  newedges[j]=newedge;
	  (*m1).edge[newedge*eg_w1+MCT1EGNOD1  ]=
	    (*m2).edge[old1*eg_w2+MCT2EGNODM];
	  (*m1).edge[newedge*eg_w1+MCT1EGNOD1+1]=
	    (*m2).edge[old2*eg_w2+MCT2EGNODM];
	  (*m1).edge[newedge*eg_w1+MCT1EGCHL1]=-1;
	  (*m1).edge[newedge*eg_w1+MCT1EGBND ]=-1;
	  (*m1).edge[newedge*eg_w1+MCT1EGPCSU]=-1;

	  /*build the face */
	  (*m1).face[(4*i+j)*fc_w1+MCT1FCEDG1  ]=chld1;
	  (*m1).face[(4*i+j)*fc_w1+MCT1FCEDG1+1]=newedge;
	  (*m1).face[(4*i+j)*fc_w1+MCT1FCEDG1+2]=chld2;
	  (*m1).face[(4*i+j)*fc_w1+MCT1FCRHSF]=
	    (*m2).face[i*fc_w2+MCT2FCRHSF];

	  /*if needed, build the pcvol */
	  if ((*m2).face[i*fc_w2+MCT2FCPCVL]!=-1)
	    {
	      FIDX new_,old,bla;
	      new_=(*m1).pv_nr;
	      old=(*m2).face[i*fc_w2+MCT2FCPCVL];

	      for (bla=0; bla<MCXXPVLN; bla++)
		{
		  (*m1).pcvol[new_*pv_w1 + bla]= (*m2).pcvol[old*pv_w2 + bla];
		}
	      (*m1).pcvol[new_*pv_w1 + MCXXPVVOLM]=(4*i+j);

	      (*m1).pv_nr++;
	      (*m1).face[(4*i+j)*fc_w1+MCT1FCPCVL]= new_;
	    }
	  else
	    {
	      (*m1).face[(4*i+j)*fc_w1+MCT1FCPCVL]= -1;
	    }
	    
	  /* check if the edges of the new face are pcsurf edges, if so,
	     let them know the name of the new face */
	  for (elj=0; elj<3; elj++)
	    {
	      FIDX eg, pcsu;
	      eg=(*m1).face[(4*i+j)*fc_w1+MCT1FCEDG1+elj];
	      pcsu=(*m1).edge[eg*eg_w1+MCT1EGPCSU];
	      if (pcsu!=-1)
		{
		  if ((*m1).pcsurf[pcsu*ps_w1+MCXXPSVOLE]==-1)
		    {
		      (*m1).pcsurf[pcsu*ps_w1+MCXXPSVOLE]=(4*i+j);
		      (*m1).pcsurf[pcsu*ps_w1+MCXXPSVOLE+1]=-1;
		    }
		  else
		    {
		      (*m1).pcsurf[pcsu*ps_w1+MCXXPSVOLE+1]=(4*i+j);
		    }
		}
	    }
	  
	  /* build the element */
	  (*m1).elem[(4*i+j)*el_w1+MCT1ELNOD1  ]=nodes[j];
	  (*m1).elem[(4*i+j)*el_w1+MCT1ELNOD1+1]=
	    (*m2).edge[old1*eg_w2+MCT2EGNODM];
	  (*m1).elem[(4*i+j)*el_w1+MCT1ELNOD1+2]=
	    (*m2).edge[old2*eg_w2+MCT2EGNODM];

	  /* build the inner face + element */
	  (*m1).face[(4*i+3)*fc_w1+MCT1FCEDG1+j]=newedges[j];
	  (*m1).face[(4*i+3)*fc_w1+MCT1FCRHSF]=
	    (*m2).face[i*fc_w2+MCT2FCRHSF];
	  (*m1).face[(4*i+3)*fc_w1+MCT1FCPCVL]=
	    (*m2).face[i*fc_w2+MCT2FCPCVL];
	  (*m1).elem[(4*i+3)*el_w1+MCT1ELNOD1+j]=
	    (*m2).edge[old1*eg_w2+MCT2EGNODM];
	}
      
    }
  (*m1).fc_nr = 4*(*m2).fc_nr;
  (*m1).el_nr = (*m1).fc_nr;



  /* no volumes */

  /* pccrit, just copy */
  for (i=0; i<(*m2).pc_nr; i++)
    {
      for (j=0; j<MCXXPCLN; j++)
	{
	  (*m1).pccrit[i*pc_w1 + j]= (*m2).pccrit[i*pc_w2 + j];
	}
    }
  (*m1).pc_nr=(*m2).pc_nr;

  /* function, just copy */
  for (i=0; i<(*m2).fu_nr; i++)
    {
      for (j=0; j<MC2XFULN; j++)
	{
	  (*m1).func[i*fu_w1 + j] = (*m2).func[i*fu_w2 + j];
	}
    }
  (*m1).fu_nr=(*m2).fu_nr;

  /* parameter, just copy */
  for (i=0; i<(*m2).pa_nr; i++)
    {
      for (j=0; j<MC2XPALN; j++)
	{
	  (*m1).para[i*pa_w1 + j] = (*m2).para[i*pa_w2 + j];
	}
    }
  (*m1).pa_nr=(*m2).pa_nr;


  /* shape segments, just copy */
  for (i=0; i<(*m2).sg_nr; i++)
    {
      for (j=0; j<MC2XSGLN; j++)
	{
	  (*m1).sseg[i*sg_w1 + j] = (*m2).sseg[i*sg_w2 + j];
	}
    }
  (*m1).sg_nr=(*m2).sg_nr;


  /* shape parameter, just copy */
  for (i=0; i<(*m2).sp_nr; i++)
    {
      for (j=0; j<MC2XSPLN; j++)
	{
	  (*m1).spar[i*sp_w1 + j] = (*m2).spar[i*sp_w2 + j];
	}
    }
  (*m1).sp_nr=(*m2).sp_nr;


  return SUCCESS;
}


/********************************************************************/
/*                                                                  */
/*         element type independend stuff                           */
/*                                                                  */
/********************************************************************/


/*FUNCTION*/
int mesh_func_eval ( struct mesh *m, FIDX tfu, double *point,
		     FIDX reslen, double *result
/* evaluates a function as defined by the mesh

   Input:  m       - the mesh
           tfu     - id of the function, points to the function
                     within m.func
	   point   - point where the function has to be evaluated
	             (vector of length m.dim)
	   reslen  - length of the provided result vector,
	             if the function type doesn't match reslen
	             this function will return FAIL

   Output: result  - vector of length reslen (provided by reference),
                     holds the result of the function evaluation on
                     SUCCESSful return

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){
  FIDX i, type, len, fu_w;
  int  err;

  fu_w=(*m).fu_w;

  if (tfu>=(*m).fu_nr)
    {
      fprintf(stderr,
	      "mesh_func_eval: tfu=%d out of range(<%d)!\n",
	      (int) tfu, (int) (*m).fu_nr);
      return FAIL;
    }
  
  type=(FIDX)(*m).func[tfu*fu_w+MC2XFUTYPE];

  if (type<100)
    { len = 1; }
  else
    {
      len   = (*m).dim;
      type -= 100;
    }

  if (len!=reslen)
    {
      fprintf(stderr,
	      "mesh_func_eval: reslen doesn't match function type!\n");
      return FAIL;
    }

  /* for the individual components of the result */
  for (i=0; i<len; i++)
    {
      switch(type)
	{
	case 0:
	  /* constant function */
	  result[i]=(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX];
	  break;
	case 1:
	  /* linear function */
	  /* constant part first, then x and y */
	  result[i] = (*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+2]
	    +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+0]*point[0]
	    +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+1]*point[1];
	  break;
	case 2:
	  /* quadratic function */
	  /* f(x,y)=c[5] + (c[4]+c[1]*y)*y+(c[2]*y+c[3]+c[0]*x)*x */
	  result[i] = (*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+5]
	    +( (*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+4]
	       +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+1]*point[1]
	       )*point[1]
	    +( (*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+2]*point[1]
	       +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+3]
	       +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+0]*point[0]
	       )*point[0];
	  break;
	case 10:
	  /* compund function 1 */
	  /* first, check which function to use */
	  if ((*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+2]
	      +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+0]*point[0]
	      +(*m).func[tfu*fu_w+MC2XFUDAT1+i*MC2XFUDATX+1]*point[1]
	      >=0)
	    {
	      /* use function found in MC2XFUDATX+3 */
	      err=mesh_func_eval( m, (FIDX) (*m).func[tfu*fu_w+MC2XFUDAT1
						     +i*MC2XFUDATX+3],
				  point, 1, &result[i] );
	      FUNCTION_FAILURE_HANDLE( err, mesh_func_eval,
				       mesh_func_eval);
	    }
	  else
	    {
	      /* use function found in MC2XFUDATX+4 */
	      err=mesh_func_eval( m, (FIDX) (*m).func[tfu*fu_w+MC2XFUDAT1
						     +i*MC2XFUDATX+4],
				  point, 1, &result[i] );
	      FUNCTION_FAILURE_HANDLE( err, mesh_func_eval,
				       mesh_func_eval);
	    }
	  break;
	default:
	  /* unimplemented type, cry */
	  fprintf(stderr,
		  "mesh_func_eval: function %d is of unknow function "
		  "type=%d  !\n", (int) tfu,
		  (int) (*m).func[tfu*fu_w+MC2XFUTYPE]);
	  return FAIL;
	}
    }
  return SUCCESS;
}


/*FUNCTION*/
int mesh_sseg_adjust_tx( struct mesh *m, int insert,
			 FIDX nI, struct vector *dIdx,
			 struct vector *dIdF,
			 int type
/* Adjusts the shape segment parts of the boundary to match the
   current parameters or evaluates the derivatives of the node
   positions and applies the chain rule to evlauate dIdF where x=x(F),
   and dIdx is given, dIdF:=dIdx*dxdF.

   In/Out: m       - the mesh, vertices are adjusted (moved), 
                     only modified if dIdF==NULL

   Input:  insert  - wether or not points may be inserted on the
                     curved boundary, ==1 may instert, ==0 not,
		     ==1 only usefull if in the pre-mesh-generator
		     stage, as this could destroy the connectivity
	   nI      - number of performance criteria, size of dIdF and
                     dIdx, ignored if dIdF==NULL
           dIdx    - derivatives of performance criteria wrt node
                     positions, i.e. dIdx[i].V[d*vx_nr+j] is
                     derivative of criterion I[i] (i=0..nI) wrt the
                     d-th coordinate of node j, ignored if dIdF==NULL
	   type    - type of the mesh, type==1 ==> T1, type==2 ==> T2

   Output: dIdF    - chain rule is applied as describe above, 
		     ignored if dIdF==NULL, otherwise, dIdF[i].V[j] is
		     the derivative of performance criterion I[i] wrt
		     spar[j]

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
			 ){
  FIDX i, j, rcnt, refines;
  int  err;
  FIDX seg, eg;

  FIDX vx_nr, eg_nr, eg_w, bd_nr, bd_w, sg_nr, sg_w;

  struct ilist **eglist; /* list of edges for each segment */
  FIDX *nd_nbs;  /* nd_nbs[i*2+j]=k,  j=1,2:  node i has neigbour k */
  FIDX *nodes;   /* list of nodes on the segment */
  FIDX *nd_segs; /* details the segment in which a node has been set,
		   -1 if it hasn't yet been set */
  FIDX ignore1, ignore2; /* at most two nodes may be necessary to
			   ignore because they have been set
			   previously */

  FIDX l_mcBDSSEG, l_mcBDEDGE, l_mcEGCHL1, l_mcVXSTRT;

  switch (type)
    {
    case 1: /* T1 mesh */
      l_mcVXSTRT=MCT1VXSTRT;
      l_mcBDSSEG=MCT1BDSSEG;
      l_mcBDEDGE=MCT1BDEDGE;
      l_mcEGCHL1=MCT1EGCHL1;
      break;
    case 2: /* T2 mesh */
      l_mcVXSTRT=MCT2VXSTRT;
      l_mcBDSSEG=MCT2BDSSEG;
      l_mcBDEDGE=MCT2BDEDGE;
      l_mcEGCHL1=MCT2EGCHL1;
      break;
    default:
      fprintf(stderr,"mesh_sseg_adjust_tx: unknown type=%d\n",type);
      return FAIL;
    }

  if (dIdF!=NULL)
    {
      if (insert!=0)
	{
	  fprintf(stderr,"mesh_sseg_adjust_tx: insert!=0 "
		  "does not make sense if the derivatives are "
		  "to be evaluated\n");
	  return FAIL;
	}
	
      /* init the derivatives */
      for (i=0; i<nI; i++)
	for (j=0; j< m->sp_nr; j++)
	  dIdF[i].V[j]=0.0;
    }


  sg_nr = m->sg_nr;

  eg_w  = m->eg_w;
  bd_w  = m->bd_w;
  sg_w  = m->sg_w;

  TRY_MALLOC( eglist, sg_nr, struct ilist*, mesh_sseg_adjust_tx);
  /* init the list */
  for (i=0; i<sg_nr; i++)
    eglist[i]=NULL;


  /* set the number of refines according to parameter insert */
  if (insert==1)
    refines=2;
  else
    refines=0;

  /******************************************************************/
  /* refine curved segments and build an edge list for each segment */
  /******************************************************************/
  for (rcnt=0; rcnt<=refines; rcnt++)
    {
      /* build a list of edges for each segment */
      for (i=0; i<(*m).bd_nr; i++)
	{
	  seg = m->bound[i*bd_w+l_mcBDSSEG];
	  eg  = m->bound[i*bd_w+l_mcBDEDGE];

	  if ((seg!=-1)&&(m->edge[eg*eg_w+l_mcEGCHL1]==-1))
	    {
	      ilist_sorted_insert( &eglist[seg], eg);
	    }
	}
      /* refine */
      if (rcnt<refines)
	{
	  for (seg=0; seg<sg_nr; seg++)
	    {
	      FIDX stype=m->sseg[seg*sg_w+MC2XSGTYPE];

	      if (stype==2)
		{
		  struct ilist *this_;
		  
		  this_=eglist[seg];

		  /* split each edge of this segment */
		  while (this_!=NULL)
		    {
		      err=mesh_split_edge_t1( m, this_->data);
		      FUNCTION_FAILURE_HANDLE(err, mesh_split_edge_t1,
					      mesh_sseg_adjust_tx);
		      this_=this_->next;
		    }
		}
	      /* the edge list is outdated now, so we start over */
	      ilist_free(&eglist[seg]);
	      eglist[seg]=NULL;
	    } /* end loop segments */
	  
	} /* end if still refine */
      
    }/* end loop possible refinements + build edge list */

  /* thevertex,  edge, ... number may have changed up to here */
  vx_nr = m->vx_nr;
  eg_nr = m->eg_nr;
  bd_nr = m->bd_nr;
  sg_nr = m->sg_nr;
  
  TRY_MALLOC( nd_nbs, 2*vx_nr, FIDX, mesh_sseg_adjust_tx);
  TRY_MALLOC( nodes, vx_nr, FIDX, mesh_sseg_adjust_tx);
  TRY_MALLOC( nd_segs, vx_nr, FIDX, mesh_sseg_adjust_tx);

  /* clear nd_segs, it will be set when the nodes of the segment are
     sorted, along with ignore1, and ignore2 */
  for (i=0; i<vx_nr; i++)
    nd_segs[i]=-1;

  for (seg=0; seg<sg_nr; seg++)
    {
      struct ilist *this_;
      FIDX seg_nds, tnode, pnode;      

      /**********************************************************/
      /* for each node on the edges mark its neighbouring nodes */
      /**********************************************************/
      this_=eglist[seg];

      /* clear nd_nbs */
      for (i=0; i<2*vx_nr; i++)
	nd_nbs[i]=-1;

      while (this_!=NULL)
	{
	  FIDX eg=this_->data;

	  if (type==1)
	    {
	      FIDX n1= m->edge[eg*eg_w+MCT1EGNOD1  ];
	      FIDX n2= m->edge[eg*eg_w+MCT1EGNOD1+1];

	      if (nd_nbs[n1*2]==-1)
		nd_nbs[n1*2]=n2;
	      else if (nd_nbs[n1*2+1]==-1)
		nd_nbs[n1*2+1]=n2;
	      else 
		{
		  fprintf(stderr,"mesh_sseg_adjust_tx: "
			  "node multiplicity error\n");
		  return FAIL;
		}

	      if (nd_nbs[n2*2]==-1)
		nd_nbs[n2*2]=n1;
	      else if (nd_nbs[n2*2+1]==-1)
		nd_nbs[n2*2+1]=n1;
	      else 
		{
		  fprintf(stderr,"mesh_sseg_adjust_tx: "
			  "node multiplicity error\n");
		  return FAIL;
		}
	    } /* end type=1 edges */
	  else if (type==2)
	    {
	      FIDX n1= m->edge[eg*eg_w+MCT2EGNOD1  ];
	      FIDX n2= m->edge[eg*eg_w+MCT2EGNOD1+1];
	      FIDX nm= m->edge[eg*eg_w+MCT2EGNODM];

	      if (nd_nbs[n1*2]==-1)
		nd_nbs[n1*2]=nm;
	      else if (nd_nbs[n1*2+1]==-1)
		nd_nbs[n1*2+1]=nm;
	      else 
		{
		  fprintf(stderr,"mesh_sseg_adjust_tx: "
			  "node multiplicity error\n");
		  return FAIL;
		}

	      if (nd_nbs[n2*2]==-1)
		nd_nbs[n2*2]=nm;
	      else if (nd_nbs[n2*2+1]==-1)
		nd_nbs[n2*2+1]=nm;
	      else 
		{
		  fprintf(stderr,"mesh_sseg_adjust_tx: "
			  "node multiplicity error\n");
		  return FAIL;
		}
	      /* the midnode really should have only two neighbours */
	      nd_nbs[nm*2  ]=n1;
	      nd_nbs[nm*2+1]=n2;
	    } /* end type=2 edges */
	  else
	    {
	      fprintf(stderr,"mesh_sseg_adjust_tx: "
		      "unknown type in node neigbours search\n");
	      return FAIL;
	    }

	  this_=this_->next;
	}


      /**********************************************************/
      /* order the segment                                      */
      /**********************************************************/
      ignore1=-1;                         /* no nodes to be ignored yet */
      ignore2=-1;
      seg_nds=0;                          /* nodes of this segment */
      pnode=-2;                           /* previous node */
      tnode=m->sseg[seg*sg_w+MC2XSGNOD1]; /* this node     */
      while ((tnode>=0)&&(seg_nds<vx_nr))
	{
	  FIDX nb;

	  nb=nd_nbs[tnode*2  ];
	  /* if this neighbour is the previous node, use other
	     neighbour */
	  if (nb==pnode)
	    nb=nd_nbs[tnode*2+1];

	  /* check if this node has already been set */
	  if (nd_segs[tnode]==-1)
	    { /* has not been set yet, mark as set, continue */
	      nd_segs[tnode]=seg;
	    }
	  else
	    { /* has already been set, mark to be ignored */
	      if (ignore1==-1)
		ignore1=seg_nds;
	      else if (ignore2==-1)
		ignore2=seg_nds;
	      else
		{ /* something is wrong, we've ran out of ignores */
		  fprintf(stderr,"mesh_sseg_adjust_tx: "
			  "node %d in > 2 segments\n", (int) tnode);
		  return FAIL;
		}
	    }
	  
	  /* add this node to the node list */
	  nodes[seg_nds]=tnode;
	  seg_nds++;

	  /* continue with the neighbour */
	  pnode=tnode;
	  tnode=nb;
	}

      /* printf("seg=%d,  found %d nodes\n", (int) seg, (int) seg_nds); /* */

      /* check we didn't run into trouble */
      if (seg_nds>=vx_nr)
	{
	  fprintf(stderr,"mesh_sseg_adjust_tx: " 
		  "node ordering did not terminate\n");
	  return FAIL;
	}


      
      /* nodes holds now a list of the seg_nds nodes of this segment */

      /**********************************************************/
      /* adjust the nodes on the ordered segment or             */
      /* eval the derivatives                                   */
      /**********************************************************/
      err=mesh_sseg_adjust_nodes(m, seg, seg_nds, nodes,
				 ignore1, ignore2,
				 l_mcVXSTRT, nI, dIdx, dIdF);
      FUNCTION_FAILURE_HANDLE(err, mesh_sseg_adjust_nodes,
			      mesh_sseg_adjust_tx);
      /* next segment */
    }/* end loop segments */

  /* free local data */
  for (i=0; i<sg_nr; i++)
    {
      ilist_free( &eglist[i]);
      eglist[i]=NULL;
    }
  free(eglist);
  free(nd_nbs);
  free(nodes);
  free(nd_segs);

  return SUCCESS;
}




/*FUNCTION*/
int mesh_sseg_adjust_nodes( struct mesh *m, FIDX seg, FIDX seg_nds,
			    FIDX *nodes, FIDX ignore1, FIDX ignore2,
			    FIDX l_mcVXSTRT, 
			    FIDX nI, struct vector *dIdx,
			    struct vector *dIdF
/* Adjusts the nodes listed in order in nodes according to shape
   segment seg to match the current parameters or evaluates the
   derivatives of the node positions and applies the chain rule to
   evlauate dIdF where x=x(F), and dIdx is given. 

   In/Out: m       - the mesh, vertices are adjusted (moved), 
                     only modified if dIdF==NULL
           dIdF    - chain rule is applied, adding the contributions
                     of this segment, has to be initialised outside!,
		     ignored if dIdF==NULL, otherwise, dIdF[i].V[j] is
		     the derivative of performance criterion I[i] wrt
		     spar[j]

   Input:  seg     - id of the segment
	   seg_nds - number of nodes in the list nodes
           nodes   - an ordered list of nodes
	   ignore1 - see ignore2
	   ignore2 - pointers into nodes for nodes which are to be
	             ignored, because they are set in another segment
	   l_mcVXSTRT
	           - start of the coordinates in a vertex entry in
	             this mesh
           nI      - number of performance criteria, size of dIdF and
                     dIdx, ignored if dIdF==NULL
           dIdx    - derivatives of performance criteria wrt node
                     positions, i.e. dIdx[i].V[d*vx_nr+j] is
                     derivative of criterion I[i] (i=0..nI) wrt the
                     d-th coordinate of node j, ignored if dIdF==NULL


   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
			 ){
  FIDX i, j, d;
  
  const FIDX vx_w=m->vx_w, sg_w=m->sg_w, sp_w=m->sp_w, vx_nr=m->vx_nr;
  const FIDX dim=m->dim;

  const double dt= 1.0/((double)seg_nds-1);

  switch (m->sseg[seg*sg_w+MC2XSGTYPE])
    {
    case 1: /* line */
      {
	if (dIdF==NULL)
	  {
	    double lpar[4];
	    
	    for (i=0; i<4; i++)
	      lpar[i]= m->spar[m->sseg[seg*sg_w+MC2XSGPAR1+i]*sp_w];

	    for (i=0; i<seg_nds; i++)
	      {
		const FIDX nd=nodes[i];
		const double t=i*dt;
		const double omt=1.0-t;

		if ((i!=ignore1)&&(i!=ignore2))
		  for (d=0; d<dim; d++)
		    {
		      m->vertex[nd*vx_w+l_mcVXSTRT+d]=
			omt*lpar[  d] +  t* lpar[2+d];
		    }
	      }
	  }
	else /* if deriv!=NULL */
	  {
	    for (j=0; j<nI; j++) /* loop criteria */
	      {
		double dlpar[4];
	    
		for (i=0; i<4; i++)
		  dlpar[i]= 0.0;

		for (i=0; i<seg_nds; i++)
		  {
		    const FIDX nd=nodes[i];
		    const double t=i*dt;
		    const double omt=1.0-t;

		    if ((i!=ignore1)&&(i!=ignore2))
		      for (d=0; d<dim; d++)
			{
			  dlpar[  d]+=omt*dIdx[j].V[d*vx_nr+nd];
			  dlpar[2+d]+=t*dIdx[j].V[d*vx_nr+nd];
			}
		  } /* end loop segment nodes / t */
		
		/* add up the derivative */
		for (i=0; i<4; i++)
		  dIdF[j].V[m->sseg[seg*sg_w+MC2XSGPAR1+i]]+=dlpar[i];

	      } /* end loop criteria */
	  } /* end if deriv!=NULL */
      }
      break;
    case 2: /* b-spline */
      {
	if (dIdF==NULL)
	  {
	    double lpar[8];
	    
	    for (i=0; i<8; i++)
	      lpar[i]= m->spar[m->sseg[seg*sg_w+MC2XSGPAR1+i]*sp_w];

	    /* the sxi and syi need to be adjusted for easy evaluation */
	    /* the first two direction need just adding the point, to give
	       control points */
	    lpar[4]+=lpar[0];
	    lpar[5]+=lpar[1];
	    /* the later two require turning around and adding the point */
	    lpar[6]=lpar[2]-lpar[6];
	    lpar[7]=lpar[3]-lpar[7]; 
	    for (i=0; i<seg_nds; i++)
	      {
		const FIDX nd=nodes[i];
		const double t=i*dt;
		const double omt=1.0-t;
		
		if ((i!=ignore1)&&(i!=ignore2))
		  for (d=0; d<dim; d++)
		    {
		      double d11, d12, d13, d21, d22;
		
		      /* eval according to de Casteljau's algorithm */
		      d11=omt*lpar[  d]+t*lpar[4+d];
		      d12=omt*lpar[4+d]+t*lpar[6+d];
		      d13=omt*lpar[6+d]+t*lpar[2+d];
		    
		      d21=omt*d11+t*d12;
		      d22=omt*d12+t*d13;

		      m->vertex[nd*vx_w+l_mcVXSTRT+d]=
			omt*d21+t*d22;
		    }
	      } /* end loop segment nodes / t */
	  }
	else /* if (deriv!=NULL) */
	  {
	    for (j=0; j<nI; j++) /* loop over criteria */
	      {
		double dlpar[8];
		for (i=0; i<8; i++)
		  dlpar[i]=0.0;

		for (i=0; i<seg_nds; i++)
		  {
		    const FIDX nd=nodes[i];
		    const double t=i*dt;
		    const double omt=1.0-t;

		    double dd21d1,dd21d2,dd21d3,  dd22d2,dd22d3,dd22d4;
		    double dxdl1,dxdl2,dxdl3,dxdl4;

		    /* derivatives of Casteljau's algorithm wrt parmeters */
		    dd21d1 = omt*omt; 
		    dd21d2 = 2*omt*t;
		    dd21d3 = t*t;

		    dd22d2 = omt*omt; 
		    dd22d3 = 2*omt*t;
		    dd22d4 = t*t;

		    dxdl1 = omt*dd21d1           ;
		    dxdl2 = omt*dd21d2 + t*dd22d2;
		    dxdl3 = omt*dd21d3 + t*dd22d3;
		    dxdl4 =              t*dd22d4;

		    if ((i!=ignore1)&&(i!=ignore2))
		      for (d=0; d<dim; d++)
			{
			  const double dIdnode = dIdx[j].V[d*vx_nr+nd];
			  dlpar[ +d] += dIdnode*dxdl1;
			  dlpar[4+d] += dIdnode*dxdl2;
			  dlpar[6+d] += dIdnode*dxdl3;
			  dlpar[2+d] += dIdnode*dxdl4;
			}
		  } /* end loop segment nodes / t */

		/* compensate for the interdependancy (chain rule bwd) */
		dlpar[0]+=dlpar[4];
		dlpar[1]+=dlpar[5];
		
		dlpar[2]+=dlpar[6]; dlpar[6]*=-1.0;
		dlpar[3]+=dlpar[7]; dlpar[7]*=-1.0;

		/* add up */
		for (i=0; i<8; i++)
		  dIdF[j].V[m->sseg[seg*sg_w+MC2XSGPAR1+i]] += dlpar[i];
	      } /* end loop over criteria */
	  } /* end if deriv!=NULL */
      } /* end b-spline block */
      break;
    default:
      fprintf(stderr,"mesh_sseg_adjust_nodes: "
	      "unknown segment type\n");
      return FAIL;
    }
  return SUCCESS;
}





/*FUNCTION*/
int mesh_more_vertices ( struct mesh *m, FIDX need
/* increases the number of vertices that can be held in the mesh by
   reallocating memory

   Input:  need       - the number of additional vertices that are
                        required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX vx_w, vx_max, newmax;

  vx_w=(*m).vx_w;
  vx_max=(*m).vx_max;

  /* just get n times as much */
  newmax=4*vx_max;
  TRY_REALLOC( (*m).vertex, newmax*vx_w, double, mesh_more_vertices);

  (*m).vx_max=newmax;

  return SUCCESS;
}



/*FUNCTION*/
int mesh_more_edges ( struct mesh *m , FIDX need
/* increases the number of edges that can be held in the mesh by
   reallocating memory

   Input:  need       - the number of additional edges that are
                        required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX eg_w, eg_max, newmax;

  eg_w=(*m).eg_w;
  eg_max=(*m).eg_max;

  /* just get n times as much */
  newmax=4*eg_max;
  TRY_REALLOC( (*m).edge, newmax*eg_w, FIDX, mesh_more_edges);

  (*m).eg_max=newmax;

  return SUCCESS;
}



/*FUNCTION*/
int mesh_more_faces ( struct mesh *m , FIDX need
/* increases the number of faces that can be held in the mesh by
   reallocating memory

   Input:  need       - the number of additional faces that are
                        required
   In/Out: m          - the mesh 

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX fc_w, fc_max, newmax;

  fc_w=(*m).fc_w;
  fc_max=(*m).fc_max;

  /* just get n times as much */
  newmax=4*fc_max;
  TRY_REALLOC( (*m).face, newmax*fc_w, FIDX, mesh_more_faces);

  (*m).fc_max=newmax;

  return SUCCESS;
}



/*FUNCTION*/
int mesh_more_elems ( struct mesh *m , FIDX need
/* increases the number of elems that can be held in the mesh by
   reallocating memory

   Input:  need       - the number of additional elems that are
                        required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX el_w, el_max, newmax;

  el_w=(*m).el_w;
  el_max=(*m).el_max;

  /* just get n times as much */
  newmax=4*el_max;
  TRY_REALLOC( (*m).elem, newmax*el_w, FIDX, mesh_more_elems);

  (*m).el_max=newmax;

  return SUCCESS;
}


/*FUNCTION*/
int mesh_more_boundary ( struct mesh *m , FIDX need
/* increases the number of boundary entries that can be held in the
   mesh by reallocating memory

   Input:  need       - the number of additional boundary entries that
                        are required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX bd_w, bd_max, newmax;

  bd_w=(*m).bd_w;
  bd_max=(*m).bd_max;

  /* just get n times as much */
  newmax=2*bd_max;
  TRY_REALLOC( (*m).bound, newmax*bd_w, FIDX, mesh_more_boundary);

  (*m).bd_max=newmax;

  return SUCCESS;
}

/*FUNCTION*/
int mesh_more_pcsurf ( struct mesh *m , FIDX need
/* increases the number of pcsurf entries that can be held in the
   mesh by reallocating memory

   Input:  need       - the number of additional pcsurf entries that
                        are required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX ps_w, ps_max, newmax;

  ps_w=(*m).ps_w;
  ps_max=(*m).ps_max;

  /* just get n times as much */
  newmax=3*ps_max;
  TRY_REALLOC( (*m).pcsurf, newmax*ps_w, FIDX, mesh_more_pcsurf);

  (*m).ps_max=newmax;

  return SUCCESS;
}

/*FUNCTION*/
int mesh_more_pcvol ( struct mesh *m , FIDX need
/* increases the number of pcvol entries that can be held in the
   mesh by reallocating memory

   Input:  need       - the number of additional pcvol entries that
                        are required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX pv_w, pv_max, newmax;

  pv_w=(*m).pv_w;
  pv_max=(*m).pv_max;

  printf("pv_nr=%d, pv_max=%d, need=%d\n", (*m).pv_nr, (*m).pv_max, need);

  /* just get n times as much */
  newmax=4*pv_max;
  TRY_REALLOC( (*m).pcvol, newmax*pv_w, FIDX, mesh_more_pcvol);

  (*m).pv_max=newmax;

  return SUCCESS;
}



/*FUNCTION*/
int mesh_more_hierarchy ( struct mesh *m , FIDX need
/* increases the number of hierarchy entries that can be held in the
   mesh by reallocating memory

   Input:  need       - the number of additional hierarchy entries that
                        are required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX hi_w, hi_max, newmax;

  hi_w=(*m).hi_w;
  hi_max=(*m).hi_max;

  /* just get n times as much */
  newmax=4*hi_max;
  TRY_REALLOC( (*m).hier, newmax*hi_w, FIDX, mesh_more_hierarchy);

  (*m).hi_max=newmax;

  return SUCCESS;
}


/*FUNCTION*/
int mesh_more_elhiers ( struct mesh *m , FIDX need
/* increases the number of elementwise hierarchy entries that can be
   held in the mesh by reallocating memory

   Input:  need       - the number of additional elementwise hierarchy
                        entries that are required
   In/Out: m          - the mesh

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		      ){
  FIDX eh_w, eh_max, newmax;

  eh_w=(*m).eh_w;
  eh_max=(*m).eh_max;

  /* just get n times as much */
  newmax=5*eh_max;
  TRY_REALLOC( (*m).elhier, newmax*eh_w, FIDX, mesh_more_elhiers);

  (*m).eh_max=newmax;

  return SUCCESS;
}




/*FUNCTION*/
void mesh_free ( struct mesh *m
/* frees the memory internally allocated in the mesh data struct */
		 ){
  (*m).vx_nr  = 0;
  (*m).vx_max = 0;
  free((*m).vertex);
  (*m).vertex = NULL;

  (*m).el_nr  = 0;
  (*m).el_max = 0;
  free((*m).elem);
  (*m).elem = NULL;

  (*m).eg_nr  = 0;
  (*m).eg_max = 0;
  free((*m).edge);
  (*m).edge = NULL;

  (*m).fc_nr  = 0;
  (*m).fc_max = 0;
  free((*m).face);
  (*m).face = NULL;

  (*m).vo_nr  = 0;
  (*m).vo_max = 0;
  free((*m).vols);
  (*m).vols = NULL;

  (*m).hi_nr  = 0;
  (*m).hi_max = 0;
  free((*m).hier);
  (*m).hier = NULL;

  (*m).eh_nr  = 0;
  (*m).eh_max = 0;
  free((*m).elhier);
  (*m).elhier = NULL;

  (*m).bd_nr  = 0;
  (*m).bd_max = 0;
  free((*m).bound);
  (*m).bound = NULL;

  (*m).ps_nr  = 0;
  (*m).ps_max = 0;
  free((*m).pcsurf);
  (*m).pcsurf = NULL;

  (*m).pv_nr  = 0;
  (*m).pv_max = 0;
  free((*m).pcvol);
  (*m).pcvol = NULL;

  (*m).pc_nr  = 0;
  (*m).pc_max = 0;
  free((*m).pccrit);
  (*m).pccrit = NULL;

  (*m).fu_nr  = 0;
  (*m).fu_max = 0;
  free((*m).func);
  (*m).func = NULL;

  (*m).pa_nr  = 0;
  (*m).pa_max = 0;
  free((*m).para);
  (*m).para = NULL;

  (*m).sg_nr  = 0;
  (*m).sg_max = 0;
  free((*m).sseg);
  (*m).sseg = NULL;

  (*m).sp_nr  = 0;
  (*m).sp_max = 0;
  free((*m).spar);
  (*m).spar = NULL;

  return;
}


/*FUNCTION*/
void ilist_free( struct ilist **head
/* empties ilist (deletes all entries and defines head to be NULL */
		){
  struct ilist *this_, *next;
  
  this_ = *head;

  while (this_ != NULL)
    {
      next = this_->next;

      free(this_);

      this_ = next;
    }

  *head = NULL;
  
  return;
}


/*FUNCTION*/
int ilist_sorted_insert( struct ilist **head, FIDX newname
/* if the entry newname does not exist in the sorted list, it is
   inserted at the appropriate position */
		){
  struct ilist *this_, *next, *hlp;

  /* printf("head=%p    newname=%d\n", *head, newname);
     printf("head->data = %d    head->next=%p\n",
     (*head)->data, (*head)->next);
     printf("this_->data = %d    this_->next=%p\n", this_->data, next); */

  if ( *head == NULL)
    {
      /* first item in the list anyway */
      TRY_MALLOC( *head, 1, struct ilist, ilist_sorted_insert);
      
      (*head)->data = newname;
      (*head)->next = NULL;
      
      return SUCCESS;
    }

  /* otherwise we were given a non-empty list */
  this_ = *head;

  if ( (*head)->data > newname )
    {
      /* even the head has name > newname, so we need to insert right
	 at the head of the list */
      TRY_MALLOC( hlp, 1, struct ilist, ilist_sorted_insert);
	      
      hlp->data = newname;
      hlp->next = (*head);

      (*head) = hlp;
	      
      return SUCCESS;
    }


  /* we are really going inside the list */
  while (this_ != NULL)
    {
      next = this_->next;

      if ( this_->data == newname )
	{
	  /* exists alredy */
	  return SUCCESS;
	}

      if ( next != NULL ) 
	{
	  if ( next->data <= newname )
	    {
	      /* next one still <= newname  ==> just keep searching for
		 the appropriate spot */
	      this_ = next;
	    }
	  else
	    {
	      /* the next one is > newname, so this_ is the right spot
		 to insert */
	      TRY_MALLOC( hlp, 1, struct ilist, ilist_sorted_insert);
	      
	      hlp ->data = newname;
	      hlp ->next = next;

	      this_->next = hlp;
	      
	      return SUCCESS;
	    }
	}
      else /* next == NULL */
	{
	  /* no next one, so append newname at the end */
	  TRY_MALLOC( hlp, 1, struct ilist, ilist_sorted_insert);
	      
	  hlp ->data = newname;
	  hlp ->next = NULL;

	  this_->next = hlp;
	  
	  return SUCCESS;
	}

      this_ = next;
    }

  fprintf(stderr, "ilist_sorted_insert: unexpected end!\n");
  return FAIL;
}




/*FUNCTION*/
int comp_intdouble_d(const void *a1, const void *b1
/* helper function for qsort, compares the double part of the
   intdouble struct, designed for descending order

   Input:  a1      - pointer to intdouble a
           b1      - pointer to intdouble b

   Return: -1   if a.d >  b.d
            0   if a.d == b.d
            1   if a.d <  b.d
*/
		     ){
  struct intdouble *a, *b;

  a=(struct intdouble *) a1;
  b=(struct intdouble *) b1;

  if ((*a).d>(*b).d)
    {
      return -1;
    }
  else if ((*a).d<(*b).d)
    {
      return 1;
    }
  else
    {
      return 0;
    }
}


/*FUNCTION*/
int mesh_write_solution_grape_t2(struct mesh *m,struct vector *x,
				FIDX ncompo,
				FIDX namlen, char *pathname
/* writes the mesh and the solution into a GRAPE readable
   file (for visualisation) 
   
   Input:  m         - the mesh
           x         - vector with data associated to the nodes
	   ncompo    - dimension of the solution vector per node which
	               shall be writen to the file, (or number of
	               solution components), the components of the
	               vector x have to be numbered such that
	               x[j*vx_nr+i] gives the j-th component in the
	               i-th node of the mesh
	   namlen    - maximal useable length of name
	   pathname  - pathname for the grape file post0.dat


   Output: (writes the files)

   Return: SUCCESS - success
           FAIL    - failure, see error message, output will not be
                     valid
*/
		     ){

  FIDX i, j, k, el_w, vx_w, vx_nr,eg_nr, eg_w,el_nr, eg_id, nd_id,nd0,nd1,nd2, *E2F;
  int flag, err;

  int *clockwise;

  double d2[2];
  long l6[6];
  long l3[3];
  double p, a,b,c,d,det;

  char *prename, *postname;
  FILE *prefile, *postfile;


  el_w =(*m).el_w;
  vx_w =(*m).vx_w;
  vx_nr=(*m).vx_nr;
  eg_nr=(*m).eg_nr;
  eg_w=(*m).eg_w;
  el_nr=(*m).el_nr;

  TRY_MALLOC( E2F,2*eg_nr, FIDX, mesh_write_solution_grape_t2);
  TRY_MALLOC( clockwise,el_nr, int, mesh_write_solution_grape_t2);

  /* initialise clockwise with 0 */
  for (i=0; i<el_nr; i++)
    {
      clockwise[i] = 0;
    }

  if (ncompo*vx_nr!=(*x).len)
    {
      fprintf( stderr,
	       "\nmesh_write_solution_grape_t2: nodenr in mesh and "
	       "dimension of vector must agree!\n");
      return FAIL;
    }

  /* prepare the names */
  TRY_MALLOC( prename, (namlen+8), char, 
	      mesh_write_solution_grape_t2 );
  TRY_MALLOC( postname, (namlen+9), char, 
	      mesh_write_solution_grape_t2 );
 
  strncpy( prename, pathname, namlen );
  strncpy( postname, pathname, namlen );

  /* make sure no overread happens */
  prename[namlen-1]='\0';
  postname[namlen-1]='\0';

  /* append ".dat" */
  strcat( prename, "pre0.dat" );
  strcat( postname, "post0.dat" );

  /* clear the buffer */
  fflush(stdout);
  
  /* open the post file */
  postfile=fopen(postname, "w");
  if (postfile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_solution_grape_t2: error opening file %s\n",
	      postname);
      return FAIL;
    }

  /* write the header */
  flag = 1;    /* Flag=1 -> write geometry */
  fwrite(&flag,1,sizeof(int),postfile);
  fwrite(&vx_nr,1,sizeof(FIDX),postfile);
  fwrite(&el_nr,1,sizeof(FIDX),postfile);

  for (i=0; i<vx_nr; i++)
    {
      /* write the node positions */
      for (j=0; j<ncompo-1; j++)
	{
	  d2[j] = (*m).vertex[i*vx_w+MCT2VXSTRT+j];
	}
      fwrite(&d2,1,2*sizeof(double),postfile);

      /* write the velocities */
      for (j=0; j<ncompo-1; j++)
	{
	  d2[j] = (*x).V[j*vx_nr+i];
	}
      fwrite(&d2,1,2*sizeof(double),postfile);
     
      /* write the pressure */
      p = (*x).V[(ncompo-1)*vx_nr+i];
      fwrite(&p,1,sizeof(double),postfile);
    }

  putchar('.');
  fflush(stdout);

  /* write nodes of each element in counter-clockwise order */
  for (i=0; i<el_nr; i++)
    {
      for (j=0; j<6; j++)
	{
	  l6[j] = (*m).elem[i*el_w+MCT2ELNOD1+j];
	}
      nd0 = (*m).elem[i*el_w+MCT2ELNOD1+0];
      nd1 = (*m).elem[i*el_w+MCT2ELNOD1+1];
      nd2 = (*m).elem[i*el_w+MCT2ELNOD1+2];

      a = (*m).vertex[nd1*vx_w+MCT2VXSTRT+0] - (*m).vertex[nd0*vx_w+MCT2VXSTRT+0]; /* x(nd1) - x(nd0) */
      b = (*m).vertex[nd2*vx_w+MCT2VXSTRT+0] - (*m).vertex[nd0*vx_w+MCT2VXSTRT+0]; /* x(nd2) - x(nd0) */
      c = (*m).vertex[nd1*vx_w+MCT2VXSTRT+1] - (*m).vertex[nd0*vx_w+MCT2VXSTRT+1]; /* y(nd1) - y(nd0) */
      d = (*m).vertex[nd2*vx_w+MCT2VXSTRT+1] - (*m).vertex[nd0*vx_w+MCT2VXSTRT+1]; /* y(nd2) - y(nd0) */
      det = (a*d-b*c);

      if (det < 0.0) /* no counter-clockwise order  */
	{
	  /* swap vertices 0 and 1 */
	  nd0 = l6[0];
	  l6[0] = l6[1];
	  l6[1] = nd0;
	  /* and swap vertices 4 and 5 */
	  nd0 = l6[4];
	  l6[4] = l6[5];
	  l6[5] = nd0;
	  /* mark this element as clockwise oriented for later */
	  clockwise[i] = 1;
	}
      fwrite(&l6,1,6*sizeof(long),postfile);
    }

  /* fill E2F */
  for (i=0; i<2*eg_nr; i++)
    {
      E2F[i] = -1;
    }
  for (i=0; i<el_nr; i++)
    {
      for (j=0; j<3; j++)
	{
	  eg_id = (*m).face[i*(*m).fc_w+MCT2FCEDG1+j];
	  if (E2F[2*eg_id]==-1) /* no neighbor face found yet */
	    {
	      E2F[2*eg_id] = i;
	    }
	  else  /* already one neighbor face determined */
	    {
	      E2F[2*eg_id+1] = i;
	    }
	}
    }

  /* write neighbor faces of each edge of each element */
  for (i=0; i<el_nr; i++) /* loop over all elements */
    {
      for (j=0; j<3; j++) /* loop over all edges */ 
	{
	  for (k=0; k<3; k++) /* sort edges according to opposing node */
	    {
	      eg_id = (*m).face[i*(*m).fc_w+MCT2FCEDG1+k];
	      nd_id = (*m).elem[i*el_w+MCT2ELNOD1+j];
	      /* not the right edge, try next one */
	      if ((nd_id==(*m).edge[eg_id*eg_w+MCT2EGNOD1  ]) ||
		  (nd_id==(*m).edge[eg_id*eg_w+MCT2EGNOD1+1]))
		{
		  continue;
		}
	      else /* found right edge */
		{
		  if (E2F[2*eg_id]==i) /* find the edge's neighbor face */
		    {
		      l3[j] = E2F[2*eg_id+1] ;
		    }
		  else
		    {
		      l3[j] = E2F[2*eg_id] ;
		    }
		}
	    }
	}
      /* if this element has clockwise orientation -> swap faces */
      if (clockwise[i]==1)
	{
	  nd0 = l3[0];
	  l3[0] = l3[1];
	  l3[1] = nd0;
	}
      fwrite(&l3,1,3*sizeof(long),postfile);
    }

  putchar('.');
  fflush(stdout);

  /* close post file */
  fclose(postfile);

  /* free some memory  */
  free(postname);
  free(E2F);
  free(clockwise);


#ifdef ADAPTIVEMESH

  /* open the pre file */
  prefile=fopen(prename, "w");
  if (prefile==NULL)
    {
      fprintf(stderr,
	      "mesh_write_solution_grape_t2: error opening file %s\n",
	      prename);
      return FAIL;
    }

  /* write the header */
  flag = 0;   /* Flag=0 -> write no geometry */
  fwrite(&flag,1,sizeof(int),prefile); 
  fwrite(&vx_nr,1,sizeof(FIDX),prefile);
  fwrite(&el_nr,1,sizeof(FIDX),prefile);

  for (i=0; i<vx_nr; i++)
    {
      /* write the velocities */
      for (j=0; j<ncompo-1; j++)
	{
	  d2[j] = (*x).V[j*vx_nr+i];
	}
      fwrite(&d2,1,2*sizeof(double),prefile);
     
      /* write the pressure */
      p = (*x).V[(ncompo-1)*vx_nr+i];
      fwrite(&p,1,sizeof(double),prefile);
    }

  putchar('.');
  fflush(stdout);

  /* close pre file */
  fclose(prefile);

#endif

  free(prename);

  return SUCCESS;
}
